const DownloadTask = require("./DownloadTask");
const TarGZArchiver = require("./Archivers/TarGZArchiver");
const ZipArchiver = require("./Archivers/ZipArchiver");
const Packager = require("./Packager");
const path = require("path");
const fs = require("fs-extra");
const os = require("os");
const assert = require("assert");
const NodeSSH = require("node-ssh");

module.exports=
	class Publisher
	{
		/**
		 * @param {string} strSSHKeyPath
		 */
		constructor(strSSHKeyPath)
		{
			if (!strSSHKeyPath)
			{
				throw new Error("You must provide the ssh key path!");
			}

			this._strSSHKeyPath = strSSHKeyPath;
			this._strAgentVersionsFileName = "agentVersions.json";
		}

		/**
		 * @param {string} strAbsolutePath
		 */
		async _createFolder(strAbsolutePath)
		{
			try
			{
				await fs.mkdir(strAbsolutePath);
			}
			catch (error)
			{
				// no need to throw if folder is already created
				if (error.code !== "EEXIST")
				{
					throw error;
				}
			}
		}

		/**
		 * @param  {string} strURL
		 */
		getFileNameFromURL(strURL)
		{
			let lastSlashPosition = strURL.lastIndexOf("/");

			return strURL.slice(lastSlashPosition + 1);
		}

		/**
		 * @param {arguments} packageFolder + list of folder paths to copy in this folder.
		 */
		async createPackageFolder()
		{
			assert(arguments.length > 1, "Argument list incomplete");

			await fs.ensureDir(arguments[0]);

			for(let i = 1; i < arguments.length; i++)
			{
				const folderName = path.basename(arguments[i]);
				console.log("Copy " + folderName + "...");
				await fs.copy(arguments[i], path.join(arguments[0], folderName));
				console.log("Done copying " + folderName + ".");
			}

			return new Promise((resolve, reject) =>
			{
				resolve();
			});
		}

		/**
		 * @param  {array} arrItemsURLList
		 * @param  {string} strAbsolutePathOutput
		 */
		async downloadItems(arrItemsURLList, strAbsolutePathOutputFolder)
		{
			let fnResolve;
			let fnReject;

			const promise = new Promise(
				(_fnResolve, _fnReject) =>
				{
					fnResolve = _fnResolve;
					fnReject = _fnReject;
				}
			);

			let downloadTask = new DownloadTask();
			await this._createFolder(strAbsolutePathOutputFolder);

			try
			{
				let promiseArray = [];
				for(let item of arrItemsURLList)
				{
					promiseArray.push(
						downloadTask.downloadItem(item, strAbsolutePathOutputFolder)
					);
				}

				this.nRunningIntervalID=setInterval(() =>
					{
						downloadTask.printDownloadProgress();

						if(downloadTask.isDownloadFinished())
						{
							clearInterval(this.nRunningIntervalID);
							fnResolve();
						}
					},
					1000 /* milliseconds */);

			}
			catch (error)
			{
				fnReject(error);
			}

			return promise;
		}

		/**
		 * @param  {array} arrFileNameList
		 * @param  {string} strSourceFolderAbsolutePath
		 * @param  {string} strDestinationFolderAbsolutePath
		 */
		async decompress(arrFileNameList, strSourceFolderAbsolutePath, strDestinationFolderAbsolutePath)
		{
			const tarGZArchiver = new TarGZArchiver();
			const zipArchiver = new ZipArchiver();

			let fnResolve;
			let fnReject;

			const promise = new Promise(
				(_fnResolve, _fnReject) =>
				{
					fnResolve = _fnResolve;
					fnReject = _fnReject;
				}
			);

			await Promise.all(arrFileNameList.map(async (strFileName) =>
			{
				const extension = path.extname(strFileName);
				let archiver;
				if (extension === ".gz")
				{
					archiver = tarGZArchiver;
				}
				else if (extension === ".zip")
				{
					archiver = zipArchiver;
				}
				else
				{
					fnReject(new Error("Cannot find propper archiver to work with " + strFileName));
				}

				try
				{
					console.log("Decompressing " + strFileName + "...");

					const strArchivePath = path.join(strSourceFolderAbsolutePath, strFileName);

					await archiver.decompress(strArchivePath, strDestinationFolderAbsolutePath);
					console.log("Done "+ strFileName + ".");
				}
				catch (error)
				{
					fnReject(error);
				}
			}));

			fnResolve();

			return promise;
		}

		/**
		 * @return {string} user home folder path
		 */
		async getUserHome()
		{
			const userHomeFolder = process.env[(process.platform === "win32") ? "USERPROFILE" : "HOME"];
			return new Promise((fnResolve, fnReject) =>
			{
				fnResolve(userHomeFolder);
			});
		}

		/**
		 * @param  {string} strRepoURL
		 * @param  {string} strRemoteFilePath
		 * @param  {string} strLocalFilePath
		 */
		async downloadFileFromRepo(strRepoURL, strRemoteFilePath, strLocalFilePath)
		{
			const ssh = new NodeSSH();
			let fnResolve;
			let fnReject;

			const promise = new Promise(
				(_fnResolve, _fnReject) =>
				{
					fnResolve = _fnResolve;
					fnReject = _fnReject;
				}
			);

			try
			{
				await ssh.connect({
					host: strRepoURL,
					username: "root",
					privateKey: this._strSSHKeyPath
				});

				await ssh.getFile(strLocalFilePath, strRemoteFilePath);

				fnResolve();
			}
			catch (error)
			{
				fnReject(error);
			}
			finally
			{
				ssh.dispose();
			}

			return promise;
		}

		/**
		 * @param  {string} strOutputRepoURL
		 * @param  {string} strOutputRepoLocalFolder
		 * @param  {string} strOutputPackageFolderPath
		 */
		async uploadDirectoryToRepo(strOutputRepoURL, strOutputRepoLocalFolder, strOutputPackageFolderPath)
		{
			const ssh = new NodeSSH();
			let fnResolve;
			let fnReject;

			const promise = new Promise(
				(_fnResolve, _fnReject) =>
				{
					fnResolve = _fnResolve;
					fnReject = _fnReject;
				}
			);

			try
			{
				await ssh.connect({
					host: strOutputRepoURL,
					username: "root",
					privateKey: this._strSSHKeyPath
				});

				const uploadStatus = await ssh.putDirectory(
					strOutputPackageFolderPath,
					strOutputRepoLocalFolder,
					{
						recursive: true,
						tick: (localPath, remotePath, error) =>
						{
							if(error)
							{
								fnReject(error);
							}
							else
							{
								console.log("Successfuly uploaded " + localPath);
							}
						}
					}
				);

				if (uploadStatus)
				{
					fnResolve();
				}
				else
				{
					fnReject(new Error("Could not upload files!"));
				}
			}
			catch (error)
			{
				fnReject(error);
			}
			finally
			{
				try
				{
					ssh.dispose()
				}
				catch (error)
				{
					fnReject(error);
				}
			}

			return promise;
		}


		/**
		 * @param  {string} strOutputRepoURL
		 * @param  {Array} arrFiles
		 */
		async uploadFilesToRepo(strOutputRepoURL, arrFiles)
		{
			const ssh = new NodeSSH();

			try
			{
				await ssh.connect({
					host: strOutputRepoURL,
					username: "root",
					privateKey: this._strSSHKeyPath
				});
				const uploadStatus = await ssh.putFiles(arrFiles);
				console.log("Upload status:", uploadStatus);
			}
			finally
			{
				ssh.dispose();
			}
		}

		async setDefaultVersion(strAgentVersion, strOutputRepoURL, strOutputRepoLocalFolder)
		{
			const strTmpDownloadPath = this._generateTmpPath(this._strAgentVersionsFileName);

			await this.downloadFileFromRepo(
				strOutputRepoURL,
				`${strOutputRepoLocalFolder}/${this._strAgentVersionsFileName}`,
				strTmpDownloadPath
			);

			const strAgentVersionsFileContent = await fs.readFile(strTmpDownloadPath);
			const objAgentVersions = JSON.parse(strAgentVersionsFileContent);
			let bVersionExists = false;
			for (const strTmpAgentVersion of Object.keys(objAgentVersions))
			{
				if (strTmpAgentVersion == strAgentVersion)
				{
					bVersionExists = true;
				}
			}

			if (!bVersionExists)
			{
				throw new Error("Requested version does not exist!");
			}

			objAgentVersions.defaultVersion = strAgentVersion;
			await fs.writeFile(strTmpDownloadPath, JSON.stringify(objAgentVersions));

			let arrFileUpdates = [
				{
					local: strTmpDownloadPath,
					remote: `${strOutputRepoLocalFolder}/${this._strAgentVersionsFileName}`
				}
			];

			await this.uploadFilesToRepo(strOutputRepoURL, arrFileUpdates);

			console.log(`Default version: ${strAgentVersion} was set for repo: ${strOutputRepoURL}, path: ${strOutputRepoLocalFolder}`);
		}

		/**
		 * @param  {string} strBsiAgentPackagePath
		 * @param  {string} strFileNameList
		 * @param  {string} strDecompressOutputFolder
		 * @param  {string} strOutputPackageFolderPath
		 * @param  {string} arrOSList
		 * @param  {string} strOutputRepoURL=null
		 * @param  {string} strOutputRepoLocalFolder=null
		 */
		async publish(
			strBsiAgentPackagePath, // folder to put to archive.
			strFileNameList, // all versions of node.
			strDecompressOutputFolder, // where to take node exe from.
			strOutputPackageFolderPath, // where to put the final package.
			arrOSList,
			strOutputRepoURL = null,
			strOutputRepoLocalFolder = null // subfolder for the package.
		)
		{
			let packager = new Packager();
			let strVersion;
			let objAgentVersionsFile;
			let agentVersionOSes = {};

			console.log("Creating packages...");
			for(let nodeArchiveFile of strFileNameList)
			{
				let extensionPositionZip = nodeArchiveFile.lastIndexOf(".zip");
				let extensionPositionTarGZ = nodeArchiveFile.lastIndexOf(".tar.gz");
				let extensionPosition = extensionPositionZip > -1 ? extensionPositionZip : extensionPositionTarGZ;
				let strNodeArchiveName = nodeArchiveFile.slice(0, extensionPosition);
				let strOSDistribution;

				for(let strOS of arrOSList)
				{
					if(strNodeArchiveName.indexOf(strOS) > -1)
					{
						strOSDistribution = strOS;
					}
				}

				// copy node binary.
				let nodeBinaryName = strOSDistribution === "win" ? "node.exe" : "bin/node";
				let nodeBinaryOutputName = strOSDistribution === "win" ? "node.exe" : "node";

				let nodeBinaryPath = path.join(
					strDecompressOutputFolder,
					strNodeArchiveName,
					nodeBinaryName
				);
				let nodeBinaryOutputPath = path.join(
					strBsiAgentPackagePath,
					nodeBinaryOutputName
				);

				await fs.copy(nodeBinaryPath, nodeBinaryOutputPath);

				//Get the version.
				const packageJSONContent = await fs.readFile(path.join(strBsiAgentPackagePath, "package.json"), {encoding: "utf8"});
				const parsedJSONContent = JSON.parse(packageJSONContent);
				strVersion = parsedJSONContent["version"];

				await fs.ensureDir(path.join(strOutputPackageFolderPath, strVersion, strOSDistribution));

				//Get architecture.
				const arrArchiveNameParts = strNodeArchiveName.split("-");
				const strArchitecture = arrArchiveNameParts[arrArchiveNameParts.length - 1].split(".")[0];

				await packager.package(
					strOSDistribution,
					"BSIAgent_" + strArchitecture,
					strBsiAgentPackagePath,
					path.join(strOutputPackageFolderPath, strVersion));

				await fs.unlink(nodeBinaryOutputPath);
				const strExtension = strOSDistribution === "win" ? ".zip" : ".tar.gz";

				if(agentVersionOSes[strOSDistribution] === undefined)
				{
					agentVersionOSes[strOSDistribution] = {};
				}
				agentVersionOSes[strOSDistribution][strArchitecture] = "BSIAgent_" + strArchitecture + strExtension;

				console.log("Created package BSIAgent_" + strArchitecture + " for " + strOSDistribution);
			};

			for (let strOSName of arrOSList)
			{
				const strJSONFilePath = path.join(strOutputPackageFolderPath, strVersion, strOSName, "agentVersions.json");
				const strFileContent = JSON.stringify(agentVersionOSes[strOSName], undefined, "\t");
				await fs.writeFile(strJSONFilePath, strFileContent);
			}

			//Get agentVersions.json from repo.
			if (strOutputRepoURL && strOutputRepoLocalFolder)
			{
				try
				{
					await this.downloadFileFromRepo(
						strOutputRepoURL,
						strOutputRepoLocalFolder + "/agentVersions.json",
						path.join(strOutputPackageFolderPath, "agentVersions.json")
					);

					const strAgentVersionsFile = await fs.readFile(path.join(strOutputPackageFolderPath, "agentVersions.json"));
					objAgentVersionsFile = JSON.parse(strAgentVersionsFile);
					objAgentVersionsFile.latestVersion = strVersion;
				}
				catch (error)
				{
					// There is no file on repo.
					// Create a minimal one.
					objAgentVersionsFile = {
						latestVersion: strVersion
					};
				}
				objAgentVersionsFile[strVersion] = {};

				for (let strOSName of arrOSList)
				{
					objAgentVersionsFile[strVersion][strOSName] = "/" + strVersion + "/" + strOSName + "/agentVersions.json";
				}

				const strAgentVersionsFile = JSON.stringify(objAgentVersionsFile, undefined, "\t");
				await fs.writeFile(path.join(strOutputPackageFolderPath, "agentVersions.json"), strAgentVersionsFile);

				console.log("Uploading file(s) to repo: " + strOutputRepoURL);

				await this.uploadDirectoryToRepo(strOutputRepoURL, strOutputRepoLocalFolder, strOutputPackageFolderPath);
			}
		}

		/**
		 * @param {string} strFileName
		 *
		 * @returns {string}
		 */
		_generateTmpPath(strFileName)
		{
			return path.join(os.tmpdir(), `${Date.now()}-${strFileName}`);
		}
	};
