const sqlite = require("sql.js");
const fs = require("fs-promise");
const path = require("path");
const pathExists = require("path-exists");
const assert = require("assert");

module.exports =
class Logger 
{
	constructor()
	{
		this.db = null;
		this._tableName = null;
		this._databaseName = null;
		this._strDatabasePath = null;
		this._nMaxDBSize = 10;
	}


	/**
	 * @param {string} strApplicationName
	 * @returns {undefined}
	 */
	async _initLoggingDirectory(strApplicationName)
	{
		let strTempDirectoryPath;

		if(process.platform === "win32")
		{
			assert(process.env.TEMP, "TEMP environment variable not set.");
			assert((await pathExists(process.env.TEMP)), "TEMP=" + JSON.stringify(process.env.TEMP) + " path not found.");

			strTempDirectoryPath = process.env.TEMP;
		}
		else if (["freebsd", "sunos", "linux", "darwin"].indexOf(process.platform) > -1)
		{
			assert(await pathExists("/var/log"), "/var/log not found. Current platform: " + JSON.stringify(process.platform) + ".");
			strTempDirectoryPath = "/var/log";
		}
		else
		{
			throw new Error("Unhandled operating system");
		}

		this._strLoggingDirectoryPath = path.join(strTempDirectoryPath, strApplicationName + "Log");

		if(!(await pathExists(this.loggingDirectoryPath)))
		{
			await fs.mkdir(this.loggingDirectoryPath, parseInt("0666", 8));
		}

		assert(await pathExists(this.loggingDirectoryPath));
	}


	/**
	 * Initializes the database and creates the table 
	 *
	 * @param {string} strApplicationName
	 * @param {string} strDatabaseName
	 * @param {string} strTableName
	 * @param {number|undefined} nMaxDBSize
	 * @returns {undefined}
	 */
	async init(strApplicationName, strDatabaseName, strTableName, nMaxDBSize)
	{
		this._tableName = strTableName;
		this._databaseName = strDatabaseName;
		if(nMaxDBSize !== undefined)
		{
			this._nMaxDBSize = nMaxDBSize;
		}

		await this._initLoggingDirectory(strApplicationName);
		await this._createDatabase();
		await this._createTable();
	}


	/**
	 * @returns {undefined}
	 */
	async _createDatabase()
	{
		if(!(await pathExists(this.databasePath)))
		{
			await fs.createWriteStream(this.databasePath);
		}
		else
		{
			let fileSize = (await fs.stat(this.databasePath)).size;
			if(fileSize > this.nMaxDBSize * 1000 * 1000)
			{
				await fs.unlink(this.databasePath);
				await fs.createWriteStream(this.databasePath);
			}
		}

		this.db = new sqlite.Database(await fs.readFile(this.databasePath));
	}


	/**
	 * @returns {undefined}
	 */
	async _createTable()
	{
		let strSqlQuery = `
			CREATE TABLE IF NOT EXISTS ` + Logger.escapeIdentifier(this.tableName) + ` (
				call_id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
				call_created_timestamp CHAR(20) DEFAULT '0000-00-00T00:00:00Z' NOT NULL,
				call_updated_timestamp CHAR(20) DEFAULT '0000-00-00T00:00:00Z' NOT NULL,
				call_client_ip_address CHAR(20) DEFAULT NULL,
				call_is_error BOOLEAN NOT NULL, 
				call_function_name VARCHAR(255) DEFAULT NULL,
				call_params_json TEXT DEFAULT NULL,
				call_exception_json TEXT DEFAULT NULL,
				call_result_json TEXT DEFAULT NULL,
				call_repeat_count INTEGER(10) DEFAULT NULL
			)`;

		await this.db.run(strSqlQuery);
	}


	/**
	 * @param {Error} error
	 * @param {string} strDate
	 * @returns {undefined}
	 */
	async _insertInTable(error, strDate)
	{
		assert(error instanceof Error, "Parameter \"error\" must be a subclass of Error");
		
		let error_json = JSON.stringify(
			{
				exception_type: error.constructor.name,
				message: error.message,
				trace: error.stack,
				code: error.code,
				errno: error.errno
			});

		let strSqlQuery = `
			INSERT INTO ` + Logger.escapeIdentifier(this.tableName) + `
			(
				call_created_timestamp,
				call_updated_timestamp,
				call_is_error,
				call_exception_json,
				call_repeat_count
			)
			VALUES
			(
				:strDate,
				:strDate,
				1,
				:error_json,
				1
			)`;
		
		await this.db.prepare(strSqlQuery).run({
			":error_json": error_json, 
			":strDate": strDate
		});
	}


	/**
	 * Check if it is a recent error. If so, just update the call_error_count, else insert a new error in database.
	 * 
	 * @param {Error} error
	 * @returns {undefined}
	 */
	async logError(error)
	{
		assert(error instanceof Error, "Parameter \"error\" must be a subclass of Error");
		
		let dateCurrent = new Date();
		
		let strCurrentDate = dateCurrent.toISOString().replace(/\.[0-9]*/, "");
		let strLastDayDate = new Date(dateCurrent.getTime() + (1000 * 3600 * -24 /* hours */)).toISOString().replace(/\.[0-9]*/, "");
		
		let error_json = JSON.stringify(
			{
				exception_type: error.constructor.name,
				message: error.message,
				trace: error.stack,
				code: error.code,
				errno: error.errno
			});

		let strSqlQuery = `
			SELECT 
				* 
			FROM ` + Logger.escapeIdentifier(this.tableName) + `
			WHERE 
				call_exception_json = :error_json
				AND
				call_created_timestamp BETWEEN :strLastDayDate AND :strCurrentDate;
			`;

		let statement = await this.db.prepare(strSqlQuery);
		statement.bind({":error_json": error_json, 
			":strLastDayDate": strLastDayDate, 
			":strCurrentDate": strCurrentDate});
		
		let bHasEntry = await statement.step();

		if(!bHasEntry) // There are no errors of that type
		{	
			await this._insertInTable(error, strCurrentDate);
		}
		else // Update the value of repeat_count
		{
			let result = await statement.getAsObject();
			
			// Retrieve call_repeat_count
			let call_repeat_count_value = result["call_repeat_count"];

			call_repeat_count_value++;
			
			strSqlQuery = `
				UPDATE ` + Logger.escapeIdentifier(this.tableName) + `
				SET
					call_repeat_count = :call_repeat_count_value,
					call_updated_timestamp = :strCurrentDate
				WHERE 
					call_exception_json = :error_json
					AND 
					call_created_timestamp BETWEEN :strLastDayDate AND :strCurrentDate
				`;
			
			await this.db.prepare(strSqlQuery).run({
				":call_repeat_count_value": call_repeat_count_value, 
				":strCurrentDate": strCurrentDate, 
				":error_json": error_json, 
				":strLastDayDate": strLastDayDate
			});
		}

		statement.free();
	}


	/**
	 * Makes data persistent.
	 * 
	 * @returns {undefined}
	 */
	async writeDataOnDisk()
	{
		await fs.writeFile(this.databasePath, new Buffer(await this.db.export()));
	}

	/**
	 * @returns {undefined}
	 */

	async logErrorPersistent(error)
	{
		await this.logError(error);
		await this.writeDataOnDisk();
	}

	/**
	 * @param {string} strFromDate
	 * @param {string} strToDate
	 * @returns {Array}
	 */
	async getDataByDateString(strFromDate, strToDate)
	{
		assert(typeof strFromDate === "string", "Parameter \"strFromDate\" must be of type string");
		assert(typeof strToDate === "string", "Parameter \"strToDate\" must be of type string");
		
		let strSqlQuery = `
			SELECT 
				* 
			FROM ` + Logger.escapeIdentifier(this.tableName) + ` 
			WHERE 
				call_created_timestamp BETWEEN :strFromDate AND :strToDate
		`;
		
		let statement = await this.db.prepare(strSqlQuery);
		statement.bind({
			":strFromDate": strFromDate, 
			":strToDate": strToDate
		});

		let arrResult = [];
		
		while(statement.step())
		{
			let objRow = await statement.getAsObject();
			arrResult.push(objRow);
		}
		statement.free();

		return arrResult;
	}


	/**
	 * @param {Date} dFromDate
	 * @param {Date} dToDate
	 * @returns {Array} 
	 */
	async getDataByDateObject(dFromDate, dToDate)
	{
		assert(dFromDate instanceof Date, "Parameter \"dFromDate\" must be of type Date");
		assert(dToDate instanceof Date, "Parameter \"dToDate\" must be of type Date");

		let strFromDate = new Date(dFromDate).toISOString().replace(/\.[0-9]*/, "");
		let strToDate = new Date(dToDate).toISOString().replace(/\.[0-9]*/, "");

		return this.getDataByDateString(strFromDate, strToDate);
	}


	/**
	 * @param {string} strExceptionMessage
	 * @returns {Array}
	 */
	async getDataByErrorMessage(strExceptionMessage)
	{
		assert(typeof strExceptionMessage === "string", "Parameter \'strExceptionMessage\' must be of type string");

		let statement = await this.db.prepare("SELECT * FROM " + Logger.escapeIdentifier(this.tableName));

		let arrResult = [];
		while(statement.step())
		{
			let objRow = await statement.getAsObject();
			
			let call_exception_json = JSON.parse(objRow["call_exception_json"]);

			if(call_exception_json["message"] === strExceptionMessage)
			{
				arrResult.push(objRow);
			}
		}

		return arrResult;
	}


	/**
	 * @param {string} strExceptionType
	 * @returns {Array}
	 */
	async getDataByErrorType(strExceptionType)
	{
		assert(typeof strExceptionType === "string", "Parameter \'strExceptionType\' must be of type string");

		let statement = await this.db.prepare("SELECT * FROM " + Logger.escapeIdentifier(this.tableName));

		let arrResult = [];
		while(statement.step())
		{
			let objRow = await statement.getAsObject();
			
			let call_exception_json = JSON.parse(objRow["call_exception_json"]);

			if(call_exception_json["exception_type"] === strExceptionType)
			{
				arrResult.push(objRow);
			}
		}

		return arrResult;
	}


	/**
	 * @returns {undefined}
	 */
	async deleteAllRecords()
	{
		await this.db.exec("DELETE FROM " + Logger.escapeIdentifier(this.tableName));
		
		// Reset PRIMARY KEY
		await this.db.run(`
			UPDATE SQLITE_SEQUENCE 
			SET 
				SEQ = 0
			WHERE 
				NAME = "` + Logger.escapeIdentifier(this.tableName) + `";
		`); 

		await this.writeDataOnDisk();
	}


	/**
	 * @returns {string}
	 */
	get loggingDirectoryPath()
	{
		if(this._strLoggingDirectoryPath === null)
		{
			throw new Error("You must call init before using this property.");
		}

		return this._strLoggingDirectoryPath;
	}


	/**
	 * @returns {string}
	 */
	get databasePath()
	{
		return path.join(this.loggingDirectoryPath, this.databaseName);
	}


	/**
	 * @returns {string}
	 */
	get tableName()
	{
		return this._tableName;
	}


	/**
	 * @returns {string}
	 */
	get databaseName()
	{
		return this._databaseName;
	}


	/**
	 * 	@returns {number}
	 */
	get nMaxDBSize()
	{
		return this._nMaxDBSize;
	}

	/**
	 * Escape function for table name
	 * 
	 * @param {string} strTableName
	 * @returns {string}
	 */
	static escapeIdentifier(strTableName)
	{
		return strTableName.replace(/[^a-zA-Z0-9_\-]+/g, "");
	}
};
