"use strict";

const EmailAddressValidatorException = require("./EmailAddressValidatorException");
const DomainNameValidator = require("../DomainName/DomainNameValidator");
const DomainNameValidatorException = require("../DomainName/DomainNameValidatorException");
const BaseValidator = require("../Base/BaseValidator");

module.exports = 
class EmailAddressValidator extends BaseValidator
{
	constructor()
	{
		throw new Error("EmailAddressValidator is a static class and cannot be instantiated.");
	}
	
	/**
	 * 
	 * @param {string} strEmailAddress 
	 * 
	 * @throws {EmailAddressValidatorException | DomainNameValidatorException}
	 */
	static assertValid(strEmailAddress)
	{
		if(strEmailAddress === undefined || strEmailAddress === null || (typeof strEmailAddress === "string" && strEmailAddress.trim().length === 0))
		{
			throw new EmailAddressValidatorException("Empty email address.");
		}

		if(typeof strEmailAddress !== "string")
		{
			throw new EmailAddressValidatorException(`Invalid value type "${typeof strEmailAddress}" for email address. Only accepts non-empty "string".`);
		}

		let arrParts = strEmailAddress.split("@");
		
		if(arrParts.length !== 2)
		{
			throw new EmailAddressValidatorException("The @ separator is specified multiple times or is missing.");
		}
		
		let strLocalPart = arrParts[0];
		let strDomainPart = arrParts[1];
		
		try
		{
			DomainNameValidator.assertValid(strDomainPart);
		}
		catch(error)
		{
			if(error instanceof DomainNameValidatorException)
			{
				error.message = "Invalid email. " + error.message;
			}

			throw error;
		}
		
		this.assertValidLocalPart(strLocalPart);

		if(strEmailAddress.length > 254)
		{
			throw new EmailAddressValidatorException(
				"The maximum length of 254 characters of an email address exceeded.", 
				EmailAddressValidatorException.MAXIMUM_254_CHARS_EXCEEDED
			);
		}
		
		return true;
	}
	
	/**
	* UTF8 characters and special characters in the local part ("(),:;<>@[\]) are not supported (characters which must be quoted or escaped)
	* by this function.
	*/
	static assertValidLocalPart(strLocalPart)
	{
		if(typeof strLocalPart !== "string" || strLocalPart.trim().length === 0)
		{
			throw new EmailAddressValidatorException(`Invalid value type "${typeof strLocalPart}" for email local part. Only accepts non-empty "string".`);
		}

		if(strLocalPart.indexOf("..") >- 1)
		{
			throw new EmailAddressValidatorException(
				"The . (dot) character appears two or more times consecutively.", 
				EmailAddressValidatorException.DOT_CHARACTER_APPEARS_TWICE_CONSECUTIVELY
			);
		}

		if(strLocalPart.substr(0, 1) == "." || strLocalPart.substr(strLocalPart.length-1) == ".")
		{
			throw new EmailAddressValidatorException("The local part cannot start or end with a . (dot) character.", EmailAddressValidatorException.LOCAL_PART_STARTS_OR_ENDS_WITH_DOT_CHARACTER);
		}
		
		
		let strSpecialCharacters="\"\\(\\),\\:;\\<\\>@\\[\\\\\\]"; //php: json_encode(preg_quote("\"(),:;<>@[\]", "/"));
		if(new RegExp("["+strSpecialCharacters+"]").test(strLocalPart))
		{
			throw new EmailAddressValidatorException(
				"The local part contains valid special (quotable) characters \"(),:;<>@[\] or email address comments, yet they are not supported to be validated by this function.",
				EmailAddressValidatorException.SPECIAL_CHARACTERS_AND_COMMENTS_ARE_UNSUPPORTED
			);
		}
		
		if(strLocalPart.length > 64)
		{
			throw new EmailAddressValidatorException("The local part maximum length of 64 characters exceeded.", EmailAddressValidatorException.LOCAL_PART_MAXIMUM_64_CHARS_EXCEEDED);
		}

		if(strLocalPart.length < 1)
		{
			throw new EmailAddressValidatorException("The local part cannot be empty (it must have at least 1 character).", EmailAddressValidatorException.LOCAL_PART_IS_EMPTY);
		}
		
		
		let strAllowedCharacters = "\\.\\!#\\$%&'\\*\\+\\\/\\=\\?\\^`\\{\\|\\}~_-"; //php: json_encode(preg_quote(".!#$%&'*+/=?^`{|}~_-", "/"));
		if(!(new RegExp("^[a-zA-Z0-9"+strAllowedCharacters+"]{1,64}$").test(strLocalPart)))
		{
			throw new EmailAddressValidatorException("The local part contains not allowed characters.", EmailAddressValidatorException.DISALLOWED_CHARACTERS);
		}
		
		return true;
	}

};
