const assert = require("assert");
const sleep = require("sleep-promise");
const JSONRPC = require("jsonrpc-bidirectional");
const WebSocket = require("ws");
const fs = require("fs-extra");
const os = require("os");
const path = require("path");
const json = require("json-promise");
const url = require("url");
const ChildProcess = require("child-process-promise");
const tcpPing = require("tcp-ping");
const CLITable = require("cli-table");

const {Encryption} = require("encryption");

const AgentIdentity = require("./Identity");

module.exports =
class BSIWebSocketClientManager
{
	/**
	 * @param {Identity} agentIdentity
	 * @param {JSONRPC.EndpointBase} endpoint
	 * @param {FactsProvider} factsProvider
	 */
	constructor(agentIdentity, endpoint, factsProvider)
	{
		this._endpoint = endpoint;
		this._factsProvider = factsProvider;

		this._jsonrpcServer = new JSONRPC.Server();
		this._jsonrpcServer.registerEndpoint(this._endpoint);

		this._jsonrpcServer.addPlugin(new JSONRPC.Plugins.Server.AuthenticationSkip());
		this._jsonrpcServer.addPlugin(new JSONRPC.Plugins.Server.AuthorizeAll());

		this._wsJSONRPCRouter = new JSONRPC.BidirectionalWebsocketRouter(this._jsonrpcServer);

		assert(agentIdentity instanceof AgentIdentity);
		this._agentIdentity = agentIdentity;

		this._objEndpointURLToTCPPingAverageMilliseconds = {};

		this._webSocket = null;

		this._promiseJSONRPCClient = null;

		this._strServerUUID = null;
	}


	/**
	 * @returns {Identity}
	 */
	get agentIdentity()
	{
		return this._agentIdentity;
	}


	/**
	 * Usage:
	 * const fnClient = await bsiWebSocketClientManager.bsiClientSingletonFunction();
	 * await (await fnClient()).divide(3, 2);
	 *
	 * @returns {Function}
	 */
	async bsiClientSingletonFunction()
	{
		if(
			this._webSocket
			&& this._webSocket.readyState === WebSocket.CLOSED
		)
		{
			assert(this._promiseJSONRPCClient);

			this._webSocket = null;
			this._promiseJSONRPCClient = null;
		}

		if(!this._promiseJSONRPCClient)
		{
			this._promiseJSONRPCClient = new Promise(async(fnResolve, fnReject) =>
			{
				try
				{
					const nSleepMillisecondsBetweenRetries = 4000;
					const nConnectRetriesMax = 200;
					let nConnectTries = 0;

					let bsiClientWebsocket = null;


					let endpointURLIterator = this.makeEndpointURLIterator();

					let strEndpointURL = this._agentIdentity.endpointURL;
					if(
						!strEndpointURL
						|| url.parse(strEndpointURL).path !== url.parse(this._endpoint.path).path
					)
					{
						do
						{
							await this._refreshEndpointURLsTCPPingAverage();
							strEndpointURL = endpointURLIterator.next().value;

							if(strEndpointURL === null)
							{
								await sleep(5000);
							}
						} while(strEndpointURL === null && ++nConnectTries <= nConnectRetriesMax);
					}

					// Helping this process be long running!
					while(++nConnectTries <= nConnectRetriesMax)
					{
						console.log(
							`Attempting to connect to BSIWebSocketServer ${strEndpointURL}, `
							+ `try no.: ${nConnectTries}. Max retries is ${nConnectRetriesMax}, at ${nSleepMillisecondsBetweenRetries} milliseconds intervals.`
						);

						try
						{
							const objHeaders = {};

							if(
								this._agentIdentity.environment !== "production"
								&& this._agentIdentity.ipAddressSpoof
							)
							{
								objHeaders["AGENT-DEVELOPMENT-IP-SPOOF"] = this._agentIdentity.ipAddressSpoof;
							}

							const webSocket = new WebSocket(
								strEndpointURL,
								/* protocols */ undefined,
								/* options */ {
									"headers": objHeaders
								}
							);

							try
							{
								if(strEndpointURL === null)
								{
									throw new Error("Failed to find a connectable WebSocket URI.");
								}

								await new Promise((_fnResolve, _fnReject) =>
								{
									webSocket.on("open", _fnResolve);
									webSocket.on("error", _fnReject);
								});

								const nWebSocketConnectionID = await this._wsJSONRPCRouter.addWebSocketSync(webSocket);

								bsiClientWebsocket = this._wsJSONRPCRouter.connectionIDToSingletonClient(nWebSocketConnectionID, this._endpoint.ReverseCallsClientClass);

								let objAgentRegisterResponse;
								if(bsiClientWebsocket.agent_register)
								{
									const strChallenge = await bsiClientWebsocket.auth_challenge();
									assert(strChallenge.trim().length >= 24, "Was expecting a longer challenge with no whitespace.");


									const encryption = new Encryption(this._agentIdentity.instanceEncryptionRotatingKeys);
									
									
									// hashing is faster.
									// const strChallengeResponse = await encryption.encrypt(strChallenge); 
									const strChallengeResponse = await encryption.hash(strChallenge); 


									objAgentRegisterResponse = await bsiClientWebsocket.agent_register(
										await this.agentVersion(),
										this._agentIdentity.agentID,
										strChallengeResponse,
										await this.serverObject()
									);


									// hashing is faster.
									//const strChallengeResponseMutual = (await encryption.decrypt(objAgentRegisterResponse.challenge_response_mutual)).toString("ascii");
									/*if(strChallengeResponseMutual !== strChallengeResponse)
									{
										throw new Error(`Mutual challenge response from server did not encrypt the same way. Server failed mutual authentication. ${JSON.stringify(strChallengeResponseMutual)} !== ${JSON.stringify(strChallengeResponse)}`);
									}
									else if(objAgentRegisterResponse.challenge_response_mutual === strChallengeResponse)
									{
										throw new Error(`Mutual challenge response from server is the same challenge response sent to server (same IV, means same message sent back). Server failed mutual authentication. ${JSON.stringify(objAgentRegisterResponse.challenge_response_mutual)} === ${JSON.stringify(strChallengeResponse)}`);
									}*/
									

									const objHashMetadata = encryption.getMetadataFromHashData(strChallengeResponse);
									if(objHashMetadata.hashAlgorithm !== "hmac-sha256")
									{
										throw new JSONRPC.Exception("Allowed hashing algorithms: hmac-sha256");
									}
									const strChallengeResponseMutual = await encryption.hash(strChallengeResponse, objHashMetadata.hashAlgorithm, objHashMetadata.keyIndex);
									if(strChallengeResponseMutual !== objAgentRegisterResponse.challenge_response_mutual)
									{
										throw new Error(`Mutual challenge response from server did not hash the same way. Server failed mutual authentication. ${JSON.stringify(strChallengeResponseMutual)} !== ${JSON.stringify(objAgentRegisterResponse.challenge_response_mutual)}`);
									}
								}
								else if(bsiClientWebsocket.bdk_agent_register)
								{
									const strChallenge = await bsiClientWebsocket.auth_challenge();
									assert(strChallenge.trim().length >= 24, "Was expecting a longer challenge with no whitespace.");

									const encryption = new Encryption(this._agentIdentity.serverEncryptionRotatingKeys);
									const strChallengeResponse = await encryption.hash(strChallenge);

									objAgentRegisterResponse = await bsiClientWebsocket.bdk_agent_register(
										await this.agentVersion(),
										await this.serverObject(),
										strChallengeResponse,
										this._agentIdentity.agentID
									);

									const objHashMetadata = encryption.getMetadataFromHashData(strChallengeResponse);
									if(objHashMetadata.hashAlgorithm !== "hmac-sha256")
									{
										throw new JSONRPC.Exception("Allowed hashing algorithms: hmac-sha256");
									}

									const strChallengeResponseMutual = await encryption.hash(strChallengeResponse, objHashMetadata.hashAlgorithm, objHashMetadata.keyIndex);
									if(strChallengeResponseMutual !== objAgentRegisterResponse.challenge_response_mutual)
									{
										throw new Error(`Mutual challenge response from server did not hash the same way. Server failed mutual authentication. ${JSON.stringify(strChallengeResponseMutual)} !== ${JSON.stringify(objAgentRegisterResponse.challenge_response_mutual)}`);
									}
								}
								else
								{
									throw new Error("Unhandled agent register scenario.");
								}

								this._agentIdentity.agentID = parseInt(objAgentRegisterResponse.agent_id, 10);
								assert(this._agentIdentity.environment === null || this._agentIdentity.environment === objAgentRegisterResponse.environment);
								this._agentIdentity.environment = objAgentRegisterResponse.environment;
								this._agentIdentity.endpointURL = strEndpointURL;
								await this._agentIdentity.saveSettings();

								console.log("agent_id: " + objAgentRegisterResponse.agent_id);
								console.log("environment: " + objAgentRegisterResponse.environment);

								// Reset because this._agentIdentity.environment may have changed and because first URLs are prefered on a retry.
								endpointURLIterator = this.makeEndpointURLIterator();


								this._webSocket = webSocket;
							}
							catch(error)
							{
								try
								{
									webSocket.close(1011 /*Internal error*/, "BSIAgent internal error: " + error.message + ". Stack: " + error.stack);
								}
								catch(ignoredError)
								{
								}

								throw error;
							}

							break;
						}
						catch(error)
						{
							await this._refreshEndpointURLsTCPPingAverage();
							strEndpointURL = endpointURLIterator.next().value;


							if(
								error instanceof TypeError
								|| error instanceof SyntaxError
								|| error instanceof EvalError
								|| error instanceof ReferenceError
								|| error instanceof RangeError
								|| error instanceof URIError
							)
							{
								console.log(error);
								throw error;
							}

							console.error(error);

							if(nConnectTries < nConnectRetriesMax)
							{
								await sleep(nSleepMillisecondsBetweenRetries);
							}
							else
							{
								throw error;
							}
						}
					}

					if(!this._webSocket || this._webSocket.readyState === WebSocket.CLOSED)
					{
						throw new Error("Failed to find a connectable WebSocket URI or WebSocket was closed right after authentication (agent_register() or equivalent).");
					}

					console.log("Connected to BSIWebSocketServer " + strEndpointURL);

					this._webSocket.on(
						"error",
						async(error) =>
						{
							this._promiseJSONRPCClient = null;

							console.error("BSI error: ");
							console.error(error);

							try
							{
								console.error("Cleanup. Attempting to close the BSI WebSocket connection.");

								await this._webSocket.end();
								//this._webSocket.terminate();
							}
							catch(error)
							{
								console.error(error);
							}

							this._webSocket = null;
						}
					);

					fnResolve(bsiClientWebsocket);
				}
				catch(error)
				{
					this._promiseJSONRPCClient = null;
					this._webSocket = null;

					fnReject(error);
				}
			});
		}

		return async() =>
		{
			if(
				this._promiseJSONRPCClient === null
				|| (
					this._webSocket
					&& this._webSocket.readyState !== WebSocket.OPEN
				)
			)
			{
				console.error("WebSocket state " + JSON.stringify(this._webSocket.readyState) + ". Waiting 2 seconds and will retry.");
				await sleep(2000);

				return (await this.bsiClientSingletonFunction())();
			}

			//return this._webSocket && this._webSocket.connection.state === "connected" ? this._webSocket : this._promiseJSONRPCClient;
			return this._promiseJSONRPCClient;
		};
	}

	/**
	 * Returns the version from the package.json.
	 *
	 * @returns {string}
	 */
	async agentVersion()
	{
		return new Promise(
			async(fnResolve, fnReject) =>
			{
				const strAgentRootFolder = path.dirname(path.dirname(path.dirname(__filename)));
				const strPackageJSONPath = path.join(strAgentRootFolder, "package.json");
				const packageContent = await fs.readFile(strPackageJSONPath, {encoding: "utf8"});
				const parsedObject = await json.parse(packageContent);
				const version = parsedObject["version"];
				if (version != null)
				{
					fnResolve(version);
				}
				fnReject(new Error("Cannot find the version in package json"));
			}
		);
	}


	/**
	 * @returns {Object}
	 */
	async serverObject()
	{
		const objServer = {
			//server_id: null,
			server_uuid: await this.serverUUID(),
			//server_network_total_capacity_mbps: null,
			//server_power_status: "on",
			//server_processor_core_count: 1,
			//server_processor_core_mhz: 1000,
			//server_processor_count: 1,
			//server_ram_gbytes: 1,
			//server_disk_count: 0,
			//server_disk_size_mbytes: 512,
			//server_disk_type: "none",
			//server_processor_name: "CPU name",
			//server_product_name: "Dummy server?",
			//server_type_id: null,
			server_interfaces: [],
			//server_disks: [],
			type: "Server"
		};

		for(let strMACAddress of await this.networkInterfacesMACAddresses())
		{
			objServer["server_interfaces"].push({server_interface_mac_address: strMACAddress});
		}

		return objServer;
	}


	/**
	 * Returns the MAC addresses of interfaces with an associated IP address.
	 *
	 * @returns {string[]}
	 */
	async networkInterfacesMACAddresses()
	{
		if(this._agentIdentity.macAddressSpoof)
		{
			return [this._agentIdentity.macAddressSpoof];
		}

		const arrMACAddresses = [];
		const arrInterfaces = await this._factsProvider.getInterfaces();

		for(let networkInterface of arrInterfaces)
		{
			if(
				networkInterface.macAddress !== "00:00:00:00:00:00"
				&& !arrMACAddresses.includes(networkInterface.macAddress)
			)
			{
				arrMACAddresses.push(networkInterface.macAddress);
			}
		}

		return arrMACAddresses;
	}


	async serverUUID()
	{
		if(this._strServerUUID !== null)
		{
			return this._strServerUUID;
		}

		if(this._agentIdentity.serverUUIDSpoof)
		{
			this._strServerUUID = this._agentIdentity.serverUUIDSpoof;
			return this._agentIdentity.serverUUIDSpoof;
		}

		let strCommand = "dmidecode | grep -i UUID"; // Linux, BSD and others.

		if(os.platform() === "win32")
		{
			strCommand = "wmic csproduct get \"UUID\"";
		}
		else if(os.platform() === "darwin")
		{
			// @TODO Not tested:
			strCommand = "system_profiler SPHardwareDataType | awk '/UUID/ { print $3; }'";
		}

		const strCommandOutput = (await ChildProcess.exec(strCommand, {maxBuffer: 50 * 1024 * 1024})).stdout;

		const arrMatches = strCommandOutput.match(/[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}/i);
		if(arrMatches === null)
		{
			throw new Error("Regex didn't match an UUID. " + strCommand + ": " + strCommandOutput);
		}

		this._strServerUUID = arrMatches[0];
		return arrMatches[0];
	}


	/**
	 * @returns {undefined}
	 */
	async _refreshEndpointURLsTCPPingAverage()
	{
		const arrPromises = [];

		for(let strEnvironment in BSIWebSocketClientManager.knownJSONRPCEndpointRoots)
		{
			if(
				this._agentIdentity.environment !== null
				&& strEnvironment !== this._agentIdentity.environment
			)
			{
				continue;
			}

			for(let strEndpointURLRoot of BSIWebSocketClientManager.knownJSONRPCEndpointRoots[strEnvironment])
			{
				if(!strEndpointURLRoot)
				{
					continue;
				}

				arrPromises.push(new Promise((fnResolve, fnReject) => 
				{
					const objURL = url.parse(strEndpointURLRoot);

					tcpPing.ping(
						{
							address: objURL.hostname,
							port: parseInt(objURL.port ? objURL.port : (["wss:", "https:"].includes(objURL.protocol) ? 443 : 80), 10),
							timeout: 5000 /*milliseconds*/,
							attempts: 2
						},
						(error, objResults) => 
						{
							if(error || /*library normally has NaN*/ Number.isNaN(objResults.avg) || objResults.avg === undefined || objResults.avg === null)
							{
								if(error)
								{
									console.error("tcp-ping", strEndpointURLRoot, error);
								}

								this._objEndpointURLToTCPPingAverageMilliseconds[strEndpointURLRoot] = null;
							}
							else
							{
								this._objEndpointURLToTCPPingAverageMilliseconds[strEndpointURLRoot] = objResults.avg;
							}

							fnResolve();
						}
					);
				}));
			}
		}

		await Promise.all(arrPromises);
		const table = new CLITable({head: ["Endpoint URL", "TCP ping average latency"], chars: {"mid": "", "left-mid": "", "mid-mid": "", "right-mid": ""}});
		for(let strEndpoint in this._objEndpointURLToTCPPingAverageMilliseconds)
		{
			const strLatency = this._objEndpointURLToTCPPingAverageMilliseconds[strEndpoint] === null ? "timed out" : String(Math.round(this._objEndpointURLToTCPPingAverageMilliseconds[strEndpoint], 2)) + " milliseconds";
			table.push([strEndpoint, String(strLatency)]);
		}
		console.log(table.toString());
	}


	/**
	 * @returns {iterator<string>}
	 */
	* makeEndpointURLIterator()
	{
		let bInfiniteLoop = false;

		while(true)
		{
			if(bInfiniteLoop)
			{
				console.error("No connectable endpoints to iterator on. Sleep for a few seconds then try again.");
				yield null;
			}

			bInfiniteLoop = true;

			for(let strEnvironment in BSIWebSocketClientManager.knownJSONRPCEndpointRoots)
			{
				if(
					this._agentIdentity.environment !== null
					&& strEnvironment !== this._agentIdentity.environment
				)
				{
					continue;
				}

				for(let strEndpointURLRoot of BSIWebSocketClientManager.knownJSONRPCEndpointRoots[strEnvironment])
				{
					// Not connectable
					if(this._objEndpointURLToTCPPingAverageMilliseconds[strEndpointURLRoot] === null)
					{
						continue;
					}
					
					bInfiniteLoop = false;

					assert(this._endpoint.path[0] === "/");
					yield strEndpointURLRoot + this._endpoint.path;
				}
			}
		}
	}


	/**
	 * Unique anycast IP.
	 *
	 * @returns {string}
	 */
	static get anycastIPAddress()
	{
		return "172.23.255.254";
	}


	/**
	 * Known BSI hostnames, in preferred order.
	 *
	 * @returns {Object.<string, string[]>}
	 */
	static get knownJSONRPCEndpointRoots()
	{
		return {
			"development": [
				"ws://bsi.vmware.local",
				"ws://192.168.137.2"
			],

			"integration": [
				// "wss://" + BSIWebSocketClientManager.anycastIPAddress + "",
				// "ws://" + BSIWebSocketClientManager.anycastIPAddress + "",

				"wss://lab-master-bucharest-bsiintegration.bigstepcloud.com",
				"wss://lab-slave-bucharest-bsiintegration.bigstepcloud.com"//,

				// "wss://89.36.24.19", // lab-master-bucharest-bsiintegration.bigstepcloud.com
				// "wss://10.255.236.11" // lab-slave-bucharest-bsiintegration.bigstepcloud.com
			],

			"production": [
				"wss://" + BSIWebSocketClientManager.anycastIPAddress + "",
				"ws://" + BSIWebSocketClientManager.anycastIPAddress + "",

				"wss://fullmetal.bigstep.com",
				//"wss://84.40.58.36",

				"wss://uk-reading-fullmetal.bigstep.com",
				//"wss://84.40.58.36",

				"wss://us-chicago-fullmetal.bigstep.com",
				//"wss://176.223.248.36",

				"wss://de-frankfurt-fullmetal.bigstep.com"
				//"wss://185.90.48.36"
			]
		};
	}

	/**
	 * Known BSI repos
	 *
	 * @returns {Object.<string, string[]>}
	 */
	static get knownRepositoryEndpointRoots()
	{
		return {
			"development": [
				"192.168.137.3"
			],

			"integration": [
				"192.168.137.3"
			],

			"production": [
				"192.168.137.3"
			]
		};
	}
};
