const JSONRPC = require("jsonrpc-bidirectional");

module.exports = 
class AFCMakeCallEndpointBase extends JSONRPC.EndpointBase
{
	/**
	 * @param {string} strName
	 * @param {string} strPath
	 * @param {Object} objReflection
	 * @param {Class|null} classReverseCallsClient
	 */
	constructor(strName, strPath, objReflection, classReverseCallsClient)
	{
		super(strName, strPath, objReflection, classReverseCallsClient);

		this._objAFCPromises = {};
		this._mapUniqueReconnectableCallsPromises = new Map();
	}


	/**
	 * Proxy function which calls functions on this endpoint. 
	 * 
	 * It reuses promises for the same AFC ID to prevent duplicate calls 
	 * and to allow re-establishing a connection with BSIWebSocketServer's AFC service.
	 * 
	 * Failed promises are removed from the cache and are allowed to be retried.
	 * Successful promises remain in the cache for a while or forever, and are NOT allowed to be retried.
	 * 
	 * @param {JSONRPC.IncomingRequest} incomingRequest
	 * @param {number} nAFCID
	 * @param {string} strFunctionName
	 * @param {Array} arrParams
	 * 
	 * @returns {null}
	 */
	async afc_make_call(incomingRequest, nAFCID, strFunctionName, arrParams)
	{
		if(typeof nAFCID !== "number" || nAFCID <= 0)
		{
			throw new JSONRPC.Exception("nAFCID must be of type number and >= 1.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(typeof strFunctionName !== "string")
		{
			throw new JSONRPC.Exception("strFunctionName must be of type string.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(arrParams.constructor.name !== "Array")
		{
			throw new JSONRPC.Exception("arrParams must be of type Array.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(typeof this[strFunctionName] !== "function")
		{
			throw new JSONRPC.Exception("Method " + JSON.stringify(strFunctionName) + " not found on endpoint " + JSON.stringify(this.path) + ".", JSONRPC.Exception.METHOD_NOT_FOUND);
		}


		// @TODO: read mxResult (associated to nAFCID) from a persistent store. If found, return it.

		if(!this._objAFCPromises.hasOwnProperty(nAFCID))
		{
			this._objAFCPromises[nAFCID] = new Promise(async(fnResolve, fnReject) => 
			{
				try
				{
					const mxResult = await this[strFunctionName].apply(this, [incomingRequest].concat(arrParams));

					// @TODO: write mxResult to a persistent store, associated to nAFCID.

					// Waits only 5 minutes for BSI to ask this promise.
					setTimeout(
						() => 
						{
							if (this._objAFCPromises[nAFCID] != null)
							{
								delete this._objAFCPromises[nAFCID];
							}
						},
						5 * 60 * 1000 // 5 minutes
					);

					fnResolve(mxResult);
				}
				catch(error)
				{
					delete this._objAFCPromises[nAFCID];
					fnReject(error);
				}
			});
		}

		return this._objAFCPromises[nAFCID];
	}


	/**
	 * Proxy function which calls functions on this endpoint. 
	 * 
	 * It reuses promises for the same AFC ID to prevent duplicate calls 
	 * and to allow re-establishing a connection with BSIWebSocketServer's AFC service.
	 * 
	 * Promises remain in the cache for a while or forever, preventing a function retry (the promise is returned instead of invoking the function).
	 * 
	 * It is good practice to have the caller's hostname in the uniqueID in addition to unique numbers and such.
	 * 
	 * @param {JSONRPC.IncomingRequest} incomingRequest
	 * @param {number} strUniqueID
	 * @param {string} strFunctionName
	 * @param {Array} arrParams
	 * @param {null|{agentID: number, workerID: string|number, threadID: string|number, pid: number}} objTargetWorker = null
	 * 
	 * @returns {null}
	 */
	async unique_reconnectable_call(incomingRequest, strUniqueID, strFunctionName, arrParams, objTargetWorker = null)
	{
		if(objTargetWorker === null)
		{
			throw new Error("The agent should not use workers.");
		}

		return await this._unique_reconnectable_call(incomingRequest, strUniqueID, strFunctionName, arrParams, objTargetWorker, /*bKeepErrors*/ true);
	}


	/**
	 * Clears the cached result.
	 * 
	 * Returns true if a promise was found and was cleared.
	 * Returns false otherwise.
	 * 
	 * @param {JSONRPC.IncomingRequest} incomingRequest
	 * @param {number} strUniqueID
	 * 
	 * @returns {boolean}
	 */
	async unique_reconnectable_call_clear(incomingRequest, strUniqueID)
	{
		return this._mapUniqueReconnectableCallsPromises.delete(strUniqueID);
	}


	/**
	 * Proxy function which calls functions on this endpoint. 
	 * 
	 * It reuses promises for the same unique call ID to prevent duplicate calls 
	 * in the event of re-establishing the connection with BSIWebSocketServer.
	 * 
	 * If bKeepErrors is false, promises which threw an error are removed from the cache automatically and are allowed to be retried.
	 * Regardless of bKeepErrors, successful promises remain in the cache for a while or forever, and are NOT allowed to be retried.
	 * 
	 * @param {JSONRPC.IncomingRequest} incomingRequest
	 * @param {string} strUniqueID
	 * @param {string} strFunctionName
	 * @param {Array} arrParams
	 * @param {null|{agentID: number, workerID: string|number, threadID: string|number, pid: number}} objTargetWorker
	 * @param {boolean} bKeepErrors
	 * 
	 * @returns {null}
	 */
	async _unique_reconnectable_call(incomingRequest, strUniqueID, strFunctionName, arrParams, objTargetWorker = null, bKeepErrors = false)
	{
		if(typeof strUniqueID !== "string" || strUniqueID.trim() === "")
		{
			throw new JSONRPC.Exception("strUniqueID must be of type string and not empty.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(typeof strFunctionName !== "string")
		{
			throw new JSONRPC.Exception("strFunctionName must be of type string.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(arrParams.constructor.name !== "Array")
		{
			throw new JSONRPC.Exception("arrParams must be of type Array.", JSONRPC.Exception.INVALID_PARAMS);
		}

		if(typeof this[strFunctionName] !== "function")
		{
			throw new JSONRPC.Exception("Method " + JSON.stringify(strFunctionName) + " not found on endpoint " + JSON.stringify(this.path) + ".", JSONRPC.Exception.METHOD_NOT_FOUND);
		}

		if(typeof bKeepErrors !== "boolean")
		{
			throw new JSONRPC.Exception("bKeepErrors must be of type boolean.", JSONRPC.Exception.INVALID_PARAMS);
		}


		if(!this._mapUniqueReconnectableCallsPromises.has(strUniqueID))
		{
			this._mapUniqueReconnectableCallsPromises.set(
				strUniqueID, 
				new Promise(async(fnResolve, fnReject) => 
				{
					try
					{
						const mxResult = await this[strFunctionName].apply(this, [incomingRequest].concat(arrParams));
	
						// Waits only 5 minutes for BSI to ask this promise.
						setTimeout(
							() => 
							{
								this._mapUniqueReconnectableCallsPromises.delete(strUniqueID);
							},
							5 * 60 * 1000 // 5 minutes
						);
	
						fnResolve(mxResult);
					}
					catch(error)
					{
						this._mapUniqueReconnectableCallsPromises.delete(strUniqueID);
						fnReject(error);
					}
				})
			);
		}

		return this._mapUniqueReconnectableCallsPromises.get(strUniqueID);
	}

	
	/**
	 * Returns the agent ID of this datacenter agent.
	 * 
	 * @param {JSONRPC.IncomingRequest} incomingRequest
	 * 
	 * @returns {{agentID: number, workerID: string|number, threadID: string|number, pid: number}}
	 */
	async worker_elect(incomingRequest)
	{
		if(typeof this._agentIdentity.agentID !== "number")
		{
			throw new JSONRPC.Exception(`agentID is not set yet or is not a number. ${JSON.stringify(this._agentIdentity.agentID)}`);
		}

		return {
			agentID: this._agentIdentity.agentID,
			workerID: "master",
			threadID: "main",
			pid: process.pid
		};
	}
};
