const sleep = require("sleep-promise");

const BSIDeployAgentClient = require("../EndpointClients/BSIDeployAgent");
const AFCMakeCallEndpointBase = require("./AFCMakeCallEndpointBase");
const Plays = require("../Plays");
const Modules = require("../Modules");
const ISCSIDConfiguration = require("../Models/ISCSIDConfiguration");
const ISCSIInitiatorName = require("../Models/ISCSIInitiatorName");
const VLAN = require("../Models/VLAN");
const ISCSIDriveCredentials = require("../Models/ISCSIDriveCredentials");
const FilesystemInformation = require("../Models/FilesystemInformation");
const {Encryption} = require("encryption");

module.exports =
	class BSIDeployAgent extends AFCMakeCallEndpointBase
	{
		constructor(deployAgentUpdater, agentIdentity)
		{
			super(
				/*strName*/ "BSIDeployAgent",
				/*strPath*/ "/api-ws/BSIDeployAgent",
				/*objReflection*/ {},
				/*classReverseCallsClient*/ BSIDeployAgentClient
			);

			this._deployAgentUpdater = deployAgentUpdater;
			this._agentIdentity = agentIdentity;
			this._instanceEncryption = new Encryption(this._agentIdentity.instanceEncryptionRotatingKeys);
			this._objAFCPromises = {};
			this._bUpdatePrerequisiteFinished = false;
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForce = false
		 * @param {number} nTimeoutMilliseconds = 0
		 *
		 * @returns {string}
		 */
		async instance_agent_reboot(incomingRequest, bForce, nTimeoutMilliseconds)
		{
			await Modules.PowerSingleton.powerOff(/*bReboot*/ true, bForce, nTimeoutMilliseconds);

			return "Server will now reboot.";
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {boolean} bForce = false
		 * @param {number} nTimeoutMilliseconds = 0
		 *
		 * @returns {string}
		 */
		async instance_agent_shutdown(incomingRequest, bForce, nTimeoutMilliseconds)
		{
			await Modules.PowerSingleton.powerOff(/*bReboot*/ false, bForce, nTimeoutMilliseconds);

			return "Server will now shutdown.";
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {number} nSleepSeconds
		 *
		 * @returns {string}
		 */
		async instance_agent_test(incomingRequest, nSleepSeconds)
		{
			const strStartTime = new Date().toISOString();
			await sleep(nSleepSeconds * 1000);

			return "Hello! Test start time: " + strStartTime + ". End time: " + (new Date().toISOString());
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 *
		 * @returns {string}
		 */
		async instance_agent_self_reset(incomingRequest)
		{
			await Plays.SelfResetPlaySingleton.run();

			return "Agent data will be deleted";
		}

		/**
		 * @param  {JSONRPC.IncomingRequest} incomingRequest
		 * @param  {string} strUpdateURL
		 *
		 * @returns {string}
		 */
		async instance_agent_update(incomingRequest, strUpdateURL)
		{
			try
			{
				this._bUpdatePrerequisiteFinished = false;
				this._nConfirmUpdateReady = setInterval(
					async() =>
					{
						if (this._bUpdatePrerequisiteFinished === true && Object.keys(this._objAFCPromises).length === 0)
						{
							// Exit to alow update.
							process.exit(0);
						}
					},
					30 * 1000 // thirty seconds.
				);

				await this._deployAgentUpdater.update(strUpdateURL);

				this._bUpdatePrerequisiteFinished = true;

				return "Agent update started.";
			}
			catch (error)
			{
				console.error(error);

				return "Agent could not be updated !";
			}
			finally
			{
				clearInterval(this._nConfirmUpdateReady);
			}
		}


		/**
		 * Exits the current process after 2 seconds using setTimeout().
		 *
		 * Should return before the setTimeout's callback executes. Two seconds should be enough time to return.
		 *
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 */
		async instance_agent_process_exit(incomingRequest)
		{
			setTimeout(
				() =>
				{
					console.log("Was asked to process.exit(1) through the API.");
					process.exit(1);

					return "Will process.exit(1) in 2 seconds.";
				},
				2000
			);
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {bool} bEnabled
		 * @returns {string}
		 */
		async instance_agent_persistent_logging_enable(incomingRequest, bEnabled)
		{
			await Plays.EnablePersistentLoggingPlaySingleton.run({"enabled": bEnabled});

			return "Persistent logging enabled: " + JSON.stringify(bEnabled);
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strBondName
		 * @param {Object} arrVLANObjects
		 * @returns {string}
		 */
		async instance_agent_bonding_configure(incomingRequest, strBondName, arrVLANObjects)
		{
			let arrVLANs = [];
			for (const objVLAN of arrVLANObjects)
			{
				arrVLANs.push(VLAN.createFromJSON(objVLAN));
			}

			const bIsChanged = await Plays.ConfigureNetworkingPlaySingleton.run({
				"bondName": strBondName,
				"VLANs": arrVLANs
			});

			let strResponse = null;
			if (bIsChanged)
			{
				strResponse = "Network bonding configured!";
			}
			else
			{
				strResponse = "Network bonding already configured";
			}

			return strResponse;

		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objVLAN
		 * @returns {string}
		 */
		async instance_agent_vlan_add(incomingRequest, objVLAN)
		{
			const moduleResponse = await Modules.NetworkSingleton.addVLAN(VLAN.createFromJSON(objVLAN));

			let strResponse = null;
			if (moduleResponse.isChanged())
			{
				strResponse = "Vlan added!";
			}
			else
			{
				strResponse = "Vlan already exists!";
			}

			return strResponse;
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strVlanName
		 * @param {string} strBondName
		 * @returns {string}
		 */
		async instance_agent_vlan_remove(incomingRequest, strVlanName, strBondName)
		{
			const moduleResponse = await Modules.NetworkSingleton.removeVLAN(strVlanName, strBondName);

			let strResponse = null;
			if (moduleResponse.isChanged())
			{
				strResponse = "Vlan removed!";
			}
			else
			{
				strResponse = "Vlan already removed!";
			}

			return strResponse;
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strNetworkInterfaceName
		 * @returns {string}
		 */
		async instance_agent_network_interface_enable(incomingRequest, strNetworkInterfaceName)
		{
			const moduleResponse = await Modules.NetworkSingleton.ifup(strNetworkInterfaceName);

			let strResponse = null;
			if (moduleResponse.isChanged())
			{
				strResponse = "NetworkInterface up!";
			}
			else
			{
				strResponse = "NetworkInterface already up!";
			}

			return strResponse;
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strNetworkInterfaceName
		 * @returns {string}
		 */
		async instance_agent_network_interface_disable(incomingRequest, strNetworkInterfaceName)
		{
			const moduleResponse = await Modules.NetworkSingleton.ifdown(strNetworkInterfaceName);

			let strResponse = null;
			if (moduleResponse.isChanged())
			{
				strResponse = "NetworkInterface down!";
			}
			else
			{
				strResponse = "NetworkInterface already down!";
			}

			return strResponse;
		}

		/**
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objNodeInformation
		 * @returns {Object}
		 */
		async instance_agent_iscsi_setup_and_login(incomingRequest, objNodeInformation)
		{
			/**
			 * objNodeInformation contains the node information
			 * It must contain the following parrameters: iscsi_initiator_iqn, isns_address, isns_port, iscsi_authmethod,
			 * iscsi_username, iscsi_password_instanceencrypted, nr_sessions and startup
			 */
			const strDecryptedISCSIPassword = (await this._instanceEncryption.decrypt(objNodeInformation["iscsi_password_instanceencrypted"])).toString("ascii");
			
			const nodeInformation = new ISCSIDConfiguration(
				objNodeInformation["isns_address"],
				objNodeInformation["isns_port"],
				objNodeInformation["iscsi_authmethod"],
				objNodeInformation["iscsi_username"],
				strDecryptedISCSIPassword,
				objNodeInformation["nr_sessions"],
				objNodeInformation["startup"]
			);

			const ISCSIInitiatorInformation = new ISCSIInitiatorName(
				objNodeInformation["iscsi_initiator_iqn"]
			);

			const objParams = {
				"node_information": nodeInformation,
				"iscsi_initiator_name": ISCSIInitiatorInformation
			};

			let objReturnInfo = {};

			try
			{
				const bIsChanged = await Plays.ISCSISetupAndLoginPlaySingleton.run(objParams);

				let strResponse = null;

				if(bIsChanged)
				{
					strResponse = "ISCSI setup and login completed successfully.";
				}
				else
				{
					strResponse = "ISCSI setup and login already completed.";
				}

				objReturnInfo = {
					"message": strResponse,
					"status": "successful",
					"data": null
				};
			}
			catch(error)
			{
				error.message = "ISCSI setup and login failed with: " + error.message;
				throw error;
			}

			return objReturnInfo;
		}

		/**
		 * Partitions a drive with the given credentials. Drive must already exist in /dev/disk/by-path, after the iSCSI login
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objDriveCredentials
		 * @param {Object} objPartitionTable
		 * @returns {Object}
		 */
		async instance_agent_drive_partition(incomingRequest, objDriveCredentials, objPartitionTable)
		{
			//objDriveCredentials must have: drive_id, lun_id, storage_ip_address, storage_port and target_iqn
			const diskCredentials = new ISCSIDriveCredentials(
				objDriveCredentials["drive_id"],
				objDriveCredentials["lun_id"],
				objDriveCredentials["storage_ip_address"],
				objDriveCredentials["storage_port"],
				objDriveCredentials["target_iqn"]
			);

			const driveModule = Modules.DriveSingleton;

			const moduleResponse = driveModule.getDiskPath(diskCredentials, false);
			const strDiskPath = moduleResponse.getOutput();

			return await this.instance_agent_disk_partition(incomingRequest, strDiskPath, objPartitionTable);
		}

		/**
		 * Partitions a generic disk. Disk path must be a resolved link
		 *  
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strDiskPath
		 * @param {Object} objPartitionTable
		 * @returns {Object}
		 */
		async instance_agent_disk_partition(incomingRequest, strDiskPath, objPartitionTable)
		{
			const diskModule = Modules.DiskSingleton;

			//objPartitionTable must have: type and partitions
			const partitionTable = diskModule.toPartitionTable(objPartitionTable);

			const objParams = {
				"disk_path": strDiskPath,
				"partition_table": partitionTable
			};

			let objReturnInfo = {};

			try
			{
				await Plays.DiskPartitionPlaySingleton.run(objParams);
				objReturnInfo = {
					"message": strDiskPath + " partitioned successfully!",
					"status": "successful",
					"data": null
				};
			}
			catch(error)
			{
				error.message = "Disk partitioning failed with: " + error.message;
				throw error;
			}

			return objReturnInfo;
		}

		/**
		 * Formats a drive partition with the given filesystem information. Drive must already exist
		 * in /dev/disk/by-path, after the iSCSI login and must already be partitioned.
		 * objFilesystemInformation needs to be an object sent by BSI, that holds filesystem_type and block_size
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objDriveCredentials
		 * @param {number} nPartitionIndex
		 * @param {Object} objFilesystemInformation
		 * @returns {Object}
		 */
		async instance_agent_drive_partition_format(incomingRequest, objDriveCredentials, nPartitionIndex, objFilesystemInformation)
		{
			//objDriveCredentials must have: drive_id, lun_id, storage_ip_address, storage_port and target_iqn
			const diskCredentials = new ISCSIDriveCredentials(
				objDriveCredentials["drive_id"],
				objDriveCredentials["lun_id"],
				objDriveCredentials["storage_ip_address"],
				objDriveCredentials["storage_port"],
				objDriveCredentials["target_iqn"]
			);

			const moduleResponse = Modules.DriveSingleton.getDiskPath(diskCredentials, false);
			const strPartitionPath = moduleResponse.getOutput() + "-part" + nPartitionIndex;

			return await this.instance_agent_disk_partition_format(incomingRequest, strPartitionPath, objFilesystemInformation);
		}

		/**
		 * Formats the disk partition with the given filesystem information.
		 *  
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strDiskPartitionPath
		 * @param {Object} objFilesystemInformation
		 * @returns {Object}
		 */
		async instance_agent_disk_partition_format(incomingRequest, strDiskPartitionPath, objFilesystemInformation)
		{
			const filesystemInformation = new FilesystemInformation(
				objFilesystemInformation["filesystem_type"],
				objFilesystemInformation["block_size"]
			);

			const objParams = {
				"disk_partition_path": strDiskPartitionPath,
				"filesystem_information": filesystemInformation
			};

			let objReturnInfo = {};

			try
			{
				await Plays.DiskPartitionFormatPlaySingleton.run(objParams);
				objReturnInfo = {
					"message": "Succesfully formated disk partition " + strDiskPartitionPath + ".",
					"status": "successful",
					"data": null
				};
			}
			catch(error)
			{
				error.message = "Disk partition formatting failed with: " + error.message;
				throw error;
			}

			return objReturnInfo;
		}

		/**
		 * Mounts the given partition of the drive using systemd.mount.
		 * Mount point can be specified or automatically generated (/mnt/drive{drive_id})
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {Object} objDriveCredentials
		 * @param {number} nPartitionIndex
		 * @param {Object} objFilesystemInformation
		 * @param {string|null} strMountPoint
		 * @returns {Object}
		 */
		async instance_agent_drive_partition_mount(incomingRequest, objDriveCredentials, nPartitionIndex, objFilesystemInformation, strMountPoint = null)
		{
			//objDriveCredentials must have: drive_id, lun_id, storage_ip_address, storage_port, target_iqn
			const driveCredentials = new ISCSIDriveCredentials(
				objDriveCredentials["drive_id"],
				objDriveCredentials["lun_id"],
				objDriveCredentials["storage_ip_address"],
				objDriveCredentials["storage_port"],
				objDriveCredentials["target_iqn"]
			);

			const driveModule = Modules.DriveSingleton;

			/**
			 * Partition mount point is automatically generated if not sent as the last parameter.
			 * Generated mount point has the format "/mnt/drive{drive_id}".
			 */
			const strPartitionMountPoint = strMountPoint === null ? driveModule.getMountPointPath(driveCredentials.driveID) : strMountPoint;

			//Get the unresolved symlink from /dev/disk/by-path
			const moduleResponse = driveModule.getDiskPath(driveCredentials, false);
			const strDrivePartitionPath = moduleResponse.getOutput() + "-part" + nPartitionIndex;

			return await this.instance_agent_disk_partition_mount(incomingRequest, strDrivePartitionPath, objFilesystemInformation, strPartitionMountPoint);
		}

		/**
		 * Mounts the given disk partition using systemd.mount.
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strDiskPartitionPath
		 * @param {Object} objFilesystemInformation
		 * @param {string} strMountPoint
		 * @returns {Object}
		 */
		async instance_agent_disk_partition_mount(incomingRequest, strDiskPartitionPath, objFilesystemInformation, strMountPoint)
		{
			//objFilesystemInformation needs to be an object sent by BSI, that holds filesystem_type and block_size
			const filesystemInformation = new FilesystemInformation(
				objFilesystemInformation["filesystem_type"],
				objFilesystemInformation["block_size"]
			);

			const objParams = {
				"disk_partition_path": strDiskPartitionPath,
				"filesystem_information": filesystemInformation,
				"mount_point": strMountPoint
			};

			let objReturnInfo = {};

			try
			{
				await Plays.DiskPartitionAutomaticMountPlaySingleton.run(objParams);
				objReturnInfo = {
					"message": "Successfully mounted partition " + strDiskPartitionPath + " at " + strMountPoint + ".",
					"status": "successful",
					"data": {
						"mount_point": strMountPoint
					}
				};
			}
			catch(error)
			{
				error.message = "Disk partition mounting failed with: " + error.message;
				throw error;
			}

			return objReturnInfo;
		}

		/**
		 * Unmount the disk partition at given mount point.
		 * 
		 * @param {JSONRPC.IncomingRequest} incomingRequest
		 * @param {string} strMountPoint
		 * @returns {Object}
		 */
		async instance_agent_disk_partition_unmount(incomingRequest, strMountPoint)
		{
			const objParams = {
				"mount_point": strMountPoint
			};

			let objReturnInfo = {};

			try
			{
				await Plays.DiskPartitionAutomaticUnmountPlaySingleton.run(objParams);
				objReturnInfo = {
					"message": "Successfully unmounted partition at mount point " + strMountPoint + ".",
					"status": "successful",
					"data": null
				};
			}
			catch(error)
			{
				error.message = "Disk partition unmounting failed with: " + error.message;
				throw error;
			}

			return objReturnInfo;
		}
	};
