const fs = require("fs-extra");
const url = require("url");
const path = require("path");
const pathExists = require("path-exists");
const assert = require("assert");

module.exports = 
class Identity
{
	/**
	 * @param {string} strAgentMode
	 * @param {string|null} strMACAddressSpoof
	 * @param {string|null} strIPAddressSpoof
	 * @param {string|null} strUUIDSpoof
	 */
	constructor(strAgentMode, strMACAddressSpoof, strIPAddressSpoof, strUUIDSpoof)
	{
		assert(["deploy", "bdk"].includes(strAgentMode), "Unknown agent mode " + JSON.stringify(strAgentMode));
		this._strAgentMode = strAgentMode;

		this._strSettingsDirectoryPath = null;

		this._objSettings = null;

		this._agentRegisterResponse = null;

		this._strMACAddressSpoof = strMACAddressSpoof;
		this._strIPAddressSpoof = strIPAddressSpoof;
		this._strUUIDSpoof = strUUIDSpoof;
	}


	/**
	 * Current instance's AES keys for decrypting received sensitive data. 
	 * Also used to authenticate to BSI using a challenge mechanism.
	 * 
	 * AES keys for JSONRPC signing and encrypting, for a specific instance.
	 * 
	 * @readonly
	 * 
	 * @returns {Object<string,Object>}
	 */
	get instanceEncryptionRotatingKeys()
	{
		if(!this._objSettings.instanceEncryptionRotatingKeys)
		{
			throw new Error("instanceEncryptionRotatingKeys is not defined in config. Please init and download the agent settings.json config manually into " + this.settingsJSONPath);
		}

		return this._objSettings.instanceEncryptionRotatingKeys;
	}

	/**
	 * Returns the AES keys for either corresponding server or the allocated pre-registration
	 * 
	 * @readonly
	 * 
	 * @returns {Object<string,Object>}
	 */
	get bdkEncryptionRotatingKeys()
	{
		const encryptionRotatingKeys = this._objSettings.serverEncryptionRotatingKeys;
		
		if(!encryptionRotatingKeys)
		{
			throw new Error("serverEncryptionRotatingKeys or serverPreregisterEncryptionRotatingKeys are not defined in config. Please init and download the agent settings.json config manually into " + this.settingsJSONPath);
		}

		return encryptionRotatingKeys;
	}


	get serverEncryptionRotatingKeys()
	{
		if(!this._objSettings.serverEncryptionRotatingKeys)
		{
			throw new Error("serverEncryptionRotatingKeys is not defined in config. Please init and download the agent settings.json config manually into " + this.settingsJSONPath);
		}

		return this._objSettings.serverEncryptionRotatingKeys;	
	}

	get serverPreregisterEncryptionRotatingKeys()
	{
		if(!this._objSettings.serverPreregisterEncryptionRotatingKeys)
		{
			throw new Error("serverPreregisterEncryptionRotatingKeys is not defined in config. Please init and download the agent settings.json config manually into " + this.settingsJSONPath);
		}

		return this._objSettings.serverPreregisterEncryptionRotatingKeys;	
	}

	/**
	 * @returns {string|null}
	 */
	get environment()
	{
		return this._objSettings.environment;
	}


	/**
	 * @param {string|null} strEnvironment
	 */
	set environment(strEnvironment)
	{
		this._objSettings.environment = strEnvironment;
	}


	/**
	 * @returns {string|null}
	 */
	get macAddressSpoof()
	{
		return this._strMACAddressSpoof;
	}


	/**
	 * @returns {string|null}
	 */
	get serverUUIDSpoof()
	{
		return this._strUUIDSpoof;
	}


	/**
	 * @returns {string|null}
	 */
	get endpointURL()
	{
		return this._objSettings.endpointURL;
	}


	/**
	 * @param {string} strEndpointURL
	 */
	set endpointURL(strEndpointURL)
	{
		this._objSettings.endpointURL = strEndpointURL;
	}


	/**
	 * @returns {string|null}
	 */
	get ipAddressSpoof()
	{
		return this._strIPAddressSpoof;
	}


	/**
	 * Returns a directory path. This application's files and settings are stored within.
	 * 
	 * @returns {string}
	 */
	get settingsDirectoryPath()
	{
		if(this._strSettingsDirectoryPath === null)
		{
			throw new Error("You must call init before using this property.");
		}

		return this._strSettingsDirectoryPath;
	}


	/**
	 * @returns {string}
	 */
	get settingsJSONPath()
	{
		return path.join(this.settingsDirectoryPath, "settings.json");
	}


	/**
	 * @returns {number}
	 */
	get agentID()
	{
		if(!this._objSettings.agentID)
		{
			throw new Error(`${this.applicationName}s are expected to have the agent_id setup in settings.json since the installation deploy stage. nAgentID is mandatory.`);
		}

		return parseInt(this._objSettings.agentID, 10);
	}


	/**
	 * @param {number|null} nAgentID
	 */
	set agentID(nAgentID)
	{
		this._objSettings.agentID = nAgentID;
	}


	/**
	 * @returns {Object|null}
	 */
	get agentRegisterResponse()
	{
		if(this._agentRegisterResponse === null)
		{
			throw new Error("this._agentRegisterResponse was not initialized.");
		}
		
		return this._agentRegisterResponse;
	}


	/**
	 * @param {Object} agentRegisterResponse
	 * 
	 * @returns {undefined}
	 */
	set agentRegisterResponse(agentRegisterResponse)
	{
		assert(agentRegisterResponse.type === "AgentRegisterResponse");
		this._agentRegisterResponse = agentRegisterResponse;
	}


	/** 
	 * Returns a string without spaces which is this application's name.
	 * 
	 * This string will never change to ensure updated agents can still access the same settings. 
	 * 
	 * @returns {string} 
	 */
	get applicationName()
	{
		switch(this._strAgentMode)
		{
			case "deploy":
				return "BSIDeployAgent";
			case "bdk":
				return "BDKAgent";
			default:
				throw new Error("Unhandled agent mode " + JSON.stringify(this._strAgentMode));
		}
	}


	/**
	 * Initializes the settings directory and a RSA key pair.
	 * 
	 * @returns {undefined}
	 */
	async init()
	{
		await this._initSettingsDirectory();
		await this._initRSAKeysPair();
		await this._initSettingsJSONFile();
	}


	/**
	 * Saves current settings to the settings JSON file path.
	 * 
	 * @returns {undefined}
	 */
	async saveSettings()
	{
		assert(this._objSettings.agentID === null || typeof this._objSettings.agentID === "number");
		assert([null, "production", "development", "integration"].includes(this._objSettings.environment));

		if(this._objSettings.endpointURL !== null)
		{
			const objURL = url.parse(this._objSettings.endpointURL);
			assert(objURL.protocol);
			assert(objURL.host);
			assert(objURL.path);
		}
		
		await fs.writeFile(this.settingsJSONPath, JSON.stringify(this._objSettings, null, "\t"), {flag: "w"});
	}


	/**
	 * @returns {undefined}
	 */
	async _initSettingsDirectory()
	{
		let strProgramDataDirectoryPath;

		if(process.platform === "win32")
		{
			assert(process.env.ALLUSERSPROFILE, "ALLUSERSPROFILE environment variable not set.");
			assert((await pathExists(process.env.ALLUSERSPROFILE)), "ALLUSERSPROFILE=" + JSON.stringify(process.env.ALLUSERSPROFILE) + " path not found.");

			strProgramDataDirectoryPath = process.env.ALLUSERSPROFILE;
		}
		else
		{
			assert(await pathExists("/etc"), "/etc not found. Current platform: " + JSON.stringify(process.platform) + ".");
			strProgramDataDirectoryPath = "/etc";
		}


		this._strSettingsDirectoryPath = path.join(strProgramDataDirectoryPath, this.applicationName);


		if(!(await pathExists(this.settingsDirectoryPath)))
		{
			await fs.mkdir(this.settingsDirectoryPath, parseInt("0666", 8));
		}
		assert(await pathExists(this.settingsDirectoryPath));
	}


	/**
	 * @returns {undefined}
	 */
	async _initRSAKeysPair()
	{
		if(this._strAgentMode !== "deploy")
		{
			return;
		}
	}


	/**
	 * @returns {undefined}
	 */
	async _initSettingsJSONFile()
	{
		let bSaveSettings = false;

		if(await pathExists(this.settingsJSONPath))
		{
			console.log("Loading " + this.settingsJSONPath + ".");

			let strJSONSettings = await fs.readFile(this.settingsJSONPath, "utf8");
			
			if(strJSONSettings === "")
			{
				strJSONSettings = "{}";
				bSaveSettings = true;
			}

			try
			{
				this._objSettings = JSON.parse(strJSONSettings);
			}
			catch(error)
			{
				throw new Error("Failed parsing " + this.settingsJSONPath + " with error message: " + error.message);
			}
		}
		else
		{
			console.log("Could not find " + this.settingsJSONPath + ". Creating...");

			this._objSettings = {};
			bSaveSettings = true;
		}

		if(this._objSettings.environment === undefined)
		{
			this._objSettings.environment = null;
			bSaveSettings = true;
		}

		if(this._objSettings.endpointURL === undefined)
		{
			this._objSettings.endpointURL = null;
			bSaveSettings = true;
		}

		if(this._objSettings.agentID === undefined)
		{
			this._objSettings.agentID = null;
			bSaveSettings = true;
		}

		if(bSaveSettings)
		{
			await this.saveSettings();
		}

		
		// Development hard-code.
		// If this optional property is found in the settings, the calling agent_register() on the BSI endpoint is skipped and this is used as result. 
		if(this._objSettings.agentRegisterResponse)
		{
			this.agentRegisterResponse = this._objSettings.agentRegisterResponse;
		}
	}
};

