const os = require("os");
const ModuleResponse = require("./ModuleResponse");
const ModuleError = require("./ModuleError");

const LoggedChildProcessSingleton = require("../../LoggedChildProcessSingleton");

/**
 * Used for running os commands. Allows using shell on linux and spawning new processes for workloads
 */
class Command
{
	constructor()
	{
		this.arrWhitelistRegex = [
			"^ifup [a-zA-Z0-9]*",
			"^ifsup [a-zA-Z0-9]*", //testing
			"^ifdown [a-zA-Z0-9]*",
			"^ip r del default dev [a-zA-Z0-9]*"
		];
	}

	/**
	 * @param {string} strCommand
	 * @param {Object} objOptions
	 * @returns {boolean}
	 */
	async execute(strCommand, objOptions)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		objOptions = objOptions || {};
		objOptions.skipWhitelistCheck = objOptions.skipWhitelistCheck || false;
		objOptions.fnCheckResult = objOptions.fnCheckResult || null;
		objOptions.fnChangedWhen = objOptions.fnChangedWhen || null;

		let fnResolve;
		let fnReject;
		const promise = new Promise((fnPromiseResolve, fnPromiseReject) =>
		{
			fnResolve = fnPromiseResolve;
			fnReject = fnPromiseReject;
		});

		if (!objOptions.skipWhitelistCheck && !this._isWhitelisted(strCommand))
		{
			fnReject(new ModuleError(`Command "${strCommand}" is not whitelisted!`));

			return promise;
		}

		const childProcess = LoggedChildProcessSingleton.spawn(strCommand, objOptions);

		let strStdout = "";
		let strStderr = "";
		childProcess.stdout.on("data", (data) =>
		{
			strStdout += data;
		});
		childProcess.stderr.on("data", (data) =>
		{
			strStderr += data;
		});
		childProcess.on("error", (error) =>
		{
			console.log(error);
			fnReject(new ModuleError(`Could not execute command: ${strCommand}`, error));
		});

		childProcess.on("exit", (nResultCode, strSignal) =>
		{
			if (objOptions.fnCheckResult != null && !objOptions.fnCheckResult(nResultCode, strStdout, strStderr, strCommand))
			{
				const strMessage = `User defined response validation function fails when running command: "${strCommand}".`;
				fnReject(new ModuleError(strMessage));
				return;
			}

			if (strSignal !== null)
			{
				const strMessage = `Child process stopped with signal: ${strSignal}`;
				fnReject(new ModuleError(strMessage));
				return;
			}

			let bIsChanged = true;
			if (objOptions.fnChangedWhen !== null)
			{
				bIsChanged = objOptions.fnChangedWhen(nResultCode, strStdout, strStderr, strCommand);
			}

			fnResolve(new ModuleResponse(bIsChanged, strStdout, strStderr));
		});

		return promise;
	}

	/**
	 * @param {string} strCommand
	 * @returns {boolean}
	 * @private
	 */
	_isWhitelisted(strCommand)
	{
		let bIsWhitelisted = false;
		let item = null;
		for (item of this.arrWhitelistRegex)
		{
			if (strCommand.match(new RegExp(item)))
			{
				bIsWhitelisted = true;
				break;
			}
		}

		return bIsWhitelisted;
	}
}

module.exports = Command;
