const ModuleError = require("./ModuleError");
const Partition = require("../Models/Partition");
const PartitionTable = require("../Models/PartitionTable");
const partitionInformation = require("partitioninfo");
const fs = require("fs");

const CommandSingleton = require("./Singletons/CommandSingleton");

class Disk
{
	/**
	 * Creates a partition table for the disk with the given details.
	 * 
	 * @param {string} strDiskPath
	 * @param {PartitionTable} desiredPartitionTable
	 * @returns {ModuleResponse}
	 */
	async createPartitionTable(strDiskPath, desiredPartitionTable)
	{
		strDiskPath = fs.realpathSync(strDiskPath);

		//For the moment, desiredPartitionTable is not used
		const strPartitionCommand = `parted -s ${strDiskPath} mklabel gpt`;

		const objOptions = {
			skipWhitelistCheck: true,
			shell: true
		};

		const executeResponse = CommandSingleton.execute(strPartitionCommand, objOptions);
		return executeResponse;
	}

	/**
	 * Returns the partition table for the specified disk (ex: /dev/sdb).
	 * If disk path is invalid, throw an error.
	 * If no partitions found, return null.
	 * 
	 * @param {string} strDiskPath
	 * @returns {PartitionTable|null}
	 */
	async getPartitionTable(strDiskPath)
	{
		strDiskPath = fs.realpathSync(strDiskPath);

		try
		{
			const objPartitionTableInformation = await partitionInformation.getPartitions(strDiskPath);
			return this.toPartitionTable(objPartitionTableInformation);
		}
		catch(error)
		{
			/* @TODO: Better error handling. */
			if(error.code !== "ENOENT")
			{
				/**
				 * If no partition table has been found (or if the "getPartitions" command fails for some reason), return null.
				 * For example, this error is returned when there is no partition table: "SyntaxError: Invalid MBR boot signature. Expected 0xAA55, but saw 0x0."
				 */
				return null;
			}
			else
			{
				throw error;
			}
		}
	}

	/**
	 * Checks if disk has the same partition table as the given one.
	 * 
	 * @param {string} strDiskPath
	 * @param {PartitionTable} desiredPartitionTable
	 * @returns {boolean}
	 */
	async hasPartitionTable(strDiskPath, desiredPartitionTable)
	{
		strDiskPath = fs.realpathSync(strDiskPath);

		//Retrieve the partition table found on the disk
		const currentPartitionTable = await this.getPartitionTable(strDiskPath);

		if(currentPartitionTable === null)
		{
			return desiredPartitionTable === null;
		}

		return this.comparePartitionTables(currentPartitionTable, desiredPartitionTable);
	}

	/**
	 * Checks if the given partition tables are similar, with the same table type and partitions.
	 * Currently, only index is compared within a partition.
	 * Returns true if partition tables are the same.
	 * 
	 * @param {PartitionTable} objPartitionTable1
	 * @param {PartitionTable} objPartitionTable2
	 * @throws {ModuleError}
	 * @returns {boolean}
	 */
	comparePartitionTables(objPartitionTable1, objPartitionTable2)
	{
		if(objPartitionTable1 !== null && !(objPartitionTable1 instanceof PartitionTable))
		{
			throw new ModuleError("Partition table 1 is not of type PartitionTable.");
		}

		if(objPartitionTable2 !== null && !(objPartitionTable2 instanceof PartitionTable))
		{
			throw new ModuleError("Partition table 2 is not of type PartitionTable.");
		}

		if(objPartitionTable1 === null || objPartitionTable2 === null)
		{
			return objPartitionTable1 === objPartitionTable2;
		}

		if(
			objPartitionTable1.type === objPartitionTable2.type &&
			objPartitionTable1.partitionCount === objPartitionTable2.partitionCount
		)
		{
			for(let nIndex = 0; nIndex < objPartitionTable1.partitionCount; nIndex++)
			{
				let partition1 = objPartitionTable1.getPartition(nIndex);
				let partition2 = objPartitionTable2.getPartition(nIndex);

				//At the moment compare only index, in the future also compare type, size and offset!
				if(partition1.index !== partition2.index)
				{
					return false;
				}
			}
		}
		else
		{
			return false;
		}

		return true;
	}

	/**
	 * Converts an object that holds partition table information to a PartitionTable class.
	 * 
	 * @param {Object} objPartitionTable
	 * @returns {PartitionTable}
	 */
	toPartitionTable(objPartitionTable)
	{
		if(
			objPartitionTable !== null && 
			(
				!objPartitionTable.hasOwnProperty("type") || !objPartitionTable.hasOwnProperty("partitions")
			)
		)
		{
			throw new ModuleError("Partition table object must have 'type' and 'partitions' keys set.");
		}

		const strType = objPartitionTable["type"];
		const objPartitions = objPartitionTable["partitions"];
		const nPartitionsCount = objPartitions.length;

		let arrPartitions = [];

		for(let nPartitionIndex = 0; nPartitionIndex < nPartitionsCount; nPartitionIndex++)
		{
			const objPartition = objPartitions[nPartitionIndex];

			//For the moment, only index is used from the partitions.
			const partition = new Partition(
				objPartition["index"]
				//objPartition["type"],
				//objPartition["size"],
				//objPartition["offset"]
			);

			arrPartitions.push(partition);
		}

		const desiredPartitionTable = new PartitionTable(strType, nPartitionsCount, arrPartitions);

		return desiredPartitionTable;
	}

	/**
	 * Create a partition spanning the entire disk.
	 * 
	 * @param {string} strDiskPath
	 * @param {Partition} partition
	 * @returns {ModuleResponse}
	 */
	async createPartition(strDiskPath, partition)
	{
		strDiskPath = fs.realpathSync(strDiskPath);

		//For the moment, partition is not used
		const strPartitionCommand = `parted -s -a opt ${strDiskPath} mkpart primary 0% 100%`;

		const objOptions = {
			skipWhitelistCheck: true,
			shell: true
		};

		const executeResponse = CommandSingleton.execute(strPartitionCommand, objOptions);

		return executeResponse;
	}
}

module.exports = Disk;
