const ISCSIDriveCredentials = require("../Models/ISCSIDriveCredentials");
const ModuleResponse = require("./ModuleResponse");
const ModuleError = require("./ModuleError");
const os = require("os");
const fs = require("fs");

/**
 * Drive module class contains operations executed on logged-in iSCSI drives.
 */
class Drive 
{
	/**
	 * Receives an ISCSIDriveCredentials object and tries to find the name of the disk from /dev/disk/by-path.
	 * Example: ip-192.168.137.2:3260-iscsi-iqn.2013-01.com.bigstep:storage.t7z5xtn.cqur0x4.mt73s7e-lun-11  -> ../../sda
	 * If bResolveSymlink is true, it returns /dev/sda,
	 * else /dev/disk/by-path/ip-192.168.137.2:3260-iscsi-iqn.2013-01.com.bigstep:storage.t7z5xtn.cqur0x4.mt73s7e-lun-11
	 * 
	 * @param {ISCSIDriveCredentials} driveCredentials
	 * @param {boolean} bResolveSymlink
	 * @throws {ModuleError}
	 * @returns {ModuleResponse}
	 */
	getDiskPath(driveCredentials, bResolveSymlink)
	{
		if(!(driveCredentials instanceof ISCSIDriveCredentials))
		{
			throw new ModuleError("ISCSI drive credentials object is not of type ISCSIDriveCredentials.");
		}

		const strDriveName = `ip-${driveCredentials.portalAddress}:${driveCredentials.portalPort}-iscsi-${driveCredentials.targetIQN}-lun-${driveCredentials.LunIndex}`;
		let strDiskPath = this.devDiskByPath + "/" + strDriveName;

		if(!fs.existsSync(strDiskPath))
		{
			throw new ModuleError("Drive name not found. iSCSI setup and login might have failed.");
		}

		if(bResolveSymlink)
		{
			strDiskPath = fs.realpathSync(strDiskPath);
		}

		return new ModuleResponse(true, strDiskPath);
	}

	/**
	 * Returns the mount point used for the disk partition automatic mount play.
	 * For the moment, it works only for linux operating systems.
	 * 
	 * @param {number} nDriveID
	 */
	getMountPointPath(nDriveID)
	{
		let strMountPoint = null;

		if(os.platform() === "linux")
		{
			strMountPoint = "/mnt/drive" + nDriveID;
		}
		else
		{
			throw new Error("At the moment, drives are mounted only on linux operating systems.");
		}

		return strMountPoint;
	}

	/**
	 * @returns {string}
	 */
	get devDiskByPath()
	{
		return "/dev/disk/by-path";
	}

}
module.exports = Drive;
