const os = require("os");
const readline = require("readline");
const osPath = require("path");
const ModuleResponse = require("./ModuleResponse");
const ModuleError = require("./ModuleError");
const AtomicWriteFileSystemSingleton = require("../../FileSystem").AtomicWriteFileSystemSingleton;
const Path = require("./Path");

class File extends Path
{
	/**
	 * @param {string} strContent
	 * @param {string} strPath
	 * @param {number} nUserID
	 * @param {number} nGroupID
	 * @param {number} nMode
	 * @returns {ModuleResponse}
	 */
	async save(strContent, strPath, nUserID, nGroupID, nMode)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		nUserID = nUserID || 0; //should this be the default?
		nGroupID = nGroupID || 0; //should this be the default?
		nMode = nMode || 0o666;

		try
		{
			await AtomicWriteFileSystemSingleton.writeFile(strPath, strContent, {
				userID: nUserID,
				groupID: nGroupID,
				mode: nMode
			});

			return new ModuleResponse(true);
		}
		catch (error)
		{
			if (error.code === "EEXIST") // file already exists
			{
				let bIsChanged = false;

				// check if file contents are changed
				let strPreviousText = await AtomicWriteFileSystemSingleton.readFile(strPath);
				if (strPreviousText !== strContent)
				{
					await AtomicWriteFileSystemSingleton.writeFile(strPath, strContent, {
						overwrite: true,
						userID: nUserID,
						groupID: nGroupID,
						mode: nMode
					});
					bIsChanged = true;
				}

				const bChmodChanged = (await this.chmod(strPath, nMode)).isChanged();
				bIsChanged = bIsChanged || bChmodChanged;
				const bOwnerChanged = (await this.chown(strPath, nUserID, nGroupID)).isChanged();
				bIsChanged = bIsChanged || bOwnerChanged;

				return new ModuleResponse(bIsChanged);
			}

			if (error.code === "ENOENT") // path does not exist
			{
				const strFolderPath = strPath.substring(0, strPath.lastIndexOf(osPath.sep));
				await AtomicWriteFileSystemSingleton.ensureDir(strFolderPath);
				await AtomicWriteFileSystemSingleton.writeFile(strPath, strContent, {
					userID: nUserID,
					groupID: nGroupID,
					mode: nMode
				});

				return new ModuleResponse(true);
			}

			throw new ModuleError(`Can not save to path: "${strPath}"`, error);
		}
	}

	/**
	 * If no replaceLineRegex is specified or is not matched, line will be added at the end of the file
	 *
	 * @param {string} strLineContent
	 * @param {string} strPath
	 * @param {string} strReplaceLineRegex
	 * @returns {ModuleResponse}
	 */
	async addLine(strLineContent, strPath, strReplaceLineRegex)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		strReplaceLineRegex = strReplaceLineRegex || null;

		if (strReplaceLineRegex === null)
		{
			await AtomicWriteFileSystemSingleton.appendFileLine(strPath, strLineContent);

			return new ModuleResponse(true);
		}

		return await this.replaceLine(strLineContent, strPath, strReplaceLineRegex, true);
	}

	/**
	 * @param {string} strText
	 * @param {string} strPath
	 * @param {string} strRegex
	 * @param {boolean} bGlobal
	 * @returns {ModuleResponse}
	 */
	async replaceText(strText, strPath, strRegex, bGlobal)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let strFileText = await AtomicWriteFileSystemSingleton.readFile(strPath);

		let strFlags = "";
		if (bGlobal)
		{
			strFlags += "g";
		}
		const regex = new RegExp(strRegex, strFlags);
		const strNewFileText = strFileText.replace(regex, strText);
		if (strNewFileText === strFileText)
		{
			return new ModuleResponse(false);
		}

		return this.save(strNewFileText, strPath);
	}

	/**
	 * @param {string} strLineContent
	 * @param {string} strPath
	 * @param {string} strRegex
	 * @param {boolean} bAddIfMissing //If regex does not match anything, add line to end of file
	 * @returns {ModuleResponse}
	 */
	async replaceLine(strLineContent, strPath, strRegex, bAddIfMissing)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		bAddIfMissing = bAddIfMissing || false;

		let fnResolve;
		let bIsChanged = false;

		const promise = new Promise((fnPromiseResolve, fnPromiseReject) =>
		{
			fnResolve = fnPromiseResolve;
		});

		const readLineInterface = readline.createInterface({
			input: await AtomicWriteFileSystemSingleton.createReadStream(strPath)
		});


		const arrLines = [];
		let bLineExists = false;
		readLineInterface.on("line", function(strCurrentLine)
		{
			// replace only first occurence
			if (bLineExists)
			{
				arrLines.push(strCurrentLine);
				return;
			}

			if (strLineContent === strCurrentLine)
			{
				bLineExists = true;
				return;
			}

			const match = strCurrentLine.match(strRegex);
			if (match !== null)
			{
				bLineExists = true;
				bIsChanged = true;
				arrLines.push(strLineContent);
			}
			else
			{
				arrLines.push(strCurrentLine);
			}
		});

		readLineInterface.on("close", async() =>
		{
			// save file if changes occured
			if (bIsChanged)
			{
				await this.save(arrLines.join(os.EOL), strPath);
				fnResolve(new ModuleResponse(true));

				return;
			}

			if (bAddIfMissing && !bLineExists)
			{
				fnResolve(await this.addLine(strLineContent, strPath));
			}
			else
			{
				fnResolve(new ModuleResponse(false));
			}
		});

		return promise;
	}
}

module.exports = File;
