const ModuleError = require("./ModuleError");
const FilesystemInformation = require("../Models/FilesystemInformation");
const CommandSingleton = require("./Singletons/CommandSingleton");

/**
 * This class holds functions related to disk partition formating.
 */
class Filesystem
{
	/**
	 * Format the drive by choosing a filesystem for the partition.
	 * 
	 * @param {string} strDiskPartitionPath
	 * @param {FilesystemInformation} filesystemInformation
	 * @throws {ModuleError}
	 * @returns {ModuleResponse}
	 */
	async createFilesystem(strDiskPartitionPath, filesystemInformation)
	{
		if(!(filesystemInformation instanceof FilesystemInformation))
		{
			throw new ModuleError("Filesystem information parameter is not of type FilesystemInformation.");
		}

		const strFormatCommand = `mkfs -t ${filesystemInformation.filesystemType} -b ${filesystemInformation.blockSize} ${strDiskPartitionPath}`;

		const objOptions = {
			skipWhitelistCheck: true,
			shell: true
		};

		const executeResponse = await CommandSingleton.execute(strFormatCommand, objOptions);

		return executeResponse;
	}

	/**
	 * Returns the block size of the filesystem used on the given partition.
	 * Works only for ext filesystem types!
	 * 
	 * @param {string} strDiskPartitionPath
	 * @returns {FilesystemInformation|null}
	 */
	async getFilesystem(strDiskPartitionPath)
	{
		let moduleResponse = await this._getFilesystemBlockSize(strDiskPartitionPath);

		//If block size command returns error, then partition is not formated
		if(moduleResponse.getError())
		{
			return null;
		}

		const nBlockSize = parseInt(moduleResponse.getOutput().trim(), 10);

		moduleResponse = await this._getPartitionFilesystemType(strDiskPartitionPath);

		//If get filesystem type command fails or output is null, then partition is not formated
		if(moduleResponse.getError() !== null || moduleResponse.getOutput() === null)
		{
			//TODO: test this
			return null;
		}

		const strFilesystemType = moduleResponse.getOutput().trim();

		return new FilesystemInformation(strFilesystemType, nBlockSize);
	}

	/**
	 * Checks if disk has the same filesystem information as the given one.
	 * 
	 * @param {string} strDiskPartitionPath
	 * @param {FilesystemInformation} desiredFilesystemInformation
	 * @returns {boolean}
	 */
	async hasFilesystem(strDiskPartitionPath, desiredFilesystemInformation)
	{
		const currentFilesystemInformation = await this.getFilesystem(strDiskPartitionPath);

		if(currentFilesystemInformation === null)
		{
			return desiredFilesystemInformation === null;
		}

		return this.compareFilesystems(currentFilesystemInformation, desiredFilesystemInformation);
	}

	/**
	 * Checks if the given filesystem information instances are similar, with the same filesystem type and block size.
	 * Returns true if partition tables are the same.
	 * 
	 * @param {FilesystemInformation} filesystemInformation1
	 * @param {FilesystemInformation} filesystemInformation2
	 * @throws {ModuleError}
	 * @returns {boolean}
	 */
	compareFilesystems(filesystemInformation1, filesystemInformation2)
	{
		if(filesystemInformation1 !== null && !(filesystemInformation1 instanceof FilesystemInformation))
		{
			throw new ModuleError("Filesystem information 1 is not of type FilesystemInformation.");
		}

		if(filesystemInformation2 !== null && !(filesystemInformation2 instanceof FilesystemInformation))
		{
			throw new ModuleError("Filesystem information 2 is not of type FilesystemInformation.");
		}

		if(filesystemInformation1 === null || filesystemInformation2 === null)
		{
			return filesystemInformation1 === filesystemInformation2;
		}

		if(filesystemInformation1.filesystemType !== filesystemInformation2.filesystemType)
		{
			return false;
		}

		return filesystemInformation1.blockSize === filesystemInformation2.blockSize;
	}

	/**
	 * Returns the block size of the filesystem used on the given partition.
	 * Works only for ext filesystem types!
	 * 
	 * @param {string} strDiskPartitionPath
	 * @returns {ModuleResponse}
	 */
	async _getFilesystemBlockSize(strDiskPartitionPath)
	{
		const strStandardCommand = `tune2fs -l ${strDiskPartitionPath} | grep -i 'block size' | awk '{print $3}'`;

		const objOptions = {
			skipWhitelistCheck: true,
			shell: true
		};

		const executeResponse = await CommandSingleton.execute(strStandardCommand, objOptions);

		return executeResponse;
	}

	/**
	 * Returns the file system type for the drive's partition.
	 * 
	 * @param {string} strDiskPartitionPath
	 * @returns {ModuleResponse}
	 */
	async _getPartitionFilesystemType(strDiskPartitionPath)
	{
		const strStandardCommand = `blkid -o export ${strDiskPartitionPath} | grep '^TYPE' | cut -d\"=\" -f2`;

		const objOptions = {
			skipWhitelistCheck: true,
			shell: true
		};

		const executeResponse = CommandSingleton.execute(strStandardCommand, objOptions);

		return executeResponse;
	}
}

module.exports = Filesystem;
