const Path = require("./Path");
const ModuleResponse = require("./ModuleResponse");
const ModuleError = require("./ModuleError");
const os = require("os");
const AtomicWriteFileSystemSingleton = require("../../FileSystem").AtomicWriteFileSystemSingleton;

/**
 * @todo implement Windows support
 */
class Folder extends Path
{
	constructor()
	{
		super();
	}

	/**
	 * @param {string} strPath
	 * @param {number} nUserID
	 * @param {number} nGroupID
	 * @param {number} nMode
	 * @returns {ModuleResponse}
	 *
	 */
	async create(strPath, nUserID, nGroupID, nMode)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		nUserID = nUserID || 0; //should this be the default?
		nGroupID = nGroupID || 0; //should this be the default?
		nMode = nMode || 0o666;

		if (await AtomicWriteFileSystemSingleton.exists(strPath))
		{
			let bIsChanged = false;
			const objStats = await AtomicWriteFileSystemSingleton.stat(strPath);
			if (!objStats.isDirectory())
			{
				throw new ModuleError("Path is not empty and it does not represent a folder");
			}

			const bChmodChanged = (await this.chmod(strPath, nMode)).isChanged();
			bIsChanged = bIsChanged || bChmodChanged;
			const bOwnerChanged = (await this.chown(strPath, nUserID, nGroupID)).isChanged();
			bIsChanged = bIsChanged || bOwnerChanged;

			return new ModuleResponse(bIsChanged);
		}

		await AtomicWriteFileSystemSingleton.ensureDir(strPath);
		this.chmod(strPath, nMode);
		this.chown(strPath, nUserID, nGroupID);

		return new ModuleResponse(true);
	}

	/**
	 * Does not delete recursively!
	 *
	 * @param {string} strPath
	 * @param {string} strFileNamePattern
	 * @param {string} strFileContentPattern
	 */
	async grepDelete(strPath, strFileNamePattern, strFileContentPattern)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let isChanged = false;
		let arrToDeleteFilenames = await this._grep(strPath, strFileNamePattern, strFileContentPattern);
		for (const strFilename of arrToDeleteFilenames)
		{
			const response = await this.delete(strPath + strFilename);
			isChanged = isChanged || response.isChanged();
		}

		return new ModuleResponse(isChanged);
	}

	/**
	 * @param {string} strPath
	 * @param {string} strFileNamePattern
	 * @param {string} strFileContentPattern
	 * @returns {Array}
	 * @private
	 */
	async _grep(strPath, strFileNamePattern, strFileContentPattern)
	{
		strFileNamePattern = strFileNamePattern || null;
		strFileContentPattern = strFileContentPattern || null;

		const arrIfconfigFiles = await AtomicWriteFileSystemSingleton.readdir(strPath);
		let arrFoundFilenames = [];

		for (const strFileName of arrIfconfigFiles)
		{
			if (strFileNamePattern !== null)
			{
				const fileNameMatch = strFileName.match(strFileNamePattern);
				if (fileNameMatch === null)
				{
					continue;
				}
			}

			if (strFileContentPattern != null)
			{
				let strFileContent = await AtomicWriteFileSystemSingleton.readFile(strPath + "/" + strFileName, {
					"encoding": "utf-8"
				});

				const fileContentMatch = strFileContent.match(strFileContentPattern);
				if (fileContentMatch === null)
				{
					continue;
				}
			}

			arrFoundFilenames.push(strFileName);

		}

		return arrFoundFilenames;
	}
}

module.exports = Folder;
