const NetworkInterface = require("../../Facts").Models.NetworkInterface;
const ModuleResponse = require("./ModuleResponse");
const ModuleError = require("./ModuleError");
const os = require("os");
const FactsProviderSingleton = require("../../Facts").FactsProviderSingleton;
const StateHandlerSingleton = require("../State").StateHandlerSingleton;
const AtomicWriteFileSystemSingleton = require("../../FileSystem").AtomicWriteFileSystemSingleton;
const FileSingleton = require("./Singletons/FileSingleton");
const FolderSingleton = require("./Singletons/FolderSingleton");
const CommandSingleton = require("./Singletons/CommandSingleton");
const IfcfgVLANTemplateSingleton = require("./Singletons/Template/IfcfgVLANTemplateSingleton");

class Network
{
	constructor()
	{
		this._strNetworkScriptsPath = "/etc/sysconfig/network-scripts";
	}

	/**
	 * @param {string} strInterface
	 */
	async ifup(strInterface)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let ifupResponse = await CommandSingleton.execute(
			`ifup ${strInterface}`,
			{
				fnChangedWhen: (nResultCode, strStdout, strStderr, strCommand) =>
				{
					if (strStdout === null || !strStdout.trim())
					{
						return false; // interface already up
					}

					return true;
				},
				shell: true
			}
		);

		return ifupResponse;
	}

	/**
	 * @param {string} strInterface
	 */
	async ifdown(strInterface)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let ifdownResponse = await CommandSingleton.execute(
			`ifdown ${strInterface}`,
			{
				fnChangedWhen: (nResultCode, strStdout, strStderr, strCommand) =>
				{
					if (strStdout === null || !strStdout.trim())
					{
						return false; // interface already down
					}

					return true;
				},
				shell: true
			}
		);

		return ifdownResponse;
	}

	/**
	 * @param {string} strVlanName
	 * @param {string} strBondName
	 *
	 * @returns {ModuleResponse}
	 */
	async removeVLAN(strVlanName, strBondName)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		const reponse = FileSingleton.delete(`${this._strNetworkScriptsPath}/ifcfg-${strBondName}.` + strVlanName);

		const currentState = await StateHandlerSingleton.get();
		currentState.removeVLAN(strVlanName);
		await StateHandlerSingleton.save(currentState);

		return reponse;
	}

	/**
	 * @param {VLAN} vlan
	 *
	 * @returns {ModuleResponse}
	 */
	async addVLAN(vlan)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		const strBondPath = `${this._strNetworkScriptsPath}/ifcfg-${vlan.bondName}`;
		const response = await AtomicWriteFileSystemSingleton.exists(strBondPath);
		if (!response)
		{
			throw new ModuleError(`Bond ${vlan.bondName} not configured, configure bond before adding vlans!`);
		}

		const currentState = await StateHandlerSingleton.get();
		currentState.addVLAN(vlan);
		await StateHandlerSingleton.save(currentState);

		const ifcfgVLANReponse = await IfcfgVLANTemplateSingleton.renderToFile(
			`${this._strNetworkScriptsPath}/ifcfg-${vlan.bondName}.${vlan.vlanName}`,
			vlan
		);

		return ifcfgVLANReponse;
	}

	/**
	 * @param {Array} existingVlans
	 * @param {string} strBondName
	 * @returns {ModuleResponse}
	 */
	async cleanIfcfgs(existingVlans, strBondName)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let bIsChanged = false;

		let arrVLANNames = [];
		for (const vlan of existingVlans)
		{
			arrVLANNames.push(`${vlan.name }$`);
		}
		const strVLANPattern = arrVLANNames.join("|");
		const vlansDeleteResponse = await FolderSingleton.grepDelete(
			this._strNetworkScriptsPath,
			`^ifcfg\-${strBondName}\.(?:(?!${strVLANPattern}).*)$`
		);
		bIsChanged = bIsChanged || vlansDeleteResponse.isChanged();

		const arrInterfaces = await FactsProviderSingleton.getInterfaces(NetworkInterface.TYPE_PHYSICAL);
		let arrInterfaceNames = [];
		for (const networkInterface of arrInterfaces)
		{
			arrInterfaceNames.push(networkInterface.getName() + "$");
		}
		const strInterfacesPattern = arrInterfaceNames.join("|");
		const slavesDeleteResponse = await FolderSingleton.grepDelete(
			this._strNetworkScriptsPath,
			`^ifcfg\-(?:(?!${strInterfacesPattern}).*)$`,
			`MASTER=${strBondName}`
		);
		bIsChanged = bIsChanged || slavesDeleteResponse.isChanged();

		return new ModuleResponse(bIsChanged);
	}

	/**
	 * @param {Array} arrNameServers
	 * @param {Array} arrSearchDomains
	 * @returns {ModuleResponse}
	 */
	async configureResolvConf(arrNameServers, arrSearchDomains)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let bIsChanged = false;

		for (const strNameServer of arrNameServers)
		{
			const nameServerResponse = await FileSingleton.addLine(`nameserver ${strNameServer}`, "/etc/resolv.conf", `^nameserver ${strNameServer}$`);
			bIsChanged = bIsChanged || nameServerResponse.isChanged();
		}

		const arrCurrentSearchDomains = await FactsProviderSingleton.getResolvNameServers();
		const arrNewSearchDomains = arrCurrentSearchDomains.concat(arrSearchDomains);
		const arrUniqueNewSearchDomains = arrNewSearchDomains.filter((item, nPosition) => arrNewSearchDomains.indexOf(item) === nPosition);
		const strNewUniqueSearchDomains = arrUniqueNewSearchDomains.join(" ");
		const searchResponse = await FileSingleton.addLine(`search ${strNewUniqueSearchDomains}`, "/etc/resolv.conf", "^search .*");
		bIsChanged = bIsChanged || searchResponse.isChanged();

		return new ModuleResponse(bIsChanged);
	}

	/**
	 * @param {string} strRouteName
	 * @returns {boolean}
	 */
	async deleteRoute(strRouteName)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		return await CommandSingleton.execute(`ip r del default dev ${strRouteName}`, {
			fnChangedWhen: (nResultCode, strStdout, strStderr, strCommand) => (strStdout !== null) && (strStdout.indexOf("Cannot find device") !== -1),
			shell: true
		});
	}

	/**
	 * @returns {string}
	 */
	get networkScriptsPath()
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		return this._strNetworkScriptsPath;
	}
}

module.exports = Network;
