const ModuleResponse = require("./ModuleResponse");
const ModuleError = require("./ModuleError");
const os = require("os");
const AtomicWriteFileSystemSingleton = require("../../FileSystem").AtomicWriteFileSystemSingleton;

class Path
{
	/**
	 * @param {string} strPath
	 * @param {number} nMode
	 * @returns {ModuleResponse}
	 */
	async chmod(strPath, nMode)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let objStats = await AtomicWriteFileSystemSingleton.stat(strPath);

		//check if file mode has changed
		if (nMode !== (objStats.mode & 0o777))
		{
			await AtomicWriteFileSystemSingleton.chmod(strPath, nMode);

			return new ModuleResponse(true);
		}

		return new ModuleResponse(false);
	}

	/**
	 * @param {string} strPath
	 * @param {number} nUserID
	 * @param {number} nGroupID
	 * @returns {ModuleResponse}
	 */
	async chown(strPath, nUserID, nGroupID)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		let objStats = await AtomicWriteFileSystemSingleton.stat(strPath);

		//check if file owner user or group has changed
		if (objStats.uid !== nUserID || objStats.gid !== nGroupID)
		{
			await AtomicWriteFileSystemSingleton.chown(strPath, nUserID, nGroupID);
			return new ModuleResponse(true);
		}

		return new ModuleResponse(false);
	}

	/**
	 * @param {string} strPath
	 * @returns {ModuleResponse}
	 */
	async delete(strPath)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		try
		{
			await AtomicWriteFileSystemSingleton.unlink(strPath);

			return new ModuleResponse(true);
		}
		catch (error)
		{
			if (error.code === "ENOENT")
			{
				return new ModuleResponse(false); // file does not exist
			}

			throw new ModuleError(`Could not delete path: ${strPath}`, error);
		}
	}

	/**
	 * @param {string} strPath
	 * @returns {ModuleResponse}
	 */
	async exists(strPath)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}

		try
		{
			await AtomicWriteFileSystemSingleton.exists(strPath);

			return new ModuleResponse(true);
		}
		catch (error)
		{
			throw new ModuleError(`Could not check if path: ${strPath} exists`, error);
		}
	}

	/**
	 * @param {string} strTarget
	 * @param {string} strPath
	 * @returns {ModuleResponse}
	 */
	async symlink(strTarget, strPath)
	{
		if (os.platform() !== "linux")
		{
			throw new Error("Not implemented. Only linux modules avaialble at this point.");
		}
		
		try
		{
			await AtomicWriteFileSystemSingleton.symlink(strTarget, strPath);

			return new ModuleResponse(true);
		}
		catch (error)
		{
			if (error.code === "EEXIST")
			{
				return new ModuleResponse(false);
			}

			throw new ModuleError(`Could not create symlink to path: "${strPath}"`, error);
		}
	}
}

module.exports = Path;
