const os = require("os");
const sleep = require("sleep-promise");
const ChildProcess = require("child-process-promise");
const ModuleResponse = require("./ModuleResponse");
const StateHandlerSingleton = require("../State").StateHandlerSingleton;

class Power
{
	/**
	 * @param {boolean} bReboot
	 * @param {boolean} bForce
	 * @param {number} nTimeoutMilliseconds
	 *
	 * @returns {string}
	 */
	async powerOff(bReboot, bForce, nTimeoutMilliseconds)
	{
		bForce = !!bForce;

		if (!nTimeoutMilliseconds)
		{
			nTimeoutMilliseconds = 0;
		}

		let strCommand;

		if (os.platform() === "win32")
		{
			// https://technet.microsoft.com/en-us/library/cc732503(v=ws.11).aspx

			strCommand = "shutdown";

			if (bReboot)
			{
				strCommand += " /r";
			}
			else
			{
				strCommand += " /s";
			}

			if (bForce)
			{
				strCommand += " /f";
			}
			// Always delaying by 1 second to get a chance to send stdout as function return.
			strCommand += " /t " + parseInt(Math.max(nTimeoutMilliseconds, 1000) / 1000, 10);
		}
		else if (
			os.platform() === "linux"
			|| os.platform() === "darwin"
			|| os.platform() === "freebsd"
			|| os.platform() === "openbsd"
			|| os.platform() === "aix"
		)
		{
			// https://linux.die.net/man/8/shutdown
			// https://developer.apple.com/legacy/library/documentation/Darwin/Reference/ManPages/man8/shutdown.8.html
			// https://www.freebsd.org/cgi/man.cgi?shutdown(8)
			// http://man.openbsd.org/shutdown
			// https://www.ibm.com/support/knowledgecenter/ssw_aix_61/com.ibm.aix.cmds5/shutdown.htm

			strCommand = "shutdown -r"; //-P to power off, -p to power off on freebsd and openbsd and aix

			if (bReboot)
			{
				strCommand += " -r";
			}
			else
			{
				if (os.platform() === "linux")
				{
					strCommand += " -P";
				}
				else
				{
					strCommand += " -p";
				}
			}

			if (bForce)
			{
				if (os.platform() === "aix")
				{
					strCommand += " -F";
				}
				else
				{
					strCommand += " -f";
				}
			}

			if (nTimeoutMilliseconds >= 1000)
			{
				if (nTimeoutMilliseconds >= 1000 * 60)
				{
					await sleep((parseInt(nTimeoutMilliseconds / 1000, 10) % 60) * 1000);
					strCommand += " +" + parseInt(nTimeoutMilliseconds / 1000 / 60, 10);
				}
				else
				{
					// await sleep(nTimeoutMilliseconds);
					strCommand += " +1";
				}
			}
			else
			{
				strCommand += " now";
			}
		}
		else if (os.platform() === "sunos")
		{
			// https://docs.oracle.com/cd/E23823_01/html/816-5166/shutdown-1m.html

			if (bReboot)
			{
				strCommand = "shutdown -y -i6 -g" + parseInt(nTimeoutMilliseconds / 1000, 10);
			}
			else
			{
				strCommand = "shutdown -y -i5 -g" + parseInt(nTimeoutMilliseconds / 1000, 10);
			}
		}
		else
		{
			throw new Error("shutdown CLI commands not implemented for os.platform() " + os.platform());
		}

		if (bReboot)
		{
			const currentTargetState = await StateHandlerSingleton.get();
			currentTargetState.rebootRequestedDate = new Date();
			await StateHandlerSingleton.save(currentTargetState);
		}

		const result = (await ChildProcess.exec(strCommand, {maxBuffer: 500 * 1024 * 1024}));

		return new ModuleResponse(true, result.stdout, result.stderr);
	}
}

module.exports = Power;
