const BasePlay = require("./BasePlay");
const BondMaster = require("../Models").BondMaster;
const BondSlave = require("../Models").BondSlave;
const NetworkInterface = require("../../Facts").Models.NetworkInterface;
const FileSingleton = require("../Modules").FileSingleton;
const NetworkSingleton = require("../Modules").NetworkSingleton;
const IfcfgBondMasterTemplateSingleton = require("../Modules").IfcfgBondMasterTemplateSingleton;
const IfcfgBondSlaveTemplateSingleton = require("../Modules").IfcfgBondSlaveTemplateSingleton;
const CloudInitScriptTemplateSingleton = require("../Modules").CloudInitScriptTemplateSingleton;
const BondFixScriptTemplateSingleton = require("../Modules").BondFixScriptTemplateSingleton;
const BondFixPreNetworkingScriptTemplateSingleton = require("../Modules").BondFixPreNetworkingScriptTemplateSingleton;
const BondFixPreNetworkingUnitTemplateSingleton = require("../Modules").BondFixPreNetworkingUnitTemplateSingleton;
const ServiceManagerSingleton = require("../Modules").ServiceManagerSingleton;
const PowerSingleton = require("../Modules").PowerSingleton;
const StateHandlerSingleton = require("../State").StateHandlerSingleton;
const FactsProviderSingleton = require("../../Facts").FactsProviderSingleton;
const assert = require("assert");

class ConfigureNetworkingPlay extends BasePlay
{
	constructor()
	{
		super();
		this.addStep(this.addBondFixPreNetworkScript.bind(this));
		this.addStep(this.configureBond.bind(this));
		this.addStep(this.addVLANs.bind(this));
		this.addStep(this.cleanIfcfgs.bind(this));
		this.addStep(this.restart.bind(this));
	}

	async validateParams()
	{
		assert(this._objParams !== null && undefined !== this._objParams, "This play must have vlans param set!");
	}

	/**
	 * @returns {boolean}
	 */
	async addBondFixPreNetworkScript()
	{
		let bIsChanged = false;
		const addScriptResponse = await BondFixPreNetworkingScriptTemplateSingleton.renderToFile(
			`/usr/sbin/bigstep-${this._objParams.bondName}-pre-networking-fix`,
			{"bondName": this._objParams.bondName},
			{mode: 0o755}
		);
		bIsChanged = bIsChanged || addScriptResponse.isChanged();

		const addUnitResponse = await BondFixPreNetworkingUnitTemplateSingleton.renderToFile(
			`/etc/systemd/system/bigstep-${this._objParams.bondName}-pre-networking-fix.service`,
			{"bondName": this._objParams.bondName},
			{mode: 0o755}
		);
		bIsChanged = bIsChanged || addUnitResponse.isChanged();

		const unitEnabledResponse = await ServiceManagerSingleton.enable(`bigstep-${this._objParams.bondName}-pre-networking-fix`);
		bIsChanged = bIsChanged || unitEnabledResponse.isChanged();

		return bIsChanged;
	}

	/**
	 * @returns {boolean}
	 */
	async addVLANs()
	{
		let bIsChanged = false;
		for (let vlan of this._objParams.VLANs)
		{
			let ifcfgVLANReponse = await NetworkSingleton.addVLAN(vlan);
			bIsChanged = bIsChanged || ifcfgVLANReponse.isChanged();
		}

		return bIsChanged;
	}


	/**
	 * @returns {boolean}
	 */
	async restart()
	{
		await PowerSingleton.powerOff(true, true, 1000);
	}
	/**
	 * @returns {boolean}
	 */
	async cleanIfcfgs()
	{
		let bIsChanged = false;

		const cleanIfcfgResponse = await NetworkSingleton.cleanIfcfgs(this._objParams.VLANs, this._objParams.bondName);
		bIsChanged = bIsChanged || cleanIfcfgResponse.isChanged();

		return bIsChanged;
	}

	/**
	 * @returns {boolean}
	 * @private
	 */
	async configureBond()
	{
		const strBondName = this._objParams.bondName;

		const currentTargetState = await StateHandlerSingleton.get();
		currentTargetState.isBondConfigured = true;
		currentTargetState.VLANs = this._objParams.VLANs;
		await this._stateHandler.save(currentTargetState);

		let bIsChanged = false;
		const replaceGrub2Response = await FileSingleton.replaceText(
			`rd.iscsi.ibft=1 bond=${strBondName}:ALL:mode=4,miimon=100,lacp_rate=1 BOOTIF=${strBondName} ip=${strBondName}:dhcp netroot=iscsi:ibft`,
			"/boot/grub2/grub.cfg",
			"ip=ibft",
			true
		);
		bIsChanged = bIsChanged || replaceGrub2Response.isChanged();

		const replaceGrubResponse = await FileSingleton.replaceText(
			`rd.iscsi.ibft=1 bond=${strBondName}:ALL:mode=4,miimon=100,lacp_rate=1 BOOTIF=${strBondName} ip=${strBondName}:dhcp netroot=iscsi:ibft`,
			"/boot/efi/EFI/centos/grub.cfg",
			"ip=ibft",
			true
		);
		bIsChanged = bIsChanged || replaceGrubResponse.isChanged();

		const cloudInitScriptResponse = await CloudInitScriptTemplateSingleton.renderToFile(
			"/etc/dhcp/dhclient.d/bigstep_cloudinit_dhcp_hook.sh",
			{"bondName": strBondName},
			{mode: 0o755}
		);
		bIsChanged = bIsChanged || cloudInitScriptResponse.isChanged();

		const bondFixScriptResponse = await BondFixScriptTemplateSingleton.renderToFile(
			`/etc/NetworkManager/dispatcher.d/98-${strBondName}-fix`,
			{"bondName": strBondName},
			{mode: 0o755}
		);
		bIsChanged = bIsChanged || bondFixScriptResponse.isChanged();

		const ifcfgMasterResponse = await IfcfgBondMasterTemplateSingleton.renderToFile(
			`${NetworkSingleton.networkScriptsPath}/ifcfg-${strBondName}`,
			new BondMaster(strBondName)
		);
		bIsChanged = bIsChanged || ifcfgMasterResponse.isChanged();

		const arrInterfaces = await FactsProviderSingleton.getInterfaces(NetworkInterface.TYPE_PHYSICAL);
		for (let networkInterface of arrInterfaces)
		{
			const ifcfgSlaveResponse = await IfcfgBondSlaveTemplateSingleton.renderToFile(
				`${NetworkSingleton.networkScriptsPath}/ifcfg-${ networkInterface.getName()}`,
				new BondSlave(networkInterface.getName(), strBondName)
			);
			bIsChanged = bIsChanged || ifcfgSlaveResponse.isChanged();
		}

		return bIsChanged;
	}
}

module.exports = ConfigureNetworkingPlay;
