const BasePlay = require("./BasePlay");
const FilesystemInformation = require("../Models/FilesystemInformation");
const isValidPath = require("is-valid-path");
const os = require("os");
const fs = require("fs");
const assert = require("assert");
const MountManagerSingleton = require("../Modules").MountManagerSingleton;

class DiskPartitionAutomaticMountPlay extends BasePlay
{
	/**
	 * 	The play receives the disk partition path, filesystem type and mountpoint.
		Standars mountpoint is /mnt/drive<drive_id> if not specified; Example: /mnt/drive608
		Steps:
		1.check if drive is already mounted
		If not mounted, do following steps:
			2. Create mount file, for example: /etc/systemd/system//mnt-drive608.mount
			3. Create auto-mount file, similar to the first one, for example: /etc/systemd/system/mnt-drive608.automount
			4. Notify systemd there are some new files available: systemctl daemon-reload
			5. systemctl start+enable /etc/systemd/system/mnt-drive608.automount
			6. Check if mounting was successful.
	 */
	constructor()
	{
		super();
		this._nRetryCount = 0;
		this._nRetryTimeout = 3000; //3s

		this.addStep(this.setParams.bind(this));
		this.addStep(this.checkAndAutoMountDiskPartition.bind(this));
	}

	/**
	 * PartitionTable parameters are validated inside the object when they are created.
	 * The other parameters are validated here.
	 * 
	 * @throws {Error}
	 */
	validateParams()
	{
		assert(typeof this._objParams["disk_partition_path"] === "string", "Disk partition path must be set and must be a string.");
		assert(this._objParams["filesystem_information"] instanceof FilesystemInformation, "Filesystem information object is not of type FilesystemInformation.");

		if(
			this._objParams["mount_point"] !== null &&
			(
				!this._objParams["mount_point"].startsWith("/") ||
				!isValidPath(this._objParams["mount_point"])
			)
		)
		{
			if(os.platform() === "linux")
			{
				throw new Error("Mountpoint is not a valid Linux path.");
			}
			//isValidPath doesn't seem to work well with Windows paths
			else if(os.platform() === "win32")
			{
				throw new Error("Mountpoint is not a valid Windows path. Also, only linux paths are supported at the moment.");
			}
		}
	}

	/**
	 * Sets parameters for the play after validation
	 */
	setParams()
	{
		this._strDiskPartitionPath = this._objParams["disk_partition_path"];
		this._strResolvedDiskPartitionPath = fs.realpathSync(this._strDiskPartitionPath);
		this._filesystemInformation = this._objParams["filesystem_information"];
		this._strMountPoint = this._objParams["mount_point"];
	}

	/**
	 * Checks if the partition automount has already been set. If it has not already been set,
	 * execute the rest of the play.
	 */
	async checkAndAutoMountDiskPartition()
	{
		console.log("Starting disk partition automatic mount play...");

		const moduleResponse = await MountManagerSingleton.isAutomountSet(
			this._strMountPoint,
			this._strDiskPartitionPath,
			this._filesystemInformation.filesystemType
		);
		const bIsAutomountSet = moduleResponse.getOutput().trim() === "true";

		if(bIsAutomountSet)
		{
			console.log("Partition " + this._strResolvedDiskPartitionPath + " automount has already been set.");
		}
		else
		{
			console.log("Will mount partition " + this._strResolvedDiskPartitionPath + ".");
			this.addStep(this.autoMountDiskPartition.bind(this));
			this.addStep(this.checkDiskPartitionAutoMount.bind(this));
		}
	}

	/**
	 * Creates .mount and .automount unit files in /etc/systemd/system/,
	 * reloads starts the mount service
	 * and starts and enables the automount service
	 */
	async autoMountDiskPartition()
	{
		await MountManagerSingleton.mountAndSetAutomount(
			this._strMountPoint,
			this._strDiskPartitionPath,
			this._filesystemInformation.filesystemType
		);

		console.log("Automounted disk partition " + this._strDiskPartitionPath + ".");
	}

	/**
	 * Checks whether the disk partition automatic mount has been set successfully.
	 */
	async checkDiskPartitionAutoMount()
	{
		const moduleResponse = await MountManagerSingleton.isAutomountSet(
			this._strMountPoint,
			this._strDiskPartitionPath,
			this._filesystemInformation.filesystemType
		);
		const bIsAutomountSet = moduleResponse.getOutput().trim() === "true";

		if(bIsAutomountSet)
		{
			console.log("Disk partition " + this._strResolvedDiskPartitionPath + " has been mounted successfully.");
		}
		else
		{
			if(this._nRetryCount > 10)
			{
				throw new Error("Disk partition " + this._strResolvedDiskPartitionPath + " has not been mounted successfully.");
			}
			setTimeout(
				async() =>
				{
					console.log("Retrying disk partition mount check");
					this._nRetryCount++;
					await this.checkDiskPartitionAutoMount();
				},
				this._nRetryTimeout
			);
		}
	}
}

module.exports = DiskPartitionAutomaticMountPlay;
