const BasePlay = require("./BasePlay");
const isValidPath = require("is-valid-path");
const os = require("os");

const MountManagerSingleton = require("../Modules").MountManagerSingleton;

class DiskPartitionAutomaticUnmountPlay extends BasePlay
{
	/**
	 * The play receives the mount point for which to remove the auto-mount service and the mount files
	 * from /etc/systemd/system/. Mount point should be a valid path.
	 * Stop and disable the services and remove the mount files from /etc/system/systemd.
	 */
	constructor()
	{
		super();
		this.addStep(this.setParams.bind(this));
		this.addStep(this.checkAndAutoUnmountDiskPartition.bind(this));
	}

	/**
	 * Mount point should be a valid linux or windows path.
	 * 
	 * @throws {Error}
	 */
	validateParams()
	{
		if(!this._objParams["mount_point"].startsWith("/") || !isValidPath(this._objParams["mount_point"]))
		{
			if(os.platform() === "linux")
			{
				throw new Error("Mountpoint is not a valid Linux path.");
			}
			//isValidPath doesn't seem to work well with Windows paths
			else if(os.platform() === "win32")
			{
				throw new Error("Mountpoint is not a valid Windows path. Also, only linux paths are supported at the moment.");
			}
		}
	}

	/**
	 * Sets parameters for the play after validation
	 */
	setParams()
	{
		this._strMountPoint = this._objParams["mount_point"];
	}

	/**
	 * Checks if the partition has already been unmounted. If it has not already been unmounted,
	 * execute the rest of the play.
	 */
	async checkAndAutoUnmountDiskPartition()
	{
		console.log("Starting disk partition automatic unmount play...");

		const moduleResponse = await MountManagerSingleton.isAutomountSet(this._strMountPoint);
		const bIsAutomountSet = moduleResponse.getOutput().trim() === "true";

		console.log(bIsAutomountSet);
		if(!bIsAutomountSet)
		{
			console.log(this._strMountPoint + " has already been unmounted.");
		}
		else
		{
			console.log("Will unmount " + this._strMountPoint + ".");
			this.addStep(this.unmountDiskPartition.bind(this));
			this.addStep(this.checkDiskPartitionAutoUnmount.bind(this));
		}
	}

	/**
	 * Checks if the automount service exists and stops and disables it. It then removes the mount files from
	 * /etc/systemd/system/ if they exist.
	 */
	async unmountDiskPartition()
	{
		await MountManagerSingleton.unmountAndUnsetAutomount(this._strMountPoint);

		console.log("Unmounted " + this._strMountPoint + ".");
	}

	/**
	 * Checks whether the disk partition automatic mount has been unset successfully.
	 */
	async checkDiskPartitionAutoUnmount()
	{
		const moduleResponse = await MountManagerSingleton.isAutomountSet(this._strMountPoint);
		const bIsAutomountSet = moduleResponse.getOutput().trim() === "true";

		if(!bIsAutomountSet)
		{
			console.log(this._strMountPoint + " has been unmounted successfully.");
		}
		else
		{
			throw new Error(this._strMountPoint + " has not been unmounted successfully.");
		}
	}
}

module.exports = DiskPartitionAutomaticUnmountPlay;
