const BasePlay = require("./BasePlay");
const FilesystemInformation = require("../Models/FilesystemInformation");
const FilesystemSingleton = require("../Modules").FilesystemSingleton;
const assert = require("assert");

class DiskPartitionFormatPlay extends BasePlay
{
	/**
	 * 	The play receives the partition disk path and filesystem information.
		It formats the disk partitions with the chosen filesystem information.
		It first checks if the partition is formated as expected. It then formats it and checks again if
		formatting was successful.
	 */
	constructor()
	{
		super();
		this._arrAllowedFilesystemTypes = ["ext3", "ext4"];

		this.addStep(this.setParams.bind(this));
		this.addStep(this.checkAndCreateFilesystem.bind(this));
	}

	/**
	 * Params are validated inside the objects when they are created
	 */
	validateParams()
	{
		assert(typeof this._objParams["disk_partition_path"] === "string", "Partition disk path must be set and must be a string.");
		assert(this._objParams["filesystem_information"] instanceof FilesystemInformation, "Filesystem information object is not of type FilesystemInformation.");

		//Additional checks for this play
		const strFilesystemType = this._objParams["filesystem_information"].filesystemType;

		if(!this._checkFilesystemType(strFilesystemType))
		{
			throw new Error(
				"Filesystem type " + strFilesystemType + " is not supported. " +
				"Allowed filesystems are " + JSON.stringify(this._arrAllowedFilesystemTypes));
		}

		if(this._objParams["filesystem_information"].blockSize < 1024)
		{
			throw new Error("Block size must be at least 1024 bytes.");
		}
	}

	/**
	 * Sets parameters for the play after validation
	 */
	setParams()
	{
		this._strDiskPartitionPath = this._objParams["disk_partition_path"];
		this._filesystemInformation = this._objParams["filesystem_information"];
	}

	/**
	 * Format the partition by choosing a filesystem type and block size.
	 * 
	 * @returns {ModuleResponse}
	 */
	async checkAndCreateFilesystem()
	{
		console.log("Starting disk partition format play...");

		const filesystemInformation = await FilesystemSingleton.getFilesystem(this._strDiskPartitionPath);

		if(FilesystemSingleton.compareFilesystems(filesystemInformation, this._filesystemInformation))
		{
			console.log("Partition " + this._strDiskPartitionPath + " is already formated as expected.");
		}
		else if(filesystemInformation !== null)
		{
			throw new Error("Partition " + this._strDiskPath + " has already been formatted and its filesystem differs from the expected one.");
		}
		else
		{
			console.log("Will format partition with the given details.");
			this.addStep(this.createFilesystem.bind(this));
			this.addStep(this.checkFilesystem.bind(this));
		}
	}

	/**
	 * Formats the partition.
	 * 
	 * @returns {ModuleResponse}
	 */
	async createFilesystem()
	{
		const moduleResponse = await FilesystemSingleton.createFilesystem(this._strDiskPartitionPath, this._filesystemInformation);

		console.log("Formated disk partition " + this._strDiskPartitionPath + " using " + this._filesystemInformation.filesystemType +
			" filesystem and " + this._filesystemInformation.blockSize + " block size.");

		return moduleResponse;
	}

	/**
	 * Checks if formating operation was successful.
	 */
	async checkFilesystem()
	{
		const bPartitionIsFormated = await FilesystemSingleton.hasFilesystem(this._strDiskPartitionPath, this._filesystemInformation);

		if(!bPartitionIsFormated)
		{
			throw new Error("Partition " + this._strDiskPartitionPath + " has not been successfully formated.");
		}
	}

	/**
	 * Checks if filesystem type is allowed or not.
	 * 
	 * @param {string} strFilesystemType
	 * @returns {boolean}
	 */
	_checkFilesystemType(strFilesystemType)
	{
		return this._arrAllowedFilesystemTypes.includes(strFilesystemType);
	}
}

module.exports = DiskPartitionFormatPlay;
