const BasePlay = require("./BasePlay");
const PartitionTable = require("../Models/PartitionTable");
const ModuleResponse = require("../Modules/ModuleResponse");
const DiskSingleton = require("../Modules").DiskSingleton;
const assert = require("assert");

/**
 * @TODO: Make it retriable.
 */
class DiskPartitionPlay extends BasePlay
{
	/**
	 * 	The play receives disk path and partition table.
		Disk path is obtained from /dev/disk/by-path for drives which are logged in via iSCSI.
		Steps:
			1.If disk has partitions, check if it has the expected partition table. If not, throw error.
				For the moment, we check if partition type is gpt and if there is one partition. In the future, more functionality
				will be added as described here https://gitlab.bigstep.com/bsi-agent/BSIAgent/merge_requests/32#note_3668

				If there are is no partition table found on disk, then:
				2. Partition drive using GPT standard
				3. Make 1 partition spanning the entire drive
	 */
	constructor()
	{
		super();
		
		this.addStep(this.setParams.bind(this));
		this.addStep(this.checkAndCreatePartitionTable.bind(this));
	}

	/**
	 * Params are validated inside the objects when they are created
	 */
	validateParams()
	{
		assert(typeof this._objParams["disk_path"] === "string", "Disk path must be set and must be a string.");
		assert(this._objParams["partition_table"] instanceof PartitionTable, "Partition table object is not of type PartitionTable.");

		//Additional checks for this play
		if(this._objParams["partition_table"].type !== "gpt")
		{
			throw new Error("Partition table must be of type gpt.");
		}

		if(this._objParams["partition_table"].nPartitionCount > 1)
		{
			throw new Error("For the moment, partition table can have a maximum of one partition.");
		}
	}

	/**
	 * Sets parameters for the play after validation
	 */
	setParams()
	{
		this._strDiskPath = this._objParams["disk_path"];
		this._partitionTable = this._objParams["partition_table"];
	}

	/**
	 * Checks if there is a single partition formated with the given filesystem type and block size.
	 * In the future, more functionality will be added as described here https://gitlab.bigstep.com/bsi-agent/BSIAgent/merge_requests/32#note_3668
	 * 
	 * @returns {ModuleResponse}
	 */
	async checkAndCreatePartitionTable()
	{
		console.log("Starting disk partition play...");

		const partitionTable = await DiskSingleton.getPartitionTable(this._strDiskPath);

		if(DiskSingleton.comparePartitionTables(partitionTable, this._partitionTable))
		{
			console.log("Disk is already partitioned the expected way. Partition table already exists and is the same as the given one.");
		}
		else if(partitionTable !== null && partitionTable.partitionCount !== 0)
		{
			throw new Error("Disk " + this._strDiskPath + " has already been partitioned and its partition table differs from the expected one.");
		}
		else
		{
			console.log("No partition table found on disk. Will partition disk with the partition table provided.");
			this.addStep(this.createPartitionTable.bind(this));
			this.addStep(this.createPartitions.bind(this));
			this.addStep(this.checkPartitionTable.bind(this));
		}
	}

	/**
	 * Partition disk using GPT standard.
	 * 
	 * @returns {ModuleResponse}
	 */
	async createPartitionTable()
	{
		const moduleResponse = await DiskSingleton.createPartitionTable(this._strDiskPath, this._partitionTable);

		console.log("Partitioned disk " + this._strDiskPath + " using GPT standard.");
		return moduleResponse;
	}

	/**
	 * Create partitions for the disk.
	 * For the moment, create a single partition spanning the entire disk.
	 * 
	 * @returns {ModuleResponse}
	 */
	async createPartitions()
	{
		let nPartitionIndex;
		let bIsChanged = true, strOutput = null, strError = null;

		for(nPartitionIndex = 0; nPartitionIndex < this._partitionTable.partitionCount; nPartitionIndex++)
		{
			const partition = this._partitionTable.getPartition(nPartitionIndex);
			const currentModuleResponse = await DiskSingleton.createPartition(this._strDiskPath, partition);

			bIsChanged = bIsChanged || currentModuleResponse.isChanged();
			strOutput += currentModuleResponse.getOutput();
			strError += currentModuleResponse.getError();

			if(currentModuleResponse.getError() !== null)
			{
				console.log("Failed to create partition " + partition.index + " for disk " + this._strDiskPath + " with error: "
							+ currentModuleResponse.getError());
				break;
			}

			console.log("Created partition " + partition.index + " for disk " + this._strDiskPath + ".");
		}

		return new ModuleResponse(bIsChanged, strOutput, strError);
	}

	async checkPartitionTable()
	{
		const bSuccesssfulPartitioning = await DiskSingleton.hasPartitionTable(this._strDiskPath, this._partitionTable);

		if(bSuccesssfulPartitioning)
		{
			console.log("Successfully partitioned disk " + this._strDiskPath);
		}
		else
		{
			throw new Error("Disk " + this._strDiskPath + " has not been partitioned correctly with the given partition table!");
		}
	}
}

module.exports = DiskPartitionPlay;
