const fs = require("fs-extra");
const BasePlay = require("./BasePlay");
const ISCSIDConfiguration = require("../Models/ISCSIDConfiguration");
const ISCSIInitiatorName = require("../Models/ISCSIInitiatorName");
const Lineinfile = require("../Models/Lineinfile");
const readLastLines = require("read-last-lines");
const assert = require("assert");

const ISCSIInitiatorNameTemplateSingleton = require("../Modules").ISCSIInitiatorNameTemplateSingleton;
const ServiceManagerSingleton = require("../Modules").ServiceManagerSingleton;
const CommandSingleton = require("../Modules").CommandSingleton;
const FileSingleton = require("../Modules").FileSingleton;

class ISCSISetupAndLoginPlay extends BasePlay
{
	constructor()
	{
		super();
		
		//These values are hardcoded for this play
		this._strInitiatorIQNFile = "/etc/iscsi/initiatorname.iscsi";
		this._strISCSIDConfFile = "/etc/iscsi/iscsid.conf";

		this.addStep(this.setParams.bind(this));
		this.addStep(this.updateInitiatorIQN.bind(this));
		this.addStep(this.updateISCSIDConfigurationFile.bind(this));
		this.addStep(this.rescanSessions.bind(this));
		this.addStep(this.loginToTargets.bind(this));
	}

	/**
	 * Params are validated inside the objects when they are created
	 */
	validateParams()
	{
		assert(this._objParams["node_information"] instanceof ISCSIDConfiguration, "Node information object is not of type ISCSIDConfiguration.");
		assert(this._objParams["iscsi_initiator_name"] instanceof ISCSIInitiatorName, "Node information object is not of type ISCSIInitiatorName.");
	}

	/**
	 * Sets parameters for the play after validation
	 */
	setParams()
	{
		this._objNodeInformation = this._objParams["node_information"];
		this._objISCSIInitiatorName = this._objParams["iscsi_initiator_name"];
	}

	/**
	 * Writes in /etc/iscsi/initiatorname.iscsi the initiator IQN
	 * 
	 * @returns {boolean}
	 */
	async updateInitiatorIQN()
	{
		console.log("Starting the iSCSI setup and login play...");

		const moduleResponse = await ISCSIInitiatorNameTemplateSingleton.renderToFile(
			this._strInitiatorIQNFile,
			this._objISCSIInitiatorName
		);

		console.log("Updated " + this._strInitiatorIQNFile + " with the initiator IQN " + this._objISCSIInitiatorName.ISCSIInitiatorIQN + ".");
		let bIsChanged = false;
		bIsChanged = bIsChanged || moduleResponse.isChanged();

		//If something changed, restart ISCSI daemon
		if(bIsChanged)
		{
			await this._restartISCSID();
		}

		return bIsChanged;
	}

	/**
	 * Restart the ISCSI daemon
	 */
	async _restartISCSID()
	{
		await ServiceManagerSingleton.restart("iscsid");
		console.log("Restarted iSCSI daemon.");
	}

	/**
	 * Rescan iSCSI session to find existing drives.
	 * 
	 * @returns {ModuleResponse}
	 */
	async rescanSessions()
	{
		const strCommand = "iscsiadm -m session --rescan";

		//Skip whitelist check
		const objOptions = {
			skipWhitelistCheck: true,
			shell: true
		};

		const executeResponse = await CommandSingleton.execute(strCommand, objOptions);

		if(executeResponse.getError() === null)
		{
			console.log("Rescanned the ISCSI session with iscsiadm.");
		}
		else
		{
			console.log("iscsiadm rescan command failed with: " + executeResponse.getError());
		}

		return executeResponse;
	}

	/**
	 * Updates /etc/iscsi/iscsid.conf with iSNS and ISCSI information from the node.
	 * 
	 * @returns {boolean}
	 */
	async updateISCSIDConfigurationFile()
	{
		//Static info for this play
		const objISCSIDConfigurationInput = [
			new Lineinfile(
				this._strISCSIDConfFile,
				"isns.address = " + this._objNodeInformation.iSNSAdress,
				"^isns.address"
			),
			new Lineinfile(
				this._strISCSIDConfFile,
				"isns.port = " + this._objNodeInformation.iSNSPort,
				"^isns.port"
			),
			new Lineinfile(
				this._strISCSIDConfFile,
				"node.session.auth.authmethod = " + this._objNodeInformation.ISCSIAuthMethod,
				"^node.session.auth.authmethod"
			),
			new Lineinfile(
				this._strISCSIDConfFile,
				"node.session.auth.username = " + this._objNodeInformation.ISCSIUsername,
				"^node.session.auth.username"
			),
			new Lineinfile(
				this._strISCSIDConfFile,
				"node.session.auth.password = " + this._objNodeInformation.ISCSIPassword,
				"^node.session.auth.password"
			),
			new Lineinfile(
				this._strISCSIDConfFile,
				"node.session.nr_sessions = " + this._objNodeInformation.nrSessions,
				"^node.session.nr_sessions"
			),
			new Lineinfile(
				this._strISCSIDConfFile,
				"node.startup = " + this._objNodeInformation.startupMode,
				"^node.startup"
			)
		];

		//Check if something has changed in the ISCSI daemon configuration file
		let bIsChanged = false;

		//Create an empty file
		if (!fs.existsSync(this._strISCSIDConfFile))
		{
			await FileSingleton.save("", this._strISCSIDConfFile);
		}

		for(const ISCSIDConfiguration of objISCSIDConfigurationInput)
		{
			const moduleResponse = await FileSingleton.addLine(
				ISCSIDConfiguration.line,
				ISCSIDConfiguration.path,
				ISCSIDConfiguration.regexp
			);

			bIsChanged = bIsChanged || moduleResponse.isChanged();
		}

		//Read the last line and check if it's empty (last line in /etc/iscsi/iscsid.conf must be empty, otherwise login command might fail)
		const strLastLine = await readLastLines.read(this._strISCSIDConfFile, 1);

		//Add empty line if last one is not
		if(!/^\s*$/.test(strLastLine))
		{
			await FileSingleton.addLine("", this._strISCSIDConfFile);
			bIsChanged = true;
		}

		console.log("Updated " + this._strISCSIDConfFile + " with iSNS and ISCSI information from the node.");

		//If something changed, restart ISCSI daemon
		if (bIsChanged)
		{
			await this._restartISCSID();
		}

		return bIsChanged;
	}

	/**
	 * Execute the shell command to login to targets and import disks.
	 * 
	 * @returns {ModuleResponse}
	 */
	async loginToTargets()
	{
		const strCommand = "iscsiadm --mode discovery --type isns -p " + this._objNodeInformation.iSNSAdress
			+ ":" + this._objNodeInformation.iSNSPort + " --login --interface=default";

		//Skip whitelist check
		const objOptions = {
			skipWhitelistCheck: true,
			shell: true
		};

		const executeResponse = await CommandSingleton.execute(strCommand, objOptions);

		if(executeResponse.getError() === null)
		{
			console.log("Imported disks by executing the iscsiadm login command.");
		}
		else
		{
			console.log("iscsiadm login command failed with: " + executeResponse.getError());
		}

		return executeResponse;
	}
}

module.exports = ISCSISetupAndLoginPlay;
