const path = require("path");
const fs = require("fs-extra");
const ChildProcess = require("child-process-promise");

const oneMinuteSeconds = 60;
const oneMinuteMilliseconds = oneMinuteSeconds * 1000;
const strAgentVersionFileName = "agentversion.json";

module.exports =
	class Service
	{
		/**
		 * @param {boolean} bRunningAsService
		 * @param {BSIWebSocketClientManager} bsiWebSocketClientManager
		 * @param {StateHandler} stateHandler
		 * @param {ModuleCallbackRunner} callbackRunner
		 * @param {FactsProvider} factsProvider
		 */
		constructor(bRunningAsService, bsiWebSocketClientManager, stateHandler, callbackRunner, factsProvider)
		{
			this.bRunningAsService = bRunningAsService;

			this._startTimeUTC = new Date();
			this._bsiWebSocketClientManager = bsiWebSocketClientManager;
			this._stateHandler = stateHandler;
			this._callbackRunner = callbackRunner;
			this._factsProvider = factsProvider;

			process.on(
				"SIGINT",
				() =>
				{
					console.log("Received SIGINT. Attempting to exit cleanly...");

					this.stop();
					process.exit(0);
				}
			);
		}

		/**
		 * Simple method to show whoever is interested that this service is alive.
		 * For the moment, it updates a simple `alive` file for other processes to check.
		 *
		 * @returns {Promise<void>}
		 */
		async _checkSanity()
		{
			let bIsSanityPassed = false;
			const currentTime = new Date();
			const timeSinceStart = (currentTime - this._startTimeUTC) / 1000; //seconds.
			if (timeSinceStart > oneMinuteSeconds && timeSinceStart <= 2 * oneMinuteSeconds)
			{
				const strAgentRootFolder = path.dirname(path.dirname(path.dirname(__filename)));
				const strWorkspaceFolder = path.dirname(path.dirname(strAgentRootFolder));
				const strSettingsDirectoryPath = this._bsiWebSocketClientManager.agentIdentity.settingsDirectoryPath;
				const strAgentVersionFilePath = path.join(strSettingsDirectoryPath, strAgentVersionFileName);
				const strAgentVersion = await this._bsiWebSocketClientManager.agentVersion();

				const strAgentVersionContent = await fs.readFile(strAgentVersionFilePath, "utf8");
				const objAgentVersionFile = JSON.parse(strAgentVersionContent);

				objAgentVersionFile.agentVersions.forEach(
					(BSIAgentEntry) =>
					{
						if (BSIAgentEntry.version === strAgentVersion)
						{
							BSIAgentEntry.sanity = "passed";
							bIsSanityPassed = true;
						}
					}
				);

				const strAgentVersionFile = JSON.stringify(objAgentVersionFile, undefined, "\t");
				await fs.writeFile(strAgentVersionFilePath, strAgentVersionFile);
				//TODO: fix for windows.
				if (bIsSanityPassed && process.platform !== "win32")
				{
					const strBSIAgentUtilityPath = path.join(
						strWorkspaceFolder,
						strAgentVersion,
						"BSIAgent",
						"scripts",
						"linux",
						"install_agent_utility.sh"
					);
					try
					{
						await ChildProcess.exec("systemctl stop BSIAgentUtility");
						await ChildProcess.exec("systemctl disable BSIAgentUtility");
						await ChildProcess.execFile(strBSIAgentUtilityPath);
					}
					catch (error)
					{
						console.error(error);
						await fs.appendFile(path.join(strWorkspaceFolder, "update_log_error"), error + "\n");
					}
				}
			}
		}

		/**
		 * @returns {null}
		 */
		async updateAgentVersionFile()
		{
			// Check if agentversion.json exists.
			// If if doesn't, create a minimal one.
			// Else do nothing => wait for the update call.
			const strSettingsDirectoryPath = this._bsiWebSocketClientManager.agentIdentity.settingsDirectoryPath;
			const strAgentVersionFilePath = path.join(strSettingsDirectoryPath, strAgentVersionFileName);

			const bExists = await fs.pathExists(strAgentVersionFilePath);
			if (!bExists)
			{
				const strAgentVersion = await this._bsiWebSocketClientManager.agentVersion();
				const strTimeNowISO = new Date().toISOString();
				const objAgentVersionFile = {
					agentVersions: [
						{
							version: strAgentVersion,
							started: strTimeNowISO,
							sanity: "unknown", // will be changed to failed by `agent utility` or passed by `agent`
							running: true
						}
					]
				};
				const strAgentVersionFile = JSON.stringify(objAgentVersionFile, undefined, "\t");
				await fs.writeFile(strAgentVersionFilePath, strAgentVersionFile);
			}
			else
			{
				// Change the started time.
				const strAgentVersion = await this._bsiWebSocketClientManager.agentVersion();
				const strTimeNowISO = new Date().toISOString();

				const strAgentVersionFile = await fs.readFile(strAgentVersionFilePath, "utf8");
				const objAgentVersionFile = JSON.parse(strAgentVersionFile);

				objAgentVersionFile.agentVersions.forEach(
					(BSIAgentEntry) =>
					{
						if (BSIAgentEntry.version === strAgentVersion)
						{
							if (BSIAgentEntry.started == null || BSIAgentEntry.started === "not started")
							{
								BSIAgentEntry.started = strTimeNowISO;
							}
							BSIAgentEntry.sanity = "unknown";
							BSIAgentEntry.running = true;
						}
						else
						{
							BSIAgentEntry.running = false;
						}
					}
				);

				const strNewAgentVersionFile = JSON.stringify(objAgentVersionFile, undefined, "\t");
				await fs.writeFile(strAgentVersionFilePath, strNewAgentVersionFile);
			}
		}

		/**
		 * @returns {undefined}
		 */
		async run()
		{
			if (this.bRunningAsService)
			{
				throw new Error("Not implemented.");
			}
			else
			{
				console.log("Running, not as a service.");
			}


			await this._handleStartupCallbacks(); // run startup calllbacks
			const fnBSIClient = await this._bsiWebSocketClientManager.bsiClientSingletonFunction();

			this._nRunningIntervalID = setInterval(
				async() =>
				{
					// Three purposes:
					// - keep this process alive;
					// - keep connection open to BSI (so BSI may call methods of this app);
					// - protect WebSocket TCP connection from load balancer timeouts.
					await (await fnBSIClient()).keep_alive();
				},
				15 * 1000
			);

			this._nCheckSanityIntervalID = setInterval(
				async() =>
				{
					await this._checkSanity();
				},
				oneMinuteMilliseconds
			);

			await this.updateAgentVersionFile();

			await this._onConnect();
		}

		/**
		 * @private
		 */
		async _onConnect()
		{
			await this._checkOSInfo();

			this._nCheckOSInfoIntervalID = setInterval(
				async() => await this._checkOSInfo(),
				15 * oneMinuteMilliseconds
			);
		}

		/**
		 * Clears the empty setInterval which allows the process to exit gracefully
		 * if it has nothing else to do.
		 *
		 * @returns {undefined}
		 */
		stop()
		{
			this._clearInterval(this._nRunningIntervalID);
			this._clearInterval(this._nCheckSanityIntervalID);
			this._clearInterval(this._nCheckOSInfoIntervalID);
		}

		/**
		 * @param {number} nInterval
		 * @private
		 */
		_clearInterval(nInterval)
		{
			clearInterval(nInterval);
			nInterval = null;
		}

		/**
		 * @private
		 */
		async _checkOSInfo()
		{
			try
			{
				const fnBSIClient = await this._bsiWebSocketClientManager.bsiClientSingletonFunction();
				const osInfo = await this._factsProvider.getOSInfo();
				const savedState = await this._stateHandler.get();
				const savedOSInfo = savedState.OSInfo;

				if (!savedOSInfo || !osInfo.isEqual(savedOSInfo))
				{
					await (await fnBSIClient()).agent_info_update(this._bsiWebSocketClientManager.agentIdentity.agentID, osInfo);
					savedState.OSInfo = osInfo;
					await this._stateHandler.save(savedState);
					console.log("OS Info updated!");
				}
				else 
				{
					console.log("OS Info NOT changed!");
				}
			}
			catch (error)
			{
				console.error(`OS Info could NOT be updated: ${error.message}`);
			}
		}
		/**
		 * @returns {boolean}
		 * @private
		 */
		async _handleStartupCallbacks()
		{
			console.log("Running startup callbacks...");
			const current = await this._stateHandler.get();

			for (const nCallbackKey in current.startupCallbacks)
			{
				/**
				 * @var callback ModuleCallback
				 */
				const callback = current.startupCallbacks[nCallbackKey];
				await this._callbackRunner.run(callback);
				if (!callback.persistent) {
					current.removeStartupCallback(nCallbackKey);
					await this._stateHandler.save(current);
				}
			}

			current.rebootRequestedDate = null;
			await this._stateHandler.save(current);

			console.log("Startup callbacks finished!");

			return true;
		}
	};
