const DownloadTask = require("publisher").DownloadTask;
const TarGZArchiver = require("publisher").TarGZArchiver;
const ZipArchiver = require("publisher").ZipArchiver;
const CRCHash = require("publisher").CRCHash;
const BSIWebSocketClientManager = require("./BSIWebSocketClientManager");
const path = require("path");
const fs = require("fs-extra");
const jskippyValidators = require("validators");
const json = require("json-promise");
const os = require("os");
const validator = require("validator");
const url = require("url");
const loggedChildProcessSingleton = require("../LoggedChildProcessSingleton");

class Updater
{
	/**
	 * @param {AgentIdentity} agentIdentity
	 */
	constructor(agentIdentity)
	{
		this._agentIdentity = agentIdentity;
		this._strSettingsDirectoryPath = agentIdentity.strSettingsDirectoryPath;
	}

	/**
	 * @param  {string} strUpdateURL
	 *
	 * @returns {bool}
	 */
	async isTrustedDomain(strUpdateURL)
	{
		const urlObject = url.parse(strUpdateURL);
		const strHost = urlObject.hostname;

		if (validator.isIP(strHost)) // check if the ip is known if the host looks like ipv4 / ipv6 pattern
		{
			const arrAllowedHosts = BSIWebSocketClientManager.knownRepositoryEndpointRoots[this._agentIdentity.environment];
			for (const strAllowedHost of arrAllowedHosts)
			{
				if (strHost === strAllowedHost)
				{
					return true;
				}
			}

			return false;
		}

		return jskippyValidators.DomainName.DomainNameValidator.hasRootDomain(strHost, ["bigstepcloud.com"]);
	}

	/**
	 * @param  {string} strFilepath
	 * @param  {string} strAttribute
	 *
	 * @returns {string}
	 */
	async getJSONAttributeValue(strFilepath, strAttribute)
	{
		const packageContent = await fs.readFile(strFilepath, {encoding: "utf8"});
		const parsedObject = await json.parse(packageContent);
		const attributeValue = parsedObject[strAttribute];
		if (attributeValue != null)
		{
			return attributeValue;
		}
		return new Error(strAttribute + " not found in file.");
	}

	/**
	 * @param {string} strNewVersion
	 * @private
	 */
	async _addAgentVersion(strNewVersion)
	{
		const strAgentVersionFilePath = path.join(this._strSettingsDirectoryPath, strNewVersion);

		let strAgentVersionFile = await fs.readFile(strAgentVersionFilePath);
		let objAgentVersions = JSON.parse(strAgentVersionFile);
		objAgentVersions.agentVersions.push({
			version: strNewVersion,
			started: "not started",
			sanity: "unknown",
			running: false
		});

		strAgentVersionFile = JSON.stringify(objAgentVersions, undefined, "\t");
		fs.writeFile(strAgentVersionFilePath, strAgentVersionFile);
	}

	/**
	 * @param {string} strNextAgentVersionURL
	 * @param {Object} objAgentData
	 *
	 * @returns {string}
	 */
	_getAgentURL(strNextAgentVersionURL, objAgentData)
	{
		let strPlatform = process.platform;
		if (strPlatform === "win32")
		{
			strPlatform = "win";
		}

		return [strNextAgentVersionURL, objAgentData["defaultVersion"], strPlatform, objAgentData["OSMetadata"][process.arch]].join("/");
	}

	/**
	 * @param {string} strRepoURL
	 * @param {string} strWorkspaceFolder
	 *
	 * @returns {Object}
	 */
	async _getAgentUpdateData(strRepoURL, strWorkspaceFolder)
	{
		let strAgentVersionURL = strRepoURL + "/agentVersions.json";
		let downloadTask = new DownloadTask();
		let strDownloadFileName = downloadTask.getFileNameFromUrl(strAgentVersionURL);
		let strDownloadedFilePath = path.join(strWorkspaceFolder, strDownloadFileName);

		if (await fs.exists(strDownloadedFilePath))
		{
			await fs.unlink(strDownloadedFilePath);
		}

		await downloadTask.downloadItem(strAgentVersionURL, strWorkspaceFolder);
		let strAgentVersionFile = await fs.readFile(strDownloadedFilePath);
		let objAgentVersionFile = JSON.parse(strAgentVersionFile);

		let strDefaultVersion = objAgentVersionFile["defaultVersion"];
		let strPlatform = process.platform;
		if (strPlatform === "win32")
		{
			strPlatform = "win";
		}

		let strNextAgentVersionURL =
			strAgentVersionURL.substring(0, strAgentVersionURL.lastIndexOf("/"))
			+ objAgentVersionFile[strDefaultVersion][strPlatform];

		await fs.unlink(strDownloadedFilePath);

		downloadTask = new DownloadTask();
		await downloadTask.downloadItem(strNextAgentVersionURL, strWorkspaceFolder);
		strDownloadFileName = downloadTask.getFileNameFromUrl(strNextAgentVersionURL);
		strDownloadedFilePath = path.join(strWorkspaceFolder, strDownloadFileName);
		strAgentVersionFile = await fs.readFile(strDownloadedFilePath);
		objAgentVersionFile = JSON.parse(strAgentVersionFile);

		await fs.unlink(strDownloadedFilePath);

		return {"OSMetadata": objAgentVersionFile, "defaultVersion": strDefaultVersion};
	}

	/**
	 * @param {string} strUpdateURL
	 *
	 * @returns {string}
	 */
	async update(strUpdateURL)
	{
		console.log("Try to self update agent...");

		if (!await this.isTrustedDomain(strUpdateURL))
		{
			console.error(new Error("Domain not trusted"));
		}

		if (os.platform() === "linux")
		{
			await this.updateLinux(strUpdateURL);
		}
		else
		{
			throw new Error("Not implemented");
		}

		return "Agent downloaded and extracted";
	}

	/**
	 * @param {string} strUpdateURL
	 */
	async updateLinux(strUpdateURL)
	{
		const strAgentVersionFolder = path.dirname(path.dirname(path.dirname(__filename)));
		const strWorkspaceFolder = path.dirname(path.dirname(strAgentVersionFolder));

		const objAgentData = await this._getAgentUpdateData(strUpdateURL, strWorkspaceFolder);
		const strAgentArchiveURL = this._getAgentURL(strUpdateURL, objAgentData);

		const strSourceCRC = objAgentData["OSMetadata"][process.arch + "_crc"];

		// Download the file.
		const downloadTask = new DownloadTask();
		await downloadTask.downloadItem(strAgentArchiveURL, strWorkspaceFolder);
		let strDownloadFileName = downloadTask.getFileNameFromUrl(strAgentArchiveURL);
		let strDownloadedFilePath = path.join(strWorkspaceFolder, strDownloadFileName);

		const bIsZipArchive = strDownloadFileName.indexOf(".zip") > -1;
		const archiver = bIsZipArchive ? new ZipArchiver() : new TarGZArchiver();
		await fs.ensureDir(path.join(strWorkspaceFolder, "NewAgent"));

		const tmpAgentFolderPath = bIsZipArchive
			? path.join(strWorkspaceFolder, "NewAgent")
			: path.join(strWorkspaceFolder, "NewAgent", "BSIAgent");

		await archiver.decompress(strDownloadedFilePath, tmpAgentFolderPath);

		const crcHasher = new CRCHash();
		const strDownloadedContentCRC = await crcHasher.generate(tmpAgentFolderPath);

		if (strDownloadedContentCRC !== strSourceCRC)
		{
			throw new Error(`CRC hash found on repo: ${strSourceCRC} does not match downloaded content CRC: ${strDownloadedContentCRC}`);
		}

		await fs.remove(strDownloadedFilePath);
		const strNewVersion = await this.getJSONAttributeValue(
			path.join(strWorkspaceFolder, "NewAgent", "BSIAgent", "package.json"),
			"version"
		);

		const strAgentVersion = await this.getJSONAttributeValue(
			path.join(strAgentVersionFolder, "package.json"),
			"version"
		);

		if (strAgentVersion === strNewVersion)
		{
			await fs.remove(path.join(strWorkspaceFolder, "NewAgent"));
			throw new Error("Cannot update to the same version");
		}

		const newPath = path.join(strWorkspaceFolder, strNewVersion, "BSIAgent");
		await fs.remove(path.join(strWorkspaceFolder, strNewVersion)); // in case of older install try
		await fs.ensureDir(path.join(strWorkspaceFolder, strNewVersion));
		await fs.move(tmpAgentFolderPath, newPath);
		await fs.remove(path.join(strWorkspaceFolder, "NewAgent"));
		
		// install new agent
		await loggedChildProcessSingleton.spawn(`sh ${newPath}/scripts/linux/install_agent.sh`, {shell: true});

		// Update the agentversion.json with the new version.
		await this._addAgentVersion(strNewVersion);
	}
};

module.exports = Updater;
