const fs = require("fs-extra");
const path = require("path");
const ChildProcess = require("child-process-promise");
const pathExists = require("path-exists");
const assert = require("assert");

module.exports =
class AgentHealthUpdate
{
	/**
	 */
	constructor()
	{
		this._objAgentVersionFile = null;
		this._strProgramDataAgentFolder = "BSIDeployAgent";
		this._strAgentVersionFileName = "agentversion.json";
		this._nOneMinuteSeconds = 60;
		this._strAgentRootFolder = path.dirname(path.dirname(__filename));
		this._strWorkspaceFolder = path.dirname(path.dirname(this._strAgentRootFolder));
	}

	/**
	* @returns {string}
	*/
	async _getProgramDataDirectoryPath()
	{
		let strProgramDataDirectoryPath;

		if(process.platform === "win32")
		{
			assert(process.env.ALLUSERSPROFILE, "ALLUSERSPROFILE environment variable not set.");
			assert((await pathExists(process.env.ALLUSERSPROFILE)), "ALLUSERSPROFILE=" + JSON.stringify(process.env.ALLUSERSPROFILE) + " path not found.");

			strProgramDataDirectoryPath = process.env.ALLUSERSPROFILE;
		}
		else
		{
			assert(await pathExists("/etc"), "/etc not found. Current platform: " + JSON.stringify(process.platform) + ".");
			strProgramDataDirectoryPath = "/etc";
		}

		return path.join(strProgramDataDirectoryPath, this._strProgramDataAgentFolder);
	}

	/**
	 * @param  {string} strStartTimeUTC
	 */
	async checkSanity(strStartTimeUTC)
	{
		let bRevert = false;
		const timeNowUTC = new Date();
		const startTimeUTC = new Date(strStartTimeUTC);

		const strAgentVersionFilePath = path.join(await this._getProgramDataDirectoryPath(), this._strAgentVersionFileName);
		let strAgentVersionFile = await fs.readFile(strAgentVersionFilePath);
		const objAgentVersionFile = JSON.parse(strAgentVersionFile);

		for (let i = 0; i < objAgentVersionFile.agentVersions.length; i++)
		{
			let BSIAgentEntry = objAgentVersionFile.agentVersions[i];
			if (BSIAgentEntry.running === true && BSIAgentEntry.sanity === "unknown")
			{
				const timeAgentStarted = new Date(BSIAgentEntry.started);
				const timeSinceAgentStarted = (timeNowUTC - timeAgentStarted) / 1000;
				const timeSinceUtilityStarted = (timeNowUTC - startTimeUTC) / 1000;

				if (timeSinceAgentStarted > 2 * this._nOneMinuteSeconds && timeSinceUtilityStarted > this._nOneMinuteSeconds)
				{
					objAgentVersionFile.agentVersions[i].sanity = "failed";
					bRevert = true;
				}
			}
		}

		if (bRevert)
		{
			// Get the version to revert to.
			let strVersion = null;
			for (let i = 0; i < objAgentVersionFile.agentVersions.length; i++)
			{
				if (objAgentVersionFile.agentVersions[i].sanity === "passed")
				{
					strVersion = objAgentVersionFile.agentVersions[i].version;
				}
			}

			for (let i = 0; i < objAgentVersionFile.agentVersions.length; i++)
			{
				if (objAgentVersionFile.agentVersions[i].version === strVersion)
				{
					objAgentVersionFile.agentVersions[i].started = "not started";
				}
			}

			if (strVersion != null)
			{
				const strInstallScriptPath = path.join(
					this._strWorkspaceFolder,
					strVersion,
					"BSIAgent",
					"scripts",
					"linux",
					"install_agent.sh"
				);
				await ChildProcess.execFile(strInstallScriptPath);
				strAgentVersionFile = JSON.stringify(objAgentVersionFile, undefined, "\t");
				await fs.writeFile(strAgentVersionFilePath, strAgentVersionFile);
			}
		}
	}

	async checkForUpdate()
	{
		const strAgentVersionFile = await fs.readFile(path.join(await this._getProgramDataDirectoryPath(), this._strAgentVersionFileName));
		const objAgentVersionFile = JSON.parse(strAgentVersionFile);

		if (this._objAgentVersionFile == null)
		{
			this._objAgentVersionFile = objAgentVersionFile;
		}
		else
		{
			//Check if there is a new version added.
			if (this._objAgentVersionFile.agentVersions.length < objAgentVersionFile.agentVersions.length)
			{
				const strNewAgentVersion = objAgentVersionFile.agentVersions[objAgentVersionFile.agentVersions.length - 1].version;
				const strOSType = process.platform === "linux" ? "linux" : "windows";
				const strAgentScriptsFolder = path.join(
					this._strWorkspaceFolder,
					strNewAgentVersion,
					"BSIAgent",
					"scripts",
					strOSType
				);
				this._objAgentVersionFile = objAgentVersionFile;

				if (strOSType === "linux")
				{
					const strAgentInstallScriptPath = path.join(strAgentScriptsFolder, "install_agent.sh");
					const strAgentUtilityInstallScriptPath = path.join(strAgentScriptsFolder, "install_agent_utility.sh");
					const strUninstallScriptPath = path.join(strAgentScriptsFolder, "uninstall.sh");

					await fs.chmod(strAgentInstallScriptPath, "0755");
					await fs.chmod(strAgentUtilityInstallScriptPath, "0755");
					await fs.chmod(strUninstallScriptPath, "0755");

					await ChildProcess.execFile(strAgentInstallScriptPath);
				}
				else
				{
					//TODO: update nssm service.
				}
			}
		}
	}

	async logError(error)
	{
		const strTimeNowUTC = new Date().toISOString();
		await fs.appendFile(path.join(await this._getProgramDataDirectoryPath(), "error.log"), error + " " + strTimeNowUTC + "\r\n");
	}
};
