const CommandExecutor = require("../../CommandExecutor");
const os = require("os");

class CommandBase
{
	constructor({bMustBeFiltered = true, bMustBeParsed = true} = {})
	{
		this._strRawCommandOutput = null;
		this._objParsedCommandOutput = {};
		this._objFilteredCommandOutput = {};
		this._objLastCommandOutput = null;

		// If there is no need to filter or parse set _bFiltered and _bParsed to true
		this._bParsed = !bMustBeParsed;
		this._bFiltered = !bMustBeFiltered;

		this._strOSPlatform = this.constructor.osPlatform;
	}

	/**
	 * Executes the actual command and stores the output in order to be parsed and filtered
	 * 
	 * @param {Object} [objExecConfig]
	 * @returns {Promise<string|Object>}
	 */
	async execCommand(objExecConfig = {})
	{
		if(this._strOSPlatform !== os.platform())
		{
			throw new Error(`Wrong platform for command ${this.constructor.name}. It should be used on ${this._strOSPlatform}.`);
		}

		this._strRawCommandOutput = await CommandExecutor.execCommandProcess(this.constructor.command, objExecConfig);

		this._objLastCommandOutput = {};
		this.bExecuted = true;

		try
		{
			if(this._bParsed !== true)
			{
				this._objLastCommandOutput = await this._parseCommandOutput();
				this._bParsed = true;
			}
	
			if(this._bFiltered !== true)
			{
				if(this._objParsedCommandOutput)
				{
					throw new Error(`Empty parsed object for command ${this.constructor.name}.`);
				}

				this._objLastCommandOutput = await this._filterCommandOutput();
				this._bFiltered = true;
			}
		}
		catch(e)
		{
			console.error(e.stack);
		}

		return this._objLastCommandOutput;
	}

	/**
	 * Filters the output, must me runned after execCommand or parseCommandOutput
	 * 
	 * @param {Object} [objResult=null]
	 * @returns {Promise<Object>} _objFilteredCommandOutput
	 */
	async _filterCommandOutput(objResult = null)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * Parses the raw output from the command, must me runned after execCommand
	 * 
	 * @param {string} [strResult=null]
	 * @returns {Promise<Object>} _objParsedCommandOutput
	 */
	async _parseCommandOutput(strResult = null)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * The actual command that is run in the shell by a child process
	 * 
	 * @returns {string}
	 */
	static get command()
	{
		throw new Error("Must me implmented");
	}

	/**
	 * @returns {string}
	 */
	static get osPlatform()
	{
		throw new Error("Must be implemented");
	}

	get rawOutput()
	{
		return this._strRawCommandOutput;
	}

	get output()
	{
		return this._objLastCommandOutput;
	}

	async test()
	{
		throw new Error("Must be implmented");
	}
}

module.exports = CommandBase;
