// @ts-check

const CommandBase = require("./CommandBase");
const NVMeDiskDetection = require("./NVMeDiskDetection");

const Utils = require("jskippy-node").Utils;

class NVMe extends CommandBase
{
	constructor(objConfig = {})
	{
		super(objConfig);
	}

	static get command()
	{
		return "nvme list -o json";
	}

	/**
	 * Parses the raw output from the command, must me runned after execCommand
	 * 
	 * @param {string} [strResult=null]
	 * @returns {Promise<Object>} _objParsedCommandOutput
	 */
	async _parseCommandOutput(strResult = null)
	{
		if(strResult)
		{
			this._strRawCommandOutput = strResult;
		}

		if(!this._strRawCommandOutput || this._strRawCommandOutput.trim() === "")
		{
			throw new Error(`Empty output string for command ${this.constructor.name}.`);
		}
		
		this._objParsedCommandOutput = Utils.safeJSONParse(this._strRawCommandOutput);
		return this._objParsedCommandOutput;
	}

	/**
	 * Filters the output, must me runned after execCommand or parseCommandOutput
	 * 
	 * @param {Object} [objResult=null]
	 * @returns {Promise<Object>} _objFilteredCommandOutput
	 */
	async _filterCommandOutput(objResult = null)
	{
		if(objResult)
		{
			this._objParsedCommandOutput = objResult;
		}

		// Acquire information from NVMeDiskDetection command
		let objNVMeDiskDetectionResult = null;
		
		try
		{
			objNVMeDiskDetectionResult = await NVMeDiskDetection.execCommand(/*objConfig*/);
		}
		catch(e)
		{
			console.error(e.stack);
		}
		
		this._objFilteredCommandOutput = {};
		const arrDisks = [];

		if(this._objParsedCommandOutput.Devices && Array.isArray(this._objParsedCommandOutput.Devices))
		{
			const arrDevices = this._objParsedCommandOutput.Devices;

			arrDevices.forEach((objDiskInfo, nIndex) => 
			{
				const objDisk = {};

				if(typeof objDiskInfo.DevicePath === "string")
				{
					const strShortDeviceName = objDiskInfo.DevicePath.replace("/dev/", "");

					objDisk.server_disk_model = objDiskInfo.ModelNumber;
					objDisk.server_disk_serial = objDiskInfo.SerialNumber;
					objDisk.server_disk_device_name = strShortDeviceName;
					objDisk.device_path = objDiskInfo.DevicePath;

					if(objNVMeDiskDetectionResult)
					{
						const objNVMEDisk = objNVMeDiskDetectionResult[strShortDeviceName];

						objDisk.server_disk_size_gb = Math.round(objNVMEDisk.size / 1024 / 1024 / 1024);
						objDisk.server_disk_vendor = objNVMEDisk.vendor;
					}

					objDisk.server_disk_type = NVMe.DISK_TYPE_NVME;
				}

				arrDisks.push(objDisk);
			});

		}
		else
		{
			throw new Error("Missing or wrong type of _objParsedCommandOutput.Devices property.");
		}

		this._objFilteredCommandOutput.disks = arrDisks;

		return this._objFilteredCommandOutput;
	}

	static get osPlatform()
	{
		return "linux";
	}

	get parsedOutput()
	{
		return this._objParsedCommandOutput;
	}

	async test()
	{
		
	}

	static get DISK_TYPE_NVME()
	{
		return "NVME";
	}
}

module.exports = new NVMe({bMustBeFiltered: true, bMustBeParsed: true});
