// @ts-check

const CommandBase = require("./CommandBase");

class NetworkInterfaces extends CommandBase
{
	constructor(objConfig = {})
	{
		super(objConfig);
	}

	static get command()
	{
		return `
			IFACES=""
			for f in /sys/class/net/*
			do
				if [ $f == "/sys/class/net/lo" ]; then continue; fi
				ifindex=\`cat $f/ifindex\`
				ifindex=$((ifindex -1))
				mac=\`cat $f/address\`
				IFACES="\${IFACES}__BSI__INTERFACE \$ifindex \$mac\n"
				name=\`echo $f | sed "s/\\\/sys\\\/class\\\/net\\\///g\"\`
				speed=\`ethtool \${name} | grep Speed\`
				IFACES="\${IFACES}\${speed}\n"
				/usr/sbin/lldptool -L -i \${name} adminStatus=rxtx
				sleep 10s
				IFACES="\${IFACES}__START_LLDP_INFO"
				LLDP=\`/usr/sbin/lldptool -t -i \${name} -n | sed 's/'"\\\$(printf '\\\\015')"'//g'\`
				if [ -z "$LLDP" ]; then
					sleep 10s
					echo "Retried LLDP on an interface once"
					LLDP=\`/usr/sbin/lldptool -t -i \${name} -n | sed 's/'"\\\$(printf '\\\\015')"'//g'\`
					if [ -z "$LLDP" ]; then
						sleep 10s
						echo "Retried LLDP on an interface twice"
						LLDP=\`/usr/sbin/lldptool -t -i \${name} -n | sed 's/'"\\\$(printf '\\\\015')"'//g'\`
					fi
				fi
				IFACES="\${IFACES}\${LLDP}"
				IFACES="\${IFACES}__END_INTERFACE_INFO"
			done
		
			echo \${IFACES}
		`;
	}

	/**
	 * Parses the raw output from the command, must me runned after execCommand
	 * 
	 * @param {string} [strResult=null]
	 * @returns {Promise<Object>} _objParsedCommandOutput
	 */
	async _parseCommandOutput(strResult = null)
	{
		if(strResult)
		{
			this._strRawCommandOutput = strResult;
		}

		if(!this._strRawCommandOutput || this._strRawCommandOutput.trim() === "")
		{
			throw new Error(`Empty output string for command ${this.constructor.name}.`);
		}
		
		const objParsedResult = {};
		const arrNetworkRows = this._strRawCommandOutput.split("__END_INTERFACE_INFO");
		const objInterfaceInformation = {};

		objInterfaceInformation.interfaces = [];
		objParsedResult.server_network_total_capacity_mbps = 0;

		arrNetworkRows.forEach((strInterfaceRow, nIndex) =>
		{
			if(strInterfaceRow.includes("__BSI__INTERFACE"))
			{
				const arrInformation = strInterfaceRow.split("__START_LLDP_INFO");
				const objLLDPInformation = this._parseLLDPInformation(arrInformation[1]);
				objInterfaceInformation.interfaces[nIndex] = objLLDPInformation;

				const arrHeaderInformation = arrInformation[0].split("Speed: ");
				const arrBasicInformation = arrHeaderInformation[0].split(" ");
				objInterfaceInformation["server_interface_index"] = parseInt(arrBasicInformation[1], 10) - 1;
				objInterfaceInformation["server_interface_mac_address"] = arrBasicInformation[2].trim();

				const strSpeed = arrHeaderInformation[1];
				const arrSpeeds = strSpeed.split("Mb/s");
				let nCurrentNetworkCapacity = null;

				if(arrSpeeds.length === 2)
				{
					nCurrentNetworkCapacity = parseInt(arrSpeeds[0], 10);
				}
				else
				{
					nCurrentNetworkCapacity = 0;
				}

				objParsedResult["server_network_total_capacity_mbps"] += nCurrentNetworkCapacity;
			}
		});
		
		this._objParsedCommandOutput = {...objParsedResult, ...objInterfaceInformation};
		return this._objParsedCommandOutput;
	}

	/**
	 * Extract the necessary LLDP information and put it in an object
	 * 
	 * @param {string} strLLDPInformation
	 * @returns {Object}
	 */
	_parseLLDPInformation(strLLDPInformation)
	{
		strLLDPInformation = strLLDPInformation.replace("/\n\t|\n /", "\t");
		const arrLLDPRows = strLLDPInformation.split("\n");

		const objRequiredLLDPInformation = {
			"Chassis ID TLV": "network_equipment_interface_mac_address",
			"Port ID TLV": "switch_port_id",
			"Port Description TLV": "switch_port_description",
			"System Name TLV": "switch_hostname",
			"System Description TLV": "network_equipment_description",
			"Port VLAN ID TLV": "switch_vlan_id",
			"VLAN Name TLV": "switch_vlan_name"	
		};

		const objTempLLDPInformation = {};
		const objLLDPInformation = {};

		objLLDPInformation["network_equipment_interface_lldp_information"] = strLLDPInformation;

		arrLLDPRows.forEach((strLLDPRow) =>
		{
			let arrLLDPRow = strLLDPRow.split("\t");

			if(arrLLDPRow.length >= 2)
			{
				objTempLLDPInformation[arrLLDPRow[0]] = arrLLDPRow.shift() && arrLLDPRow.join("");
			}
		});

		for(let strLLDPKey in objTempLLDPInformation)
		{
			if(objRequiredLLDPInformation[strLLDPKey])
			{
				let strLLDPValue = objTempLLDPInformation[strLLDPKey];
				if(strLLDPValue.includes(": "))
				{
					strLLDPValue = strLLDPValue.split(": ")[1];
				}

				objLLDPInformation[objRequiredLLDPInformation[strLLDPKey]] = strLLDPValue;
			}
		}
	}

	get parsedOutput()
	{
		return this._objParsedCommandOutput;
	}

	static get osPlatform()
	{
		return "linux";
	}

	async test()
	{
		
	}
}

module.exports = new NetworkInterfaces({bMustBeFiltered: false, bMustBeParsed: true});
