//@ts-check

const CommandExecutor = require("../CommandExecutor");

class IPMISetup
{
	constructor()
	{}

	/**
	 * This function will set up IPMI for the server.
	 * IPMI credentials can be pregenerated or allocated by the agent.
	 * 
	 * @param {Object} [objConfig]
	 * @param {Encryption} serverEncryption
	 * @returns {Promise<Object>}
	 */
	static async run(objConfig, serverEncryption)
	{
		const {
			objServer,
			objOOBSubnet,
			strIPMIUsername = IPMISetup.allocateIPMIUsername(),
			strIPMIPassword = IPMISetup.allocateIPMIPassword(16),
			strInternalAdministratorUserID = IPMISetup.INTERNAL_ADMINISTRATOR_USER_ID
		} = objConfig;

		let strCommand = `
			IPMITOOL=/usr/bin/ipmitool
			for channel in 0 1 2 3 4
			do
				result=\`$IPMITOOL lan print $channel\`
				if [ -n "$result" ]; then
					ipmichannel=$channel
				fi
			done
			echo "${IPMISetup.END_IPMICHANNEL}\${ipmichannel}${IPMISetup.START_IPMICHANNEL}"
		`;

		if(objServer && objServer.server_status !== IPMISetup.SERVER_STATUS_USED_REGISTERING)
		{
			if(!objOOBSubnet)
			{
				throw new Error("Empty object objOOBSubnet. Must provide OOBSubnet information.");
			}

			let strCommandExtension = `
				$IPMITOOL user set name ${strInternalAdministratorUserID} ${strIPMIUsername}
				$IPMITOOL user set password ${strInternalAdministratorUserID} ${strIPMIPassword}
				$IPMITOOL user priv ${strInternalAdministratorUserID} ${IPMISetup.PRIVILEGE_ADMINISTRATOR} $ipmichannel
				$IPMITOOL lan set $ipmichannel ipsrc static
				$IPMITOOL lan set $ipmichannel ipaddr ${objServer["server_ipmi_host"]}
				$IPMITOOL lan set $ipmichannel netmask ${objOOBSubnet["subnet_oob_netmask_human_readable"]}
				$IPMITOOL lan set $ipmichannel defgw ipaddr ${objOOBSubnet["subnet_oob_gateway_human_readable"]}
				$IPMITOOL lan set $ipmichannel access on
			`;

			// disabling default administrator user
			strCommandExtension += `
				$IPMITOOL user disable ${IPMISetup.DEFAULT_ADMINISTRATOR_USER_ID}
				$IPMITOOL bmc reset warm
			`;

			strCommand += strCommandExtension;
		}

		const nIPMIChannel = await IPMISetup._executeCommand(strCommand);

		const objIPMISettings = {
			server_ipmi_internal_username: strIPMIUsername,
			server_ipmi_internal_password_encrypted: serverEncryption.encrypt(strIPMIPassword),
			server_ipmi_version: 2,
			server_ipmi_channel: nIPMIChannel,
			datacenter_name: objServer.datacenter_name
		};

		return objIPMISettings;
	}


	static allocateIPMIUsername()
	{
		return `bsi${IPMISetup.allocateIPMIPassword(5)}`;
	}

	/**
	* Generates a 16 characters password for IPMI server authentication.
	* Should be used by IPMI setup functions.
	*
	* @param {number} nPasswordLength
	* @returns {string}
	*/
	static allocateIPMIPassword(nPasswordLength = 16)
	{
		return IPMISetup._generateRandomString(nPasswordLength);
	}


	/**
	 * Generates a random string.
	 * characterList - possible characters
	 *
	 * @param {number} [nLength=16]
	 * @param {{characterList: string}|any} [objOptions={}]
	 *
	 * @returns {string}
	 */
	static _generateRandomString(nLength = 16, objOptions = {})
	{
		const {
			characterList = "23456789abcdefghijkmnpqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ" //Not present characters: o, O, 0 (zero), 1 (one), l (L).
		} = objOptions;


		let strRandomText = "";
		for(let i = 0; i < nLength; i++)
		{
			strRandomText += characterList.charAt(Math.floor(Math.random() * characterList.length));
		}

		return strRandomText;
	}


	/**
	 * Executes the given commands and parses the output to extract the ipmichannel.
	 * 
	 * @param {string} strCommand 
	 * @returns {Promise<number>}
	 */
	static async _executeCommand(strCommand)
	{
		const strCommandOutput = await CommandExecutor.execCommandProcess(strCommand);
		const strIPMIChannel = strCommandOutput
			.split(IPMISetup.START_IPMICHANNEL)
			.pop()
			.split(IPMISetup.END_IPMICHANNEL)
			.shift();
		const nIPMIChannel = parseInt(strIPMIChannel, 10);

		if(Number.isNaN(nIPMIChannel))
		{
			throw new Error("IPMI channel received from the setup script is invalid.");
		}

		return nIPMIChannel;
	}

	/**
	 * @returns {string}
	 */
	static get SERVER_STATUS_USED_REGISTERING()
	{
		return "used_registering";
	}

	/**
	 * @returns {number}
	 */
	static get PRIVILEGE_ADMINISTRATOR()
	{
		return 4;
	}

	/**
	 * Function that returns the id of the internal Administrator user in the user list. This user is always present to be used to control the server.
	 * The default value for this is 2.
	 *
	 * @returns {number}
	 */
	static get INTERNAL_ADMINISTRATOR_USER_ID()
	{
		return 2;
	}

	/**
	 * Function that returns the id of the default Administrator user in the user list. This user is present after a factory reset of the server.
	 * The default value for this is 1.
	 *
	 * @returns {number}}
	*/
	static get DEFAULT_ADMINISTRATOR_USER_ID()
	{
		return 1;
	}

	/**
	 * @returns {string}
	 */
	static get START_IPMICHANNEL()
	{
		return "START_IPMICHANNEL";
	}
	
	/**
	 * @returns {string}
	 */
	static get END_IPMICHANNEL()
	{
		return "END_IPMICHANNEL";
	}
}

module.exports = IPMISetup;
