const os = require("os");
const getos = require("getos");
const NetworkInterface = require("./Models/NetworkInterface");
const OSInfo = require("./Models/OSInfo");
const AtomicWriteFileSystemSingleton = require("../FileSystem").AtomicWriteFileSystemSingleton;

class FactsProvider
{
	constructor()
	{
		this.arrPaths = {
			"interfaces": "/sys/class/net/",
			"resolvConf": "/etc/resolv.conf"
		};
	}

	/**
	 * @returns {OSInfo}
	 */
	async getOSInfo()
	{
		let fnResolve;
		let fnReject;
		const promise = new Promise((fnPromiseResolve, fnPromiseReject) =>
		{
			fnResolve = fnPromiseResolve;
			fnReject = fnPromiseReject;
		});

		getos((error, response) =>
		{
			if (error !== null)
			{
				fnReject(error);
				return;
			}

			const osInfo = new OSInfo(
				response.os,
				response.dist,
				response.release,
				os.arch()
			);

			if (os.platform() === "win32") // for windows 64 bit, still win32 ?!?
			{
				osInfo.release = os.release();
				osInfo.distribution = "Windows";
			}

			fnResolve(osInfo);
		});

		return promise;
	}

	/**
	 * @param {string} strRequiredType
	 * @returns {Array}
	 */
	async getInterfaces(strRequiredType)
	{
		strRequiredType = strRequiredType || null;

		const arrInterfaces = [];
		const arrFiles = await AtomicWriteFileSystemSingleton.readdir(this.getPath("interfaces"));
		for (const strFile of arrFiles)
		{
			if (strFile === "bonding_masters")
			{
				continue;
			}

			let strInterfaceType = null;
			let strRealPath = await AtomicWriteFileSystemSingleton.realpath(this.getPath("interfaces") + strFile);
			const bIsVirtualInterface = (strRealPath.indexOf("devices/pci") === -1);

			if (bIsVirtualInterface)
			{
				strInterfaceType = NetworkInterface.TYPE_VIRTUAL;
			}
			else
			{
				strInterfaceType = NetworkInterface.TYPE_PHYSICAL;
			}

			let strMACAddress = await AtomicWriteFileSystemSingleton.readFile(strRealPath + "/" + "address") || null;

			if (strMACAddress !== null)
			{
				strMACAddress = strMACAddress.replace(/\n$/, "");
				strMACAddress = strMACAddress.trim();
			}

			const networkInterface = new NetworkInterface(strFile, strInterfaceType, strMACAddress);

			switch (strRequiredType)
			{
				case NetworkInterface.TYPE_PHYSICAL:
					if (strInterfaceType === NetworkInterface.TYPE_PHYSICAL)
					{
						arrInterfaces.push(networkInterface);
					}
					break;
				case NetworkInterface.TYPE_VIRTUAL:
					if (strInterfaceType === NetworkInterface.TYPE_VIRTUAL)
					{
						arrInterfaces.push(networkInterface);
					}
					break;
				default:
					arrInterfaces.push(networkInterface);
			}
		}

		return arrInterfaces;
	}

	/**
	 * @returns {Array}
	 */
	async getResolvNameServers()
	{
		const strResolvConfContent = await AtomicWriteFileSystemSingleton.readFile(this.getPath("resolvConf"), {
			"encoding": "utf-8"
		});

		const match = strResolvConfContent.match(new RegExp("search (.*)"));
		if (!match)
		{
			return [];
		}

		return match[1].split(" ");
	}

	/**
	 * @param {string} strName
	 */
	getPath(strName)
	{
		return this.arrPaths[strName];
	}
}


module.exports = FactsProvider;
