const os = require("os");
const path = require("path");
const fsExtra = require("fs-extra");

/**
 * Proxies and enhances node native fs.
 * (Almost) same function signatures as native fs
 */
class AtomicWriteFileSystem
{
	/**
	 * @param {string} strPath
	 * @returns {string}
	 */
	async readFile(strPath)
	{
		return await fsExtra.readFile(strPath, {
			"encoding": "utf-8"
		});
	}

	/**
	 * @param {string} strPath
	 * @param {string} strContent
	 * @param {Object} settings
	 * @private
	 */
	async writeFile(strPath, strContent, settings)
	{
		settings = settings || {};
		settings.overwrite = settings.overwrite || false;
		settings.mode = settings.mode || null;
		settings.groupID = settings.groupID || null;
		settings.userID = settings.userID || null;

		const strTmpPath = this._getTmpFilePath(strPath);

		try
		{
			await fsExtra.writeFile(strTmpPath, strContent);
			await this.move(strTmpPath, strPath, {overwrite: settings.overwrite});
		}
		finally
		{
			await this.unlink(strTmpPath);
		}

		if (settings.groupID != null && settings.userID != null)
		{
			await this.chown(strPath, settings.userID, settings.groupID);
		}

		if (settings.mode != null)
		{
			await this.chmod(strPath, settings.mode);
		}
	}

	/**
	 * @param {string} strPath
	 * @param {string} strContent
	 */
	async appendFile(strPath, strContent)
	{
		const strTmpPath = this._getTmpFilePath(strPath);

		try
		{
			await this.copy(strPath, strTmpPath);
			await fsExtra.appendFile(strTmpPath, strContent);
			await this.move(strTmpPath, strPath, {overwrite: true});
		}
		finally
		{
			await this.unlink(strTmpPath);
		}
	}

	/**
	 * @param {string} strPath
	 * @param {string} strLineContent
	 */
	async appendFileLine(strPath, strLineContent)
	{
		return await this.appendFile(strPath, `${strLineContent}${os.EOL}`);
	}

	/**
	 *
	 * @param {string} strPath
	 * @returns {string}
	 */
	async ensureDir(strPath)
	{
		return await fsExtra.ensureDir(strPath);
	}

	/**
	 * @param {string} strPath
	 * @returns {ReadStream}
	 */
	async createReadStream(strPath)
	{
		return await fsExtra.createReadStream(strPath);
	}

	/**
	 * @param {string} strPath
	 * @returns {Array}
	 */
	async readdir(strPath)
	{
		return await fsExtra.readdir(strPath);
	}

	/**
	 * @param {string} strPath
	 * @param {number} nMode
	 * @returns {number}
	 */
	async chmod(strPath, nMode)
	{
		return await fsExtra.chmod(strPath, nMode);
	}

	/**
	 * @param {string} strPath
	 * @param {number} nUserID
	 * @param {number} nGroupID
	 * @returns {number}
	 */
	async chown(strPath, nUserID, nGroupID)
	{
		return await fsExtra.chown(strPath, nUserID, nGroupID);
	}

	/**
	 * @param {string} strPath
	 * @returns {Object}
	 */
	async stat(strPath)
	{
		return await fsExtra.stat(strPath);
	}

	/**
	 * @param {string} strPath
	 * @returns {boolean}
	 */
	async exists(strPath)
	{
		return await fsExtra.exists(strPath);
	}

	/**
	 * @param {string} strPath
	 * @returns {string}
	 */
	async realpath(strPath)
	{
		return await fsExtra.realpath(strPath);
	}

	/**
	 * @param {string} strTarget
	 * @param {string} strSource
	 * @returns {number}
	 */
	async symlink(strTarget, strSource)
	{
		return await fsExtra.symlink(strTarget, strSource);
	}

	/**
	 * @param {string} strPath
	 * @returns {number}
	 */
	async unlink(strPath)
	{
		try
		{
			await fsExtra.unlink(strPath);
		}
		catch (error)
		{
			if (error.code !== "ENOENT") // path does not exist
			{
				throw error;
			}
		}
	}


	/**
	 * @param {string} strSource
	 * @param {string} strDestination
	 *
	 * @returns {number}
	 */
	async copy(strSource, strDestination)
	{
		return await fsExtra.copy(strSource, strDestination);
	}

	/**
	 * @param {string} strSource
	 * @param {string} strDestination
	 * @param {Object} objParameters
	 *
	 * @returns {number}
	 */
	async move(strSource, strDestination, objParameters)
	{
		return await fsExtra.move(strSource, strDestination, objParameters);
	}

	/**
	 * @param {string} strFilePath
	 * @private
	 */
	_getTmpFilePath(strFilePath)
	{
		return path.join(os.tmpdir(), path.basename(strFilePath) + ".agent.tmp");
	}
}

module.exports = AtomicWriteFileSystem;
