const jsesc = require("jsesc");
const assert = require("assert");
const BaseMountManager = require("./BaseMountManager");
const DiskPartitionMountTemplate = require("../Agent/Modules/Singletons/Template/DiskPartitionMountTemplateSingleton");
const DiskPartitionAutoMountTemplate = require("../Agent/Modules/Singletons/Template/DiskPartitionAutoMountTemplateSingleton");
const SystemdSystemManagerSingleton = require("../SystemServiceManager").SystemdSingleton;
const AtomicWriteFileSystemSingleton = require("../FileSystem").AtomicWriteFileSystemSingleton;
const FileModuleSingleton = require("../Agent/Modules/Singletons/FileSingleton");

class Systemd extends BaseMountManager
{
	constructor()
	{
		super();
	}

	/**
	 * @returns {boolean}
	 */
	async isUsable()
	{
		return await SystemdSystemManagerSingleton.isUsable();
	}

	/**
	 * Creates a mount file, for example: /etc/systemd/system/<escaped-mount-point>.mount and starts its service.
	 * 
	 * @param {string} strMountPoint
	 * @param {string} strDiskPath
	 * @param {string} strFilesystemType
	 * @throws {Error}
	 */
	async mount(strMountPoint, strDiskPath, strFilesystemType)
	{
		const strPrefix = ".mount";
		const strMountFilePath = this._getFileName(strMountPoint, strPrefix);

		//If mount file does not exists, create it
		if(!await AtomicWriteFileSystemSingleton.exists(strMountFilePath))
		{
			await FileModuleSingleton.save("", strMountFilePath);
		}

		await DiskPartitionMountTemplate.renderToFile(
			strMountFilePath,
			{
				"description": "Generic drive mount file",
				"diskPath": strDiskPath,
				"mountPoint": strMountPoint,
				"fileSystem": strFilesystemType
			},
			{mode: 0o755}
		);

		let strMountServiceName = strMountFilePath.substring(strMountFilePath.lastIndexOf("/") + 1);

		//Add another '\' to each '\' found in the name, because commands with systemctl won't work otherwise.
		strMountServiceName = strMountServiceName.replace(/["\\"]/g, "\\\\");

		await SystemdSystemManagerSingleton.reloadDaemon();
		await SystemdSystemManagerSingleton.start(strMountServiceName);
	}

	/**
	 * Creates a mount file, for example: /etc/systemd/system/<escaped-mount-point>.mount and starts its service.
	 * 
	 * @param {string} strMountPoint
	 * @param {string} strDiskPath
	 * @param {string} strFilesystemType
	 * @throws {Error}
	 * @returns {boolean}
	 */
	async isMounted(strMountPoint, strDiskPath, strFilesystemType)
	{
		const strPrefix = ".mount";
		const strMountFilePath = this._getFileName(strMountPoint, strPrefix);

		if(!await AtomicWriteFileSystemSingleton.exists(strMountFilePath))
		{
			return false;
		}

		if(
			(strDiskPath !== undefined && strDiskPath !== null) && 
			(strFilesystemType !== undefined && strFilesystemType !== null)
		)
		{
			const strFileContent = await AtomicWriteFileSystemSingleton.readFile(strMountFilePath);
			const strExpectedFileContent = await DiskPartitionMountTemplate.getRenderedTemplate(
				{
					"description": "Generic drive mount file",
					"diskPath": strDiskPath,
					"mountPoint": strMountPoint,
					"fileSystem": strFilesystemType
				}
			);

			if(strFileContent.trim() !== strExpectedFileContent.trim())
			{
				return false;
			}
		}

		let strMountServiceName = strMountFilePath.substring(strMountFilePath.lastIndexOf("/") + 1);

		//Add another '\' to each '\' found in the name, because commands with systemctl won't work otherwise.
		strMountServiceName = strMountServiceName.replace(/["\\"]/g, "\\\\");

		return await SystemdSystemManagerSingleton.isRunning(strMountServiceName);
	}

	/**
	 * Unmounts and removes the mount file from /etc/systemd/system/ and stops the service.
	 * 
	 * @param {string} strMountPoint
	 */
	async unmount(strMountPoint)
	{
		const strPrefix = ".mount";
		const strMountFilePath = this._getFileName(strMountPoint, strPrefix);

		if(!await AtomicWriteFileSystemSingleton.exists(strMountFilePath))
		{
			throw new Error("Cannot unmount " + strMountFilePath + ", because file does not exist!");
		}

		let strMountServiceName = strMountFilePath.substring(strMountFilePath.lastIndexOf("/") + 1);

		//Add another '\' to each '\' found in the name, because commands with systemctl won't work otherwise.
		strMountServiceName = strMountServiceName.replace(/["\\"]/g, "\\\\");

		if(await SystemdSystemManagerSingleton.isRunning(strMountServiceName))
		{
			await SystemdSystemManagerSingleton.stop(strMountServiceName);
		}

		await AtomicWriteFileSystemSingleton.unlink(strMountFilePath);
		await SystemdSystemManagerSingleton.reloadDaemon();
	}

	/**
	 * Create a mount and automount file, for example: /etc/systemd/system/<escaped-mount-point>.[auto]mount
	 * and starts and enables its service.
	 * 
	 * @param {string} strMountPoint
	 * @param {string} strDiskPath
	 * @param {string} strFilesystemType
	 * @throws {Error}
	 * @returns {string}
	 */
	async mountAndSetAutomount(strMountPoint, strDiskPath, strFilesystemType)
	{
		await this.mount(strMountPoint, strDiskPath, strFilesystemType);

		const strPrefix = ".automount";
		const strAutoMountFilePath = this._getFileName(strMountPoint, strPrefix);

		//If automount file does not exists, create it
		if(!await AtomicWriteFileSystemSingleton.exists(strAutoMountFilePath))
		{
			await FileModuleSingleton.save("", strAutoMountFilePath);
		}

		await DiskPartitionAutoMountTemplate.renderToFile(
			strAutoMountFilePath,
			{
				"description": "Generic drive automatic mount file",
				"diskPath": strDiskPath,
				"mountPoint": strMountPoint
			},
			{mode: 0o755}
		);

		let strAutoMountServiceName = strAutoMountFilePath.substring(strAutoMountFilePath.lastIndexOf("/") + 1);

		//Add another '\' to each '\' found in the name, because commands with systemctl won't work otherwise.
		strAutoMountServiceName = strAutoMountServiceName.replace(/["\\"]/g, "\\\\");

		await SystemdSystemManagerSingleton.reloadDaemon();
		await SystemdSystemManagerSingleton.start(strAutoMountServiceName);
		await SystemdSystemManagerSingleton.enable(strAutoMountServiceName);

		return strAutoMountServiceName;
	}

	/**
	 * Checks whether mount and automount service files exist and if automount is active.
	 *
	 * @param {string} strMountPoint
	 * @param {string} strDiskPath
	 * @param {string} strFilesystemType
	 * @throws {Error}
	 * @returns {boolean}
	 */
	async isAutomountSet(strMountPoint, strDiskPath, strFilesystemType)
	{
		if(!await this.isMounted(strMountPoint, strDiskPath, strFilesystemType))
		{
			return false;
		}

		const strPrefix = ".automount";
		const strAutoMountFilePath = this._getFileName(strMountPoint, strPrefix);

		if(!await AtomicWriteFileSystemSingleton.exists(strAutoMountFilePath))
		{
			return false;
		}

		if(
			(strDiskPath !== undefined && strDiskPath !== null) && 
			(strFilesystemType !== undefined && strFilesystemType !== null)
		)
		{
			const strFileContent = await AtomicWriteFileSystemSingleton.readFile(strAutoMountFilePath);
			const strExpectedFileContent = await DiskPartitionAutoMountTemplate.getRenderedTemplate(
				{
					"description": "Generic drive automatic mount file",
					"diskPath": strDiskPath,
					"mountPoint": strMountPoint
				}
			);
			if(strFileContent.trim() !== strExpectedFileContent.trim())
			{
				return false;
			}
		}

		let strAutoMountServiceName = strAutoMountFilePath.substring(strAutoMountFilePath.lastIndexOf("/") + 1);

		//Add another '\' to each '\' found in the name, because commands with systemctl won't work otherwise.
		strAutoMountServiceName = strAutoMountServiceName.replace(/["\\"]/g, "\\\\");

		return await SystemdSystemManagerSingleton.isEnabled(strAutoMountServiceName);
	}

	/**
	 * Unmounts and removes the mount and automount files from /etc/systemd/system/ and stops and disables their services.
	 * 
	 * @param {string} strMountPoint
	 */
	async unmountAndUnsetAutomount(strMountPoint)
	{
		let strPrefix = ".automount";
		const strAutoMountFilePath = this._getFileName(strMountPoint, strPrefix);

		if(!await AtomicWriteFileSystemSingleton.exists(strAutoMountFilePath))
		{
			throw new Error("Cannot unmount " + strAutoMountFilePath + ", because file does not exist!");
		}

		let strAutoMountServiceName = strAutoMountFilePath.substring(strAutoMountFilePath.lastIndexOf("/") + 1);

		//Add another '\' to each '\' found in the name, because commands with systemctl won't work otherwise.
		strAutoMountServiceName = strAutoMountServiceName.replace(/["\\"]/g, "\\\\");

		await SystemdSystemManagerSingleton.stop(strAutoMountServiceName);
		await SystemdSystemManagerSingleton.disable(strAutoMountServiceName);

		strPrefix = ".mount";
		const strMountFilePath = this._getFileName(strMountPoint, strPrefix);

		//Remove the mount file it it exists.
		if(await AtomicWriteFileSystemSingleton.exists(strMountFilePath))
		{
			await this.unmount(strMountPoint);
		}

		await AtomicWriteFileSystemSingleton.unlink(strAutoMountFilePath);
		await SystemdSystemManagerSingleton.reloadDaemon();
	}

	/**
	 * Transforms a mount point name to a valid one used in /etc/systemd/system.
	 * Example: /mnt/a/b/c -> /mnt-a-b-c.[auto]mount
	 * Prefix can be .mount or .automount
	 * 
	 * @param {string} strMountPoint
	 * @param {string} strPrefix
	 * @throws {Error}
	 * @returns {string}
	 */
	_getFileName(strMountPoint, strPrefix)
	{
		const arrAllowedPrefixes = [".mount", ".automount"];

		if(!arrAllowedPrefixes.includes(strPrefix))
		{
			throw new Error("Given prefix " + strPrefix + " is not a valid one. Allowed prefixes are " + JSON.stringify(arrAllowedPrefixes));
		}

		const strSystemDPoint = this._escapeUnitName(strMountPoint);

		const strSystemDFileName = this.systemDPath + strSystemDPoint + strPrefix;

		return strSystemDFileName;
	}

	/**
	 * Escape strings for usage in systemd unit names.
	 * Given a path, "/" is replaced by "-", and all other characters which are not ASCII
	 * alphanumerics are replaced by C-style "\x2d" escapes (except that "_"
	 * is never replaced and "." is only replaced when it would be the first
	 * character in the escaped path). The root directory "/" is encoded as
	 * single dash, while otherwise the initial and ending "/" are removed
	 * from all paths during transformation.
	 * The escaping is irreversible.
	 * This function can also be present in the systemd system service manager file.
	 * 
	 * @param {string} strUnitName
	 *
	 * @returns {string}
	 */
	_escapeUnitName(strUnitName)
	{
		assert(typeof strUnitName === "string", "SystemD unit name must be a string. Instead it is of type " + typeof strUnitName + ".");

		//If '/' is the first or last character, remove it
		if (strUnitName.indexOf("/") === 0) 
		{
			strUnitName = strUnitName.substring(1);
		}
		if (strUnitName.lastIndexOf("/") === strUnitName.length - 1) 
		{
			strUnitName = strUnitName.substring(0, strUnitName.length - 1);
		}

		let chValue;
		for (let nIndex = 0; nIndex < strUnitName.length; nIndex++) 
		{
			chValue = strUnitName.charAt(nIndex);
			if (chValue === "/") 
			{
				strUnitName = this._replaceCharAt(strUnitName, nIndex, "-");
			}
			else if (!this._isAlphanumeric(chValue)) 
			{
				if (chValue === "." && nIndex !== 0) 
				{
					continue;
				}

				//Manually enter code for '\' because jsesc doesn't know to escape it...
				if (chValue === "\\") 
				{
					strUnitName = this._replaceCharAt(
						strUnitName,
						nIndex,
						"\\x5c"
					);
					continue;

				}

				strUnitName = this._replaceCharAt(
					strUnitName,
					nIndex,
					jsesc(chValue,
						{
							"escapeEverything": true
						}
					).toLowerCase()
				);
			}
		}

		return strUnitName;
	}

	/**
	 * Checks if given input is alphanumeric or not.
	 * 
	 * @param {string} strInput
	 *
	 * @returns {boolean}
	 */
	_isAlphanumeric(strInput) 
	{
		return !/[^0-9a-z\xDF-\xFF]/.test(strInput.toLowerCase());
	}

	/**
	 * Replaces the character at strInput[nIndex] with strReplacement.
	 * 
	 * @param {string} strInput
	 * @param {number} nIndex
	 * @param {string} strReplacement
	 *
	 * @returns {string}
	 */
	_replaceCharAt(strInput, nIndex, strReplacement) 
	{
		return strInput.substr(0, nIndex) + strReplacement + strInput.substr(nIndex + 1);
	}

	/**
	 * @returns {string}
	 */
	get systemDPath()
	{
		return "/etc/systemd/system/";
	}
}

module.exports = Systemd;
