const LoggedChildProcessSingleton = require("../LoggedChildProcessSingleton");
const AtomicWriteFileSystemSingleton = require("../FileSystem/AtomicWriteFileSystemSingleton");

/**
 * @see https://github.com/ansible/ansible/blob/devel/lib/ansible/modules/system/service.py
 */
class BaseSystemServiceManager
{
	constructor()
	{
		this._nRetryCount = 0;
		this._nRetryMaximumCount = 5;
		this._nRetryTimeout = 500; //miliseconds
	}

	/**
	 * @returns {string}
	 */
	getBinaryName()
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @returns {boolean}
	 */
	async isUsable()
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @param {string} strServiceName
	 * @param {Array} arrArguments
	 */
	async start(strServiceName, arrArguments)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @param {string} strServiceName
	 */
	async stop(strServiceName)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @param {string} strServiceName
	 * @param {Array} arrArguments
	 */
	async restart(strServiceName, arrArguments)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @param {string} strServiceName
	 */
	async enable(strServiceName)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @param {string} strServiceName
	 */
	async disable(strServiceName)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @param {string} strServiceName
	 */
	async reload(strServiceName)
	{
		throw new Error("Must be implemented");
	}

	/**
	 * @param {string} strName
	 * @param {Array} arrPaths
	 * @returns {string}
	 */
	async findFullPath(strName, arrPaths)
	{
		for (const strPath of arrPaths)
		{
			const strFullPath = `${strPath}/${strName}`;
			if (await AtomicWriteFileSystemSingleton.exists(strFullPath))
			{
				return strFullPath;
			}
		}

		return null;
	}

	/**
	 * @param {string} strCommand
	 * @param {Function} fnShouldRetry
	 * @returns {Promise}
	 */
	async executeWithRetry(strCommand, fnShouldRetry)
	{
		let fnResolve;
		let fnReject;
		let strStdout = "";
		let strStderr = "";

		const promise = new Promise((fnPromiseResolve, fnPromiseReject) =>
		{
			fnResolve = fnPromiseResolve;
			fnReject = fnPromiseReject;
		});
		const childProcess = await LoggedChildProcessSingleton.spawn(strCommand, {shell: true});

		childProcess.stdout.on("data", (data) =>
		{
			strStdout += data;
		});
		childProcess.stderr.on("data", (data) =>
		{
			strStderr += data;
		});

		childProcess.on("error", (error) =>
		{
			fnReject(new Error(`Could not execute command: ${strCommand}`, error));
		});

		childProcess.on("exit", (nResultCode, strSignal) =>
		{
			strStdout = strStdout.trim();

			if (fnShouldRetry(strStdout, strStderr, nResultCode, strSignal))
			{
				if (this._nRetryCount > this._nRetryMaximumCount)
				{
					fnReject(new Error("Max retry count exceeded!"));
					return;
				}

				setTimeout(
					async() =>
					{
						this._nRetryCount++;
						fnResolve(await this.executeWithRetry(strCommand, fnShouldRetry));
					},
					this._nRetryTimeout
				);

				return;
			}

			this._nRetryCount = 0;
			fnResolve(strStdout);
		});

		return promise;
	}
}

module.exports = BaseSystemServiceManager;
