const BaseSystemServiceManager = require("./BaseSystemServiceManager");
const AtomicWriteFileSystemSingleton = require("../FileSystem").AtomicWriteFileSystemSingleton;
const LoggedChildProcessSingleton = require("../LoggedChildProcessSingleton");

class Systemd extends BaseSystemServiceManager 
{
	/**
	 * @returns {string}
	 */
	getBinaryName() 
	{
		return "systemctl";
	}

	/**
	 * @returns {boolean}
	 */
	async isUsable() 
	{
		if (await this.getFullBinaryPath() === null) 
		{
			return false;
		}

		// this should show if systemd is the boot init system
		const arrCanaries = ["/run/systemd/system/", "/dev/.run/systemd/", "/dev/.systemd/"];
		for (const strCanary of arrCanaries)
		{
			if (await AtomicWriteFileSystemSingleton.exists(strCanary))
			{
				return true;
			}
		}

		// If all else fails, check if init is the systemd command, using comm as cmdline could be symlink
		const strContent = await AtomicWriteFileSystemSingleton.readFile("/proc/1/comm");
		if (strContent.indexOf("systemd") !== -1)
		{
			return true;
		}

		return false;
	}

	/**
	 * @returns {string|null}
	 */
	async getFullBinaryPath() 
	{
		const arrBinPaths = ["/sbin", "/usr/sbin", "/bin", "/usr/bin"];

		return await this.findFullPath(this.getBinaryName(), arrBinPaths);
	}

	/**
	 * @param {string} strServiceName
	 */
	async start(strServiceName)
	{
		await LoggedChildProcessSingleton.spawn(`${await this.getFullBinaryPath()} start ${strServiceName}`, {shell: true});
	}

	/**
	 * @param strServiceName
	 */

	async stop(strServiceName)
	{
		await LoggedChildProcessSingleton.spawn(`${await this.getFullBinaryPath()} stop ${strServiceName}`, {shell: true});
	}

	/**
	 * @param {string} strServiceName
	 */
	async enable(strServiceName)
	{
		await LoggedChildProcessSingleton.spawn(`${await this.getFullBinaryPath()} enable ${strServiceName}`, {shell: true});
	}

	/**
	 * @param {string} strServiceName
	 */
	async disable(strServiceName)
	{
		await LoggedChildProcessSingleton.spawn(`${await this.getFullBinaryPath()} disable ${strServiceName}`, {shell: true});
	}

	/**
	 * @param {string} strServiceName
	 */
	async reload(strServiceName) 
	{
		if (!await this.isRunning(strServiceName)) 
		{
			throw new Error("Can not reload a service that is not running!");
		}


		await LoggedChildProcessSingleton.spawn(`${await this.getFullBinaryPath()} reload ${strServiceName}`, {shell: true});
	}

	/**
	 * @param {string} strServiceName
	 */
	async restart(strServiceName)
	{
		await LoggedChildProcessSingleton.spawn(`${await this.getFullBinaryPath()} restart ${strServiceName}`, {shell: true});
	}

	/**
	 * @param {string} strServiceName
	 *
	 * @returns {boolean}
	 */
	async isRunning(strServiceName) 
	{
		const strCommand = `${await this.getFullBinaryPath()} is-active ${strServiceName}`;
		const strStdoutResponse = await this.executeWithRetry(strCommand, (strStdout) => strStdout === "activating");

		return strStdoutResponse === "active";
	}

	/**
	 * @param {string} strServiceName
	 *
	 * @returns {boolean}
	 */
	async isEnabled(strServiceName) 
	{
		const strCommand = `${await this.getFullBinaryPath()} is-enabled ${strServiceName}`;
		const strStdoutResponse = await this.executeWithRetry(strCommand, (strStdout) => strStdout === "enabling");

		return strStdoutResponse === "enabled";
	}

	async reloadDaemon() 
	{
		await LoggedChildProcessSingleton.spawn(`${await this.getFullBinaryPath()} daemon-reload`, { shell: true });
	}
}

module.exports = Systemd;
